#include "candidate.h"

Candidate::Candidate(int i, int j, bool _is_one) {
	is_one = _is_one;
	if (i > j) {
		candidate_pair.first = j;
		candidate_pair.second = i;
	}
	else {
		candidate_pair.first = i;
		candidate_pair.second = j;
	}

}

Candidate::Candidate(const std::pair<int, int> &_candidate_pair, bool _is_one) {
	is_one = _is_one;
	candidate_pair.first = _candidate_pair.first;
	candidate_pair.second = _candidate_pair.second;
}

bool Candidate::operator<(const Candidate &cand2) const {
	return (candidate_pair.first < cand2.candidate_pair.first || 
			(candidate_pair.first == cand2.candidate_pair.first && candidate_pair.second < cand2.candidate_pair.second));
}

// searches candidates for a pair (a,b) a<=b, resp. (b,a) b< a
bool is_summable_candidate(const std::set<Candidate> &candidates, int a, int b) {
	std::pair<int, int> my_pair;
	
	if (a <= b) {
		my_pair.first = a;
		my_pair.second = b;
	}
	else {
		my_pair.first = b;
		my_pair.second = a;
	}

	return ( candidates.find(Candidate(my_pair, false)) != candidates.end() );
}


// pair = (x,y) is a canditate pair, i.e. x + y is defined
// then 
//	x must be summable with all elements of y + 0 = 0 + y
// 		and
//	y must be summable with all elements of 0 + x
//
// ! asserts that all pairs (0,x) are already defined !

bool check_summability(const std::set<Candidate> &candidates, const std::pair<int, int> &test_pair, const std::set<int> *(&op)[MAX][MAX]) {

	// x is summable with all elements in y + 0, if x + y is summable
	for (std::set<int>::const_iterator it_set = op[0][test_pair.second]->begin(); it_set != op[0][test_pair.second]->end(); ++it_set) {
		if (*it_set == 0)
			continue;

		if (!is_summable_candidate(candidates, test_pair.first, *it_set))
			return false;
	}

	if (test_pair.first != test_pair.second) {
		// y is summable with all elements in x + 0, if x + y is summable
		for (std::set<int>::const_iterator it_set = op[0][test_pair.first]->begin(); it_set != op[0][test_pair.first]->end(); ++it_set) {
			if (*it_set == 0)
				continue;

			if (!is_summable_candidate(candidates, test_pair.second, *it_set))
				return false;
		}
	}

	return true;
}

