#include <iomanip>
#include <string.h>
#include <stdlib.h>
#include <fstream>
#include <vector>

#include "hyperEA3_common.h"
#include "hea.h"
#include "hyperEA3_parse.h"
	

using namespace std;


#if 0

void disp_elem(const int n, int elem) {
	if (elem == 0) {
		cout << 0;
		return;
	}
	if (elem == n - 1) {
		cout << 1;
		return;
	}

	if (elem >=0 && elem < n) {
		cout << (char) ('a' + elem - 1);
		return;
	}
}


void iso_image(const int n, const set<int> * const (&op)[MAX][MAX], const int (&perm)[MAX], const set<int> *setCandidates, const set<int>* (&op2)[MAX][MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++) {
		if (hea::is_summable(op, i, j)) {
			int value = 0;
			for (set<int>::const_iterator it = op[i][j]->begin(); it != op[i][j]->end(); ++it) {
				value |= (1 << perm[ *it ]);
				op2[ perm[i] ][ perm[j] ] = &setCandidates[value];
				
			}
		}
		else {
			op2[ perm[i] ][ perm[j] ] = &setCandidates[0]; // empty set
		}
	}
}

bool are_equal_op(const int n, const set<int>* const (&op)[MAX][MAX], const set<int>* const (&op2)[MAX][MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if (op[i][j] != op2[i][j])
			return false;

	return true;
}

bool parse_set_content(const char **str, const int n, int &set_number) {
	const char *pom = *str;

	int tmp_val = 0;
	while (*pom && *pom != '}') {
		// parse one elemet
		// skip ws
		while (*pom && (*pom == ' ' || *pom == '\t') )
			pom++;

		if (*pom=='}') {
			continue;
		}
		else {
			if (*pom == '0' || *pom =='1' || (*pom >='a' && *pom <='z')) {
				switch (*pom) {
					case '0':
						tmp_val |= 1;
						break;
					case '1':
						tmp_val |= (1<< (n-1));
						break;

					default:
						tmp_val |= (1<< (int)(*pom-'a' + 1));
						break;
				}
				pom++;
			}
			else {
				// unexpected character
				cout << "Unexpected character '" << *pom << "'" << endl;
				return false;
			}

			// skip ws
			while (*pom && (*pom == ' ' || *pom == '\t') )
				pom++;
			if (*pom == ',')
				pom++;
		}
	}

	if (*pom == '}') {
		set_number = tmp_val;
		//pom++;
		*str = pom;
		return true;
	}
	
	cout << "HERE *pom=='" << *pom << "'" << endl;
	return false;
}

bool parse_set(const char **str, const int n, int &set_number) {
	const char *pom = *str;
	// skip ws
	while (*pom && (*pom == ' ' || *pom == '\t') )
		pom++;

	if (*pom != '{') {
		// missing openning brace {
		cout << "Missing {" << endl;
		return false;
	}
	else
		pom++;

	if (!parse_set_content(&pom, n, set_number)) {
		cout << "Parse set content error" << endl;
		return false;
	}

	// skip ws
	while (*pom && (*pom == ' ' || *pom == '\t') )
		pom++;
	
	if (*pom != '}') {
		// missing closing brace }
		cout << "Missing }" << endl;
		return false;
	}
	else
		pom++;

	*str = pom;
	return true;
} 

bool parse_line(const char **str, const int n, int *line_nums) {
	const char *pom = *str;
	int idx = 0;

	while (*pom && *pom != '\n' && *pom !='\r') {
		

		int set_number;
		if (parse_set(&pom, n, set_number)) {
			//cout << set_number << " ";
			line_nums[idx] = set_number;
			idx++;
			
		}
		else {
			cout << "Error parsing set" << endl;
			return false;
		}

		// skip ws
		while (*pom && (*pom == ' ' || *pom == '\t') )
			pom++;

		if (*pom==',')
			pom++;

		// skip ws
		while (*pom && (*pom == ' ' || *pom == '\t') )
			pom++;
	}

	//cout <<endl;
	return true;
}

bool parse_complement(ifstream &ifs, const int n, int (&complement)[MAX]) {
	char buffer[1024];

	// skip first line
	if (ifs.good())
		ifs.getline(buffer, 1024);
	else
		// cannot read stream
		return false;

	if (ifs.good())
		ifs.getline(buffer, 1024);
	else
		// cannot read stream
		return false;
	
	//cout << "Buffer: \"" << buffer << "\"" << endl;


	char *pom = &buffer[0];
	int idx = 0;
	int tmp_val = 0;

	while (*pom && idx < n) {
		//cout << "'" << *pom << "'" ;
		// skip ws
		while (*pom && (*pom == ' ' || *pom == '\t') )
			pom++;
		//cout << "'" << *pom << "'" << endl;

		if (*pom == '0' || *pom =='1' || (*pom >='a' && *pom <='z')) {
			switch (*pom) {
				case '0':
					tmp_val = 0;
					break;
				case '1':
					tmp_val = n - 1;
					break;

				default:
					tmp_val = (int)(*pom-'a' + 1);
					break;
			}
			//cout << tmp_val << endl;
			complement[idx] = tmp_val;
			idx++;
			pom++;
		}
		else
			// Unknown character
			return false;
		
		
	}

	if (idx != n)
		// insufficient number of elements read
		return false;

	return true;
}

bool parse_op(ifstream &ifs, const int n, const set<int>* (&op)[MAX][MAX], const set<int>* setCandidates) {
	char buffer[1024];
	const char *pom = buffer;
	int nthline = 0;
	int line_nums[MAX*2];
	
	while (ifs.good() && nthline < n) {
		ifs.getline(buffer, 1024);

		if (ifs.good()) {
			if (parse_line(&pom, n, line_nums)) {
				for (int i = 0; i < n; i++)
					op[nthline][i] = &setCandidates[ line_nums[i] ];
			}
			else
				// error parsing line
				return false;
		
			nthline++;
		}
	}
	return true;

	
}

bool parse_model(ifstream &ifs, const int n, const set<int>* (&op)[MAX][MAX], int (&complement)[MAX], const set<int> *setCandidates) {
	if (parse_complement(ifs, n, complement)) {
		if (parse_op(ifs, n, op, setCandidates))
			return true;
		else
			// error parsing op
			return false;

	}
	// error parsing complement
	return false;
}


void disp_set(const int n, const set<int> &ds) {
/*
	for (set<int>::iterator it = ds.begin(); it != ds.end(); ++it) {
		cout << *it << ", ";
	}
//	cout << endl;
*/

	set<int>::iterator it = ds.begin();
	if (it != ds.end()) {
		//cout << *it;
		disp_elem(n, *it);
		++it;
		do {
			if (it != ds.end()) {
				cout << ", "; 
				disp_elem(n, *it);
				++it;
			}
		} while (it != ds.end());
	}

}
#endif

/*
bool is_summable(const set<int>* const (&op)[MAX][MAX], int a, int b) {
//	cout << "is_summable: " << a << ", " << b << endl;
	return !((op[a][b])->empty());
}

bool is_summable(const set<int>* const (&op)[MAX][MAX], int a, set<int>&set_b) {
	for (set<int>::iterator it = set_b.begin(); it != set_b.end(); ++it)
		if (!is_summable(op, a, *it))
			return false;

	return true;
}

bool is_summable(const set<int>* const (&op)[MAX][MAX], set<int> &set_a, set<int>&set_b) {
	for (set<int>::iterator it1 = set_a.begin(); it1 != set_a.end(); ++it1)
	for (set<int>::iterator it2 = set_b.begin(); it2 != set_b.end(); ++it2)
		if (!is_summable(op, *it1, *it2))
			return false;

	return true;
}
*/


/*
int is_associative(const int n, const set<int> *(&op)[MAX][MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
	for (int k = 0; k < n; k++) {
//		cout << i << ", " << j << ", " << k << endl;

		
		// exists i+j && exists (i+j)+k => exists j+k && exists i+(j+k)
		// and (i+j)+k = i + (j+k)

		if (is_summable(op, i, j)) {
			set<int> tmp (compute_sum(op, i, j));
			if (is_summable(op, k, tmp)) {
				set<int> final1(compute_sum(op, k, tmp));

				if (!is_summable(op, j, k))
					return false;

				set<int> tmp2(compute_sum(op, j, k));

				if (!is_summable(op, i, tmp2))
					return false;

				set<int> final2(compute_sum(op, i, tmp2));

				if (final1 != final2)
					return false;
			}
		}


#

		// exists i+j && exists (i+j)+k <= exists j+k && exists i+(j+k)
		// and (i+j)+k = i + (j+k)
		if (is_summable(op, j, k)) {
			set<int> tmp(compute_sum(op, j, k));
			//cout << j << " + " << k << " = tmp: "; disp_set(tmp); cout << endl;
			if (is_summable(op, i, tmp)) {
				set<int> final1(compute_sum(op, i, tmp));
				//cout << i << " + " << "(" << j << " + " << k << " = final1:" ; disp_set(final1); cout << endl;

 

				if (!is_summable(op, i, j)) {
					//cout << "not i + j" << endl;
					return false;
				}

				set<int> tmp2(compute_sum(op, i, j));

				if (!is_summable(op, k, tmp2)) {
					//cout << "not (i+j)+k" << endl;
					return false;
				}

				//cout << "tmp2: "; disp_set(tmp2); cout << endl;
				set<int> final2(compute_sum(op, k, tmp2));
				//cout << "final2: " ;disp_set(final2); cout << endl;


				if (final1 != final2) {
					//cout << "neq" << endl;
					return false;
				}
			}
		}


	}
//	cout << "HERE" << endl;
	return true;
}
*/

bool has_unique_orthosupplement_elem(const int n, const set<int>* (&op)[MAX][MAX], int a) {
	bool found = false;

	for (int i = 0; i < n ; i++)

		if (op[a][i]->find(n - 1) != op[a][i]->end()) {
			if (!found)
				found = true;
			else
				return false; /* not unique element */
			
		}

	return found;
}

bool has_unique_orthosupplement(const int n, const set<int>* (&op)[MAX][MAX]) {
	for (int i = 0; i < n; i++)
		if (!has_unique_orthosupplement_elem(n, op, i))
			return false;

	return true;
}



void compute_complement(const int n, const set<int>* (&op)[MAX][MAX], int (&complement)[MAX]) {
/*
	cout << "computing complement for: " << endl;
	disp_op(n, op);
	cout << "end complement" << endl;
*/

//cout << "HERE" << endl;
	complement[0] = n - 1;
	complement[n - 1] = 0;

	for (int i = 1; i < n - 1; i++) {
		//cout << "c: " << i;
		for (int j = n - 2; j>=1; j--) {
		//	cout << j << " ";
			//disp_set(*op[i][j]);
			if (op[i][j]->find(n-1) != op[i][j]->end()) {
				complement[i] = j;
				break;
			}
		}
		//cout << endl;

			
		
	}
}


/*
bool is_antisymmetric(const int n, const set<int>* (&op)[MAX][MAX], const int (&complement)[MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if  (is_summable(op, i, complement[j]) && is_summable(op, j, complement[i]) && i!= j)
			return false;

	return true;
}
*/

/*
bool is_transitive(const int n, const set<int>* (&op)[MAX][MAX], const int (&complement)[MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
	for (int k = 0; k < n; k++)
		if  (is_summable(op, i, complement[j]) && is_summable(op, j, complement[k]) && !is_summable(op, i, complement[k]))
			return false;

	return true;
}
*/

#if 0
// initialize set with elements equal to binary representation of number i
// it is expected to call this function for i from 0 to (1 << n) - 1.
//
// n = 4
// 3210
// 0100 -> {2}
// 3210
// 1101 -> {0,2,3}

void init_set_i(set<int> &set, int i) {
	set.clear();

	int k = 0;
	while (i > 0) {
		if (i % 2)
			set.insert(k);
		
		i = i / 2;
		k++;
	}
}

bool test_subsets(const int n, const set<int> *(&op)[MAX][MAX]) {
	for (int t = 1 ; t < n; t++)
		for (set<int>::const_iterator it = op[0][t]->begin(); it != op[0][t]->end(); ++it) {
			if (*it != 0 || *it != t) {
				for (set<int>::const_iterator it2 = op[0][*it]->begin(); it2 != op[0][*it]->end(); ++it2)
					if (op[0][t]->find(*it2) == op[0][t]->end())
						return false;
			}
		}

	return true;

}
#endif

#if 0
void solve(int x, int y, const int n, const set<int> *(&op)[MAX][MAX], int (&complement)[MAX], set<int> *setCandidates, bool transitive) {
	if (x == n - 1) {


/*
				for (int i =0 ; i < n; i++)
					cout << complement[i] << "  ";
				cout << endl;
*/

		if (is_associative(n, op) /*&& has_unique_orthosupplement(n, op)*/) {
			/*int complement[MAX];*/
			/*for (int i = 0; i < n; i++)
				complement[n-1-i] = i;*/

			/*compute_complement(n, op, complement);*/

			if (is_antisymmetric(n, op, complement) && 
				(!transitive || is_transitive(n, op, complement))
			) 
			{
				/* display complement */
				for (int i =0 ; i < n; i++) {
					disp_elem(n, i);
					cout << "  ";
				}
				cout << endl;

				for (int i =0 ; i < n; i++) {
					disp_elem(n, complement[i]);
					cout << "  ";
				}
				cout << endl;

				disp_op(cout, n, op);
				cout << "\t-M-" << endl;
			}
		}

		//disp_op(n, op);
	}
	else {
		if (x == 0) {
			/* {0} cannot be used any more */
			for (int i = 2; i < 1 << n; i++) {
				// 1 \in 0 + 1, 1 cannot be element of 0+x for x\neq 1 
				if ( ( ((y != n - 1) && (i <= (1<<(n-1)))) || ((y == n - 1) && (i>= (1<<(n-1))))) &&



/*(y != n - 1 || i < (1 << (n - 1))) &&*/ setCandidates[i].find(y) != setCandidates[i].end()) {
					bool was_used = false;

					for (int k = y - 1; k>=1;k--) {
						if (op[0][k] == &setCandidates[i]) {
							was_used = true;	
							break;
						}
					}
					
					if (was_used)
						continue;



					op[x][y] = &setCandidates[i];
					op[y][x] = &setCandidates[i];

					if (y < n - 1)
						solve(x, y + 1, n, op, complement, setCandidates, transitive);
					else {
						// test for any $t\in x + 0$, $t + 0\subseteq x + 0$
							if (test_subsets(n, op))
								solve(x + 1, x + 1, n, op, complement, setCandidates, transitive);
					}
					// empty set
					op[x][y] = &setCandidates[0];
					op[y][x] = &setCandidates[0];

				}

			}

		}
		else {
			if (y < n - 1) {
				for (int i = 0; i < 1 << n; i++) {
					op[x][y] = &setCandidates[i];
					if (x != y)
						op[y][x] = &setCandidates[i];

					solve(x, y + 1, n, op, complement, setCandidates, transitive);
					
					op[x][y] = &setCandidates[0];
					if (x!=y)
						op[y][x] = &setCandidates[0];
				}
			}
			else {
				// there is just one element containing 1
				int count = 0;
				for (int i = 0; i < n - 1; i++)
					if (op[x][i]->find(n - 1) != op[x][i]->end()) {
						count++;
						complement[x] = i;
					}

				if (count != 1)
					return;

				// x is summable with all elements in y + 0, if x + y is summable

				for (int i = x; i < n - 1; i++)
					if (is_summable(op, x, i))
						for (set<int>::const_iterator it = op[0][i]->begin(); it != op[0][i]->end(); ++it)
							if (!is_summable(op, x, *it))
								return ;

				solve(x + 1, x + 1, n, op, complement, setCandidates, transitive);
			}

		}
	}
}
#endif


#if 0
void are_isomorphic_via_perm(int level, const hea::Hea &hea1, const hea::Hea &hea2, const set<int> *setCandidates, int (&permutation)[MAX], bool &is_iso) {
	if (!is_iso) {
		if (level == hea1.n - 2) {
			/*
			for (int i = 0; i < hea1.n; i++)
				cout << permutation[i] << " ";
			cout << endl;
			*/

			const set<int> *op_iso[MAX][MAX];
			iso_image(hea1.n, hea1.op, permutation, setCandidates, op_iso);

			if (are_equal_op(hea2.n, hea2.op, op_iso)) {
				is_iso = true;
				/* write information to cerr */

				cerr << "Eliminating ... " << endl;
				disp_hea(cerr, hea2);
				cerr << "\tby" << endl;
				disp_hea(cerr, hea1);
				cerr << "Isomorphism: ";
				for (int i = 0; i < hea1.n; i++) {
					hea::disp_elem(cerr, hea1.n, permutation[i]);
					cerr << " ";
				}
				cerr << endl << "\t---" << endl;
			}	
		}
		else {
			for (int i = level; i < hea1.n - 1; i++) {
				//swap level i
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
				are_isomorphic_via_perm( level + 1, hea1, hea2, setCandidates, permutation, is_iso);
				//swap level i
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
			}
			
		}
	}
}



bool are_isomorphic(const hea::Hea &hea1, const hea::Hea &hea2, const set<int> *setCandidates) {
	if (hea1.n != hea2.n)
		return false;

	int permutation[MAX];
	for (int i = 0; i < hea1.n;i++)
		permutation[i] = i;

	bool is_iso = false;

	are_isomorphic_via_perm(1, hea1, hea2, setCandidates, permutation, is_iso);

	return is_iso;
}
#endif

#if 0
void disp_hea(const hea::Hea &hea) {
	/* display complement */
	for (int i =0 ; i < hea.n; i++) {
		disp_elem(hea.n, i);
		cout << "  ";
	}
	cout << endl;

	for (int i =0 ; i < hea.n; i++) {
		disp_elem(hea.n, hea.complement[i]);
		cout << "  ";
	}
	cout << endl;

	hea::disp_op(cout, hea.n, hea.op);
	cout << "\t-M-" << endl;
}
#endif

/* for n = 5 only BEGIN */

/* defined a + a' , b + b', c + c'  only (apart from x + 0) */
bool isClass1(const int n, const set<int> * const (&op)[MAX][MAX], const int (&complement)[MAX]) {
	for (int i = 1; i < n - 1; i++)
	for (int j = i + 1; j < n - 1; j++)
		if (hea::is_summable(op, i, complement[j]) || hea::is_summable(op, j, complement[i]))
			return false;

	return true;	
}

/* defined a + a' , b + b', c + c', b + c'  only (apart from x + 0) */
bool isClass2(const int n, const set<int> * const (&op)[MAX][MAX], const int (&complement)[MAX]) {
	int count = 0;
	for (int i = 1; i < n - 1; i++)
	for (int j = i + 1; j < n - 1; j++)
		if (hea::is_summable(op, i, complement[j]) || hea::is_summable(op, j, complement[i])) {
			count++;
		}

	if (count == 1)
		return true;
	else
		return false;
}

/* defined a + a' , b + b', c + c', a + c', b + c'  only (apart from x + 0) */
bool isClass3(const int n, const set<int> * const (&op)[MAX][MAX], const int (&complement)[MAX]) {
	bool atleastone = false;

	for (int i = 1; i < n - 1; i++)
	for (int j = 1; j < n - 1; j++)
	for (int k = 1; k < n - 1; k++) 
		if (i != j && j != k && i != k) {
			if (	hea::is_summable(op, i, complement[k]) &&
				hea::is_summable(op, j, complement[k]) ) {
					atleastone = true;

					if (hea::is_summable(op, i, complement[j]) || hea::is_summable(op, j, complement[i])) 
						return false;
					
				

			}
	

		}
	
	return atleastone;
}

/* defined a + a' , b + b', c + c', a + b', b + c' and a || c  only (apart from x + 0) */
bool isClass4(const int n, const set<int> * const (&op)[MAX][MAX], const int (&complement)[MAX]) {
	bool atleastone = false;

	for (int i = 1; i < n - 1; i++)
	for (int j = 1; j < n - 1; j++)
	for (int k = 1; k < n - 1; k++) 
		if (i != j && j != k && i != k) {
			if (	hea::is_summable(op, i, complement[j]) &&
				hea::is_summable(op, j, complement[k]) ) {
					atleastone = true;

					if (hea::is_summable(op, i, complement[k]) || hea::is_summable(op, k, complement[i])) 
						return false;
					
				

			}
	

		}
	
	return atleastone;
}

/* defined a + a' , b + b', c + c', a + b', b + c' and a + c'  only (apart from x + 0) */
bool isClass6(const int n, const set<int> * const (&op)[MAX][MAX], const int (&complement)[MAX]) {

	for (int i = 1; i < n - 1; i++)
	for (int j = 1; j < n - 1; j++)
	for (int k = 1; k < n - 1; k++) 
		if (i != j && j != k && i != k) {
			if (	hea::is_summable(op, i, complement[j]) &&
				hea::is_summable(op, j, complement[k]) &&
				hea::is_summable(op, i, complement[k]) ) {
					return true;
			}
	

		}
	
	return false;
}

/* defined a + a' , b + b', c + c', a + b', b + c' and a + c'  only (apart from x + 0) */
bool isClass7(const int n, const set<int> * const (&op)[MAX][MAX], const int (&complement)[MAX]) {

	for (int i = 1; i < n - 1; i++)
	for (int j = 1; j < n - 1; j++)
	for (int k = 1; k < n - 1; k++) 
		if (i != j && j != k && i != k) {
			if (	hea::is_summable(op, i, complement[j]) &&
				hea::is_summable(op, j, complement[k]) &&
				hea::is_summable(op, k, complement[i]) ) {
					return true;
			}
	

		}
	
	return false;
}


/* for n = 5 only END   */



int main(int argc, char **argv) {


	if (argc < 2) {
		cout << "Usage: " << argv[0] << " filename [filename2 ...]" << endl;
		return 0;
	
	}

	//vector<hea::Hea> non_iso_list;
	//bool is_first = true;

	const int n = 5;

	set<int> setCandidates[1 << MAX];
	const set<int>* op[MAX][MAX];

	// init
	for (int i = 0; i < 1 << n; i++) {
		setCandidates[i] = set<int>();
		hea::init_set_i(setCandidates[i], i);
	}
	
	int complement[MAX];
	int idx = 1;
	while (idx < argc) {
		ifstream ifs;

		ifs.open(argv[idx], ifstream::in);
		
		if (ifs.good()) {
			cerr << "Loading file " << argv[idx] << "." << endl;

			while(hea::parse_model(ifs, n, op, complement, setCandidates)) {

				hea::Hea tested_hea(n, op, complement);

				if (isClass7(n, op, complement))
					hea::disp_hea(cout, tested_hea);

				// skip one line containing -M-
				char buffer[1024];
				ifs.getline(buffer, 1024);
			}
				
		}
		else {
			cerr << "Cannot open or read from file '" << argv[idx] << "'." << endl;
		}
		ifs.close();

		idx++;
	}

	return 0;
}

