#include "hea.h"

#include "hyperEA3_common.h"

namespace hea {

Hea::Hea(const int _n, const std::set<int> * const _op[MAX][MAX], const int (&_complement)[MAX]) {
	n = _n;
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		op[i][j] = _op[i][j];	

	for (int i = 0; i < n; i++)
		complement[i] = _complement[i];
	
}

void disp_hea(std::ostream &os, const Hea &hea) {
	/* display complement */
	for (int i =0 ; i < hea.n; i++) {
		disp_elem(os, hea.n, i);
		os << "  ";
	}
	os << std::endl;

	for (int i =0 ; i < hea.n; i++) {
		disp_elem(os, hea.n, hea.complement[i]);
		os << "  ";
	}
	os << std::endl;

	disp_op(os, hea.n, hea.op);
	os << "\t-M-" << std::endl;
}

void are_isomorphic_via_perm(int level, const Hea &hea1, const Hea &hea2, const std::set<int> *setCandidates, int (&permutation)[MAX], bool &is_iso) {
	if (!is_iso) {
		if (level == hea1.n - 2) {
			/*
			for (int i = 0; i < hea1.n; i++)
				cout << permutation[i] << " ";
			cout << endl;
			*/

			const std::set<int> *op_iso[MAX][MAX];
			iso_image(hea1.n, hea1.op, permutation, setCandidates, op_iso);

			if (are_equal_op(hea2.n, hea2.op, op_iso)) {
				is_iso = true;	

				/* write information to cerr */

				std::cerr << "Eliminating ... " << std::endl;
				disp_hea(std::cerr, hea2);
				std::cerr << "\tby" << std::endl;
				disp_hea(std::cerr, hea1);
				std::cerr << "Isomorphism: ";
				for (int i = 0; i < hea1.n; i++) {
					disp_elem(std::cerr, hea1.n, permutation[i]);
					std::cerr << " ";
				}
				std::cerr << std::endl << "\t---" << std::endl;
			}
		}
		else {
			for (int i = level; i < hea1.n - 1; i++) {
				//swap level i
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
				are_isomorphic_via_perm( level + 1, hea1, hea2, setCandidates, permutation, is_iso);
				//swap level i
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
			}
			
		}
	}
}

bool are_isomorphic(const Hea &hea1, const Hea &hea2, const std::set<int> *setCandidates) {
	if (hea1.n != hea2.n)
		return false;

	int permutation[MAX];
	for (int i = 0; i < hea1.n;i++)
		permutation[i] = i;

	bool is_iso = false;

	are_isomorphic_via_perm(1, hea1, hea2, setCandidates, permutation, is_iso);

	return is_iso;
}

void generate_all_order_preserving_permutations_recursive(int level, const Hea &hea, int (&permutation)[MAX], const bool include_identity_permutation, std::vector<Permutation> &vec_order_preserve) {

	if (level == hea.n - 2) {
		
		bool is_identity = true;
		if (include_identity_permutation) {
			// include all permutations
			is_identity = false;
		}
		else {
			for (int i = 1 ; i < hea.n - 1; i++)
				if (i != permutation[i]) {
					is_identity = false;
					break;
				}
					

		}

		if (!is_identity) {
#if 0
/* TODO: check */
			for (int i = 1; i < hea.n - 1; i++)
			for (int j = 1; j < hea.n - 1; j++) {
				/* does it preserve ``order'' relation */
				if (is_summable(hea.op, i, hea.complement[j]) && !is_summable(hea.op, permutation[i], hea.complement[permutation[j]]))
					return ;
				if (is_summable(hea.op, permutation[i], hea.complement[permutation[j]]) && !is_summable(hea.op, i, hea.complement[j]))
					return ;
			}
#endif
			vec_order_preserve.push_back(Permutation(hea.n, permutation));

		}

	}
	else {
		for (int i = level; i < hea.n - 1; i++) {
			//swap level i
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
			generate_all_order_preserving_permutations_recursive( level + 1, hea, permutation, include_identity_permutation, vec_order_preserve);
			//swap level i
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
		}
		
	}

}

void generate_all_order_preserving_permutations(const Hea &hea, const bool include_identity_permutation, std::vector<Permutation> &vec_order_preserve) {
	int permutation[MAX];
	for (int i = 0; i < hea.n; i++)
		permutation[i] = i;

	generate_all_order_preserving_permutations_recursive(1, hea, permutation, include_identity_permutation, vec_order_preserve);
}


void select_plan_preserving_permutations_old(const Hea &hea, std::vector<Permutation> &vec_order_preserve) {
	std::vector<Permutation> tmp_vec_order_preserve(vec_order_preserve);
	std::vector<Permutation>::iterator itvec = vec_order_preserve.begin();

	for (std::vector<Permutation>::iterator tmp_itvec = tmp_vec_order_preserve.begin(); tmp_itvec != tmp_vec_order_preserve.end(); ++tmp_itvec) {
#ifdef DEBUG
			std::cerr << "Considering: ";
			for (int i = 0; i < hea.n; i++)
				std::cerr << itvec->permutation[i] << "  ";

			std::cerr << std::endl;
#endif

		bool preserves_plan = true;

		for (int i = 1; preserves_plan && i < hea.n - 1; i++)
		for (int j = 1; j < hea.n - 1; j++)
			if (is_summable(hea.op, i, j) && !is_summable(hea.op, itvec->permutation[i], itvec->permutation[j])) {
				preserves_plan = false;
				break;
			}

		if (!preserves_plan) {
#ifdef DEBUG
			std::cerr << "Deleting: ";
			for (int i = 0; i < hea.n; i++)
				std::cerr << itvec->permutation[i] << "  ";

			std::cerr << std::endl;
#endif
			vec_order_preserve.erase(itvec);
		}
		++itvec;
			
	}
}

void select_plan_preserving_permutations(const Hea &hea, const std::vector<Permutation> &vec_order_preserve, std::vector<Permutation> &out_vector) {
	out_vector.clear();
	
	for (std::vector<Permutation>::const_iterator itvec = vec_order_preserve.begin(); itvec != vec_order_preserve.end(); ++itvec) {
#ifdef DEBUG
		std::cerr << "Considering: ";
		for (int i = 0; i < hea.n; i++)
			std::cerr << itvec->permutation[i] << "  ";

		std::cerr << std::endl;
#endif

		bool preserves_plan = true;
		int perm_image[MAX];

		/* compute complement image */
		for (int i = 1; i < hea.n - 1; i++) {
			perm_image[itvec->permutation[i]] = itvec->permutation[hea.complement[i]];
		}
		
		/* compare complement and complement image */
		for (int i = 1; i < hea.n - 1; i++) 
			if (perm_image[i] != hea.complement[i]) {
				preserves_plan = false;
				break;
			}

		if (!preserves_plan)
			continue;

		/* i + j exists iff perm^{-1}(i) + perm^{-1}(j) exists */

		/* compute inverse perm */
		int inv_permutation[MAX];
		inv_permutation[0] = 0; inv_permutation[hea.n - 1] = hea.n - 1;
		for (int i = 1; i < hea.n - 1; i++)
			inv_permutation[itvec->permutation[i]] = i;

		/* check i + j exists iff perm^{-1}(i) + perm^{-1}(j) exists */
		for (int i = 1; preserves_plan && (i < hea.n - 1); i++)
		for (int j = 1; j < hea.n - 1; j++) {
			if (is_summable(hea.op, i, j) && !is_summable(hea.op, inv_permutation[i], inv_permutation[j])) {
				preserves_plan = false;
				break;
			}	
			if (is_summable(hea.op, inv_permutation[i], inv_permutation[j]) && !is_summable(hea.op, i, j)) {
				preserves_plan = false;
				break;
			}	
		}

		if (preserves_plan) {		
			/* store into output vector - out_vector */
			out_vector.push_back(*itvec);
		}		
	}
}


bool are_isomorphic_via_perm(const hea::Hea &hea1, const hea::Hea &hea2, const std::set<int> *setCandidates, const int (&permutation)[MAX]) {

	const std::set<int> *op_iso[MAX][MAX];
	hea::iso_image(hea1.n, hea1.op, permutation, setCandidates, op_iso);

	if (hea::are_equal_op(hea2.n, hea2.op, op_iso)) 
		return true;

	return false;

}

bool are_isomorphic_perm_list(const Hea &hea1, const Hea &hea2, const std::set<int> *setCandidates, const std::vector<Permutation> &vec_perm) {
	for (std::vector<Permutation>::const_iterator it_vec = vec_perm.begin(); it_vec != vec_perm.end(); ++it_vec)
		if (are_isomorphic_via_perm(hea1, hea2, setCandidates, it_vec->permutation)) {
			/* write information to cerr */

			std::cerr << "Eliminating ... " << std::endl;
			disp_hea(std::cerr, hea2);
			std::cerr << "\tby" << std::endl;
			disp_hea(std::cerr, hea1);
			std::cerr << "Isomorphism: ";
			for (int i = 0; i < hea1.n; i++) {
				disp_elem(std::cerr, hea1.n, it_vec->permutation[i]);
				std::cerr << " ";
			}
			std::cerr << std::endl << "\t---" << std::endl;

			return true;
		}

	return false;
}
} // end namespace hea
