#include "hyperEA3_common.h"

namespace hea {

bool is_summable(const std::set<int>* const (&op)[MAX][MAX], const int a, const int b) {
//	cout << "is_summable: " << a << ", " << b << endl;
	return !((op[a][b])->empty());
}

bool is_summable(const std::set<int>* const (&op)[MAX][MAX], const int a, const std::set<int>&set_b) {
	for (std::set<int>::iterator it = set_b.begin(); it != set_b.end(); ++it)
		if (!is_summable(op, a, *it))
			return false;

	return true;
}

bool is_summable(const std::set<int>* const (&op)[MAX][MAX], const std::set<int> &set_a, const std::set<int>&set_b) {
	for (std::set<int>::iterator it1 = set_a.begin(); it1 != set_a.end(); ++it1)
	for (std::set<int>::iterator it2 = set_b.begin(); it2 != set_b.end(); ++it2)
		if (!is_summable(op, *it1, *it2))
			return false;

	return true;
}

void disp_elem(std::ostream &os, const int n, int elem) {
	if (elem == 0) {
		os << 0;
		return;
	}
	if (elem == n - 1) {
		os << 1;
		return;
	}

	if (elem >=0 && elem < n) {
		os << (char) ('a' + elem - 1);
		return;
	}
}

void disp_set(std::ostream &os, const int n, const std::set<int> &ds) {
/*
	for (set<int>::iterator it = ds.begin(); it != ds.end(); ++it) {
		cout << *it << ", ";
	}
//	cout << endl;
*/

	std::set<int>::iterator it = ds.begin();
	if (it != ds.end()) {
		//cout << *it;
		disp_elem(os, n, *it);
		++it;
		do {
			if (it != ds.end()) {
				os << ", "; 
				disp_elem(os, n, *it);
				++it;
			}
		} while (it != ds.end());
	}

}


void disp_complement(std::ostream &os, const int n, const int (&complement)[MAX]) {
	bool was_displayed = false;
	for (int i = 0; i < n; i++) {
		if (was_displayed)
			os << "  ";
		disp_elem(os, n, i);
		was_displayed = true;
	}
	os << std::endl;

	was_displayed = false;
	for (int i = 0; i < n; i++) {
		if (was_displayed)
			os << "  ";
		disp_elem(os, n, complement[i]);
		was_displayed = true;
	}
	os << std::endl;
}

void disp_op(std::ostream &os, const int n, const std::set<int>* const (&op)[MAX][MAX]) {
	for (int i = 0; i < n; i++) {
		bool was_displayed = false;
		for (int j = 0; j < n; j++) {
			if (was_displayed)
				os << ", ";
			os << "{";
			disp_set(os, n, *op[i][j]);
			os << "}";
			was_displayed = true;
		}
		os << std::endl;
			
	}
}

std::set<int> compute_sum(const std::set<int>* const (&op)[MAX][MAX], const int a, const int b) {
	std::set<int> result = std::set<int>();

	if (!is_summable(op, a, b))
		return result;

	else
		return *(op[a][b]);
	
}

std::set<int> compute_sum(const std::set<int>* const (&op)[MAX][MAX], const int a, const std::set<int> &set_b) {
	std::set<int> result = std::set<int>();

	if (!is_summable(op, a, set_b))
		return result;


	for (std::set<int>::const_iterator it = set_b.begin(); it != set_b.end(); ++it)

		result.insert(op[a][*it]->begin(), op[a][*it]->end());
	return result;

	
}




std::set<int> compute_sum(const std::set<int>* const (&op)[MAX][MAX], const std::set<int> &set_a, const std::set<int> &set_b) {
	std::set<int> result = std::set<int>();

	if (!is_summable(op, set_a, set_b))
		return result;

	else {
		for (std::set<int>::const_iterator it1 = set_a.begin(); it1 != set_a.end(); ++it1)
		for (std::set<int>::const_iterator it2 = set_b.begin(); it2 != set_b.end(); ++it2)

			result.insert(op[*it1][*it2]->begin(), op[*it1][*it2]->end());
		return result;
	}
	
}

int is_associative(const int n, const std::set<int> * const (&op)[MAX][MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
	for (int k = 0; k < n; k++) {
//		cout << i << ", " << j << ", " << k << endl;

		
		// exists i+j && exists (i+j)+k => exists j+k && exists i+(j+k)
		// and (i+j)+k = i + (j+k)

		if (is_summable(op, i, j)) {
			std::set<int> tmp (compute_sum(op, i, j));
			if (is_summable(op, k, tmp)) {
				std::set<int> final1(compute_sum(op, k, tmp));

				if (!is_summable(op, j, k))
					return false;

				std::set<int> tmp2(compute_sum(op, j, k));

				if (!is_summable(op, i, tmp2))
					return false;

				std::set<int> final2(compute_sum(op, i, tmp2));

				if (final1 != final2)
					return false;
			}
		}


#

		// exists i+j && exists (i+j)+k <= exists j+k && exists i+(j+k)
		// and (i+j)+k = i + (j+k)
		if (is_summable(op, j, k)) {
			std::set<int> tmp(compute_sum(op, j, k));
			//cout << j << " + " << k << " = tmp: "; disp_set(tmp); cout << endl;
			if (is_summable(op, i, tmp)) {
				std::set<int> final1(compute_sum(op, i, tmp));
				//cout << i << " + " << "(" << j << " + " << k << " = final1:" ; disp_set(final1); cout << endl;

 

				if (!is_summable(op, i, j)) {
					//cout << "not i + j" << endl;
					return false;
				}

				std::set<int> tmp2(compute_sum(op, i, j));

				if (!is_summable(op, k, tmp2)) {
					//cout << "not (i+j)+k" << endl;
					return false;
				}

				//cout << "tmp2: "; disp_set(tmp2); cout << endl;
				std::set<int> final2(compute_sum(op, k, tmp2));
				//cout << "final2: " ;disp_set(final2); cout << endl;


				if (final1 != final2) {
					//cout << "neq" << endl;
					return false;
				}
			}
		}


	}
//	cout << "HERE" << endl;
	return true;
}

bool is_antisymmetric(const int n, const std::set<int>* const (&op)[MAX][MAX], const int (&complement)[MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if  (is_summable(op, i, complement[j]) && is_summable(op, j, complement[i]) && i!= j)
			return false;

	return true;
}

bool is_transitive(const int n, const std::set<int>* const (&op)[MAX][MAX], const int (&complement)[MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
	for (int k = 0; k < n; k++)
		if  (is_summable(op, i, complement[j]) && is_summable(op, j, complement[k]) && !is_summable(op, i, complement[k]))
			return false;

	return true;
}

// setCandidate[candidate] is the set 0 + y
// for any t in 0 + y, t !<! y (order in whole numbers) we known that 0 + t is already defined
// moreover it is checked that 0 + t \subseteq 0 + y
// ---
// new checks
// for any 0 + t, t !<! y (order in whole numbers) if y\in 0 + t, then 0 + y (setCandidate[candidate) must be \subseteq 0 + t 
// ---
bool test_subsets_new(const int n, int candidate, const std::set<int> * const (&op)[MAX][MAX], int y, const std::set<int> *setCandidates) {
#if 0
	for (int t = y - 1 ; t >= 0; t--) 
		if (candidate & (1<<t)) {
			/* candidate contains element t */
			for (set<int>::const_iterator it = op[0][t]->begin(); it != op[0][t]->end(); ++it) {
				/* does it contain also any element from [0][t] */
				if (!(candidate & (1 << *it))) 
					return false;

			}
		}

#endif
	for (std::set<int>::const_iterator it = setCandidates[candidate].begin(); it != setCandidates[candidate].end(); ++it) {
		if (*it < y) {
			for (std::set<int>::const_iterator it2 = op[0][ *it ]->begin(); it2 != op[0][ *it ]->end(); ++it2) {
				/* does it contain also any element from [0][*it] */
				if (setCandidates[candidate].find( *it2 ) == setCandidates[candidate].end())
					return false;
	
			}
		}
	}

	// 0 + 0 = \{0\}, thus it is not necessary to check it
	for (int i = y - 1 ; i > 0 ; i--) {
		if (op[0][i]->find(y) != op[0][i]->end()) {

			for (std::set<int>::const_iterator it = setCandidates[candidate].begin(); it != setCandidates[candidate].end();++it)
				if (op[0][i]->find(*it) == op[0][i]->end())
					return false;
		}
	}


	return true;

}

// initialize set with elements equal to binary representation of number i
// it is expected to call this function for i from 0 to (1 << n) - 1.
//
// n = 4
// 3210
// 0100 -> {2}
// 3210
// 1101 -> {0,2,3}

void init_set_i(std::set<int> &set, int i) {
	set.clear();

	int k = 0;
	while (i > 0) {
		if (i % 2)
			set.insert(k);
		
		i = i / 2;
		k++;
	}
}

void iso_image(const int n, const std::set<int> * const (&op)[MAX][MAX], const int (&perm)[MAX], const std::set<int> *setCandidates, const std::set<int>* (&op2)[MAX][MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++) {
		if (hea::is_summable(op, i, j)) {
			int value = 0;
			for (std::set<int>::const_iterator it = op[i][j]->begin(); it != op[i][j]->end(); ++it) 
				value |= (1 << perm[ *it ]);
			
			op2[ perm[i] ][ perm[j] ] = &setCandidates[value];
		}
		else {
			op2[ perm[i] ][ perm[j] ] = &setCandidates[0]; // empty set
		}
	}
}

void iso_image_full(const int n, const std::set<int> * const (&op)[MAX][MAX], const int (&complement)[MAX], const int (&perm)[MAX], const std::set<int> *setCandidates, const std::set<int>* (&op2)[MAX][MAX], int (&complement2)[MAX]) {
	// op image
	iso_image(n, op, perm, setCandidates, op2);
	
	for (int i = 0; i < n; i++)
		complement2[ perm[i] ] = perm[ complement[i] ];


}

bool are_equal_op(const int n, const std::set<int>* const (&op)[MAX][MAX], const std::set<int>* const (&op2)[MAX][MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if (op[i][j] != op2[i][j])
			return false;

	return true;
}


void dump_state(std::ostream &os, int n, const std::set<int>* const (&op)[MAX][MAX], const std::set<int> *setCandidates) {
	// dumps  state
	for (int i = 0; i < n; i++)
		for (int j = i; j < n; j++)
			if (op[i][j] != &setCandidates[0])
				os << (op[i][j] - &setCandidates[0]) << ", ";
	os << std::endl;
}

void dump_state_to_vector(int n, const std::set<int>* const (&op)[MAX][MAX], const std::set<int> *setCandidates, std::vector<int> &state_vector) {
	// dumps  state
	for (int i = 0; i < n; i++)
		for (int j = i; j < n; j++)
			if (op[i][j] != &setCandidates[0])
				state_vector.push_back(op[i][j] - &setCandidates[0]);
}

/* -1 = <  */
/*  0 = == */
/*  1 = >  */

int compare_states(const std::vector<int> &state1, const std::vector<int> &state2) {
	/* assert: state1 and state2 corresponds to the same order */

	std::vector<int>::const_iterator it1 = state1.begin();	
	std::vector<int>::const_iterator it2 = state2.begin();	

	// skip same elements in the vectors
	while (it1 != state1.end() && it2 != state2.end() && *it1 == *it2) {++it1; ++it2;}

	if (it1 == state1.end()) // all elements are the same
	{
		return 0;
	}

	if (*it1 < *it2)
		return -1;
	else
		return 1;

}

/* For the case where 0 + x subseteq \{0, x\}, checks wheater candidate for x + x meets desired conditions */
bool testCandidateSetForXPlusXContainingOne(int i, int x, const int n, const std::set<int> *(&op)[MAX][MAX], std::set<int> *setCandidates) {
	/* x + x */
	/* x + 0 = \{x\} 
		x + x = A
			A + 0 = A
			1 + 0 \subseteq A + 0

	*/
	if (op[0][x] == &setCandidates[1<<x]) {
		std::set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
	
		if (resA0 != setCandidates[i]) 
			return false;	
		
		
		int idx01 = op[0][n - 1] - &setCandidates[0];
		if ((idx01 & i) != idx01) 
			return false;	
	}
	else
	/* x + 0 = \{0, x\} 
		x + x = A
			A + 0 = A \union \{0, x\}
			1 + 0 \subseteq A + 0
	*/
	if (op[0][x] == &setCandidates[(1<<x) + 1]) {
		std::set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		std::set<int> resA0x = setCandidates[i];
		resA0x.insert(0);
		resA0x.insert(x);
		
		if (resA0 != resA0x) 
			return false;						

		int idx01 = op[0][n - 1] - &setCandidates[0];
		int idxA0 = 0;
		for (std::set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
			idxA0 += (1 << (*itA0));

/*
		int idxA0 = i | 1;
		idxA0 |= (1 << x); 
*/
		if ((idx01 & idxA0) != idx01) 
			return false;
	}
			
	
	return true;

}

/* For the case where 0 + x subseteq \{0, x\} or 0 + y subseteq \{0, y\} checks wheater candidate for x + x meets desired conditions */
bool testCandidateSetForXPlusYJustOneContainingOne(int i, int x, bool only0x, int y, const int n, const std::set<int> *(&op)[MAX][MAX], std::set<int> *setCandidates) {

	int a, b;

	if (only0x) {
		a = x;
		b = y;
	}
	else {
		a = y;
		b = x;
	}

	if (op[0][a] == &setCandidates[1<<a]) {
		std::set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		if (resA0 != setCandidates[i]) 
			return false;
		
		int idx01 = op[0][n - 1] - &setCandidates[0];
		if ((idx01 & i) != idx01) 
			return false;
	}
	else
	if (op[0][a] == &setCandidates[1+(1<<a)]) {
		std::set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		std::set<int> resA_0b = setCandidates[i];
		for (std::set<int>::const_iterator it_0b=op[0][b]->begin(); it_0b != op[0][b]->end(); ++it_0b)
			resA_0b.insert(*it_0b);
		
		if (resA0 != resA_0b) 
			return false;

		int idx01 = op[0][n - 1] - &setCandidates[0];
		int idxA0 = 0;
		for (std::set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
			idxA0 += (1 << (*itA0));
		if ((idx01 & idxA0) != idx01) 
			return false;
	}

	return true;

}

/* For the case where 0 + x subseteq \{0, x\} and 0 + y subseteq \{0, y\} checks wheater candidate for x + x meets desired conditions */
bool testCandidateSetForXPlusYBothContainingOne(int i, int x, int y, const int n, const std::set<int> *(&op)[MAX][MAX], std::set<int> *setCandidates) {

	/* x + y */
	if (op[0][x] == &setCandidates[1<<x] && op[0][y] == &setCandidates[1<<y]) {
		/* x + 0 = \{x\} && y + 0 = \{y\} => (x + y) + 0 = (x + y) */
		std::set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		if (resA0 != setCandidates[i]) 
			return false;
		
		
		int idx01 = op[0][n - 1] - &setCandidates[0];
		if ((idx01 & i) != idx01) 		
			return false;
	}
	else 
	if (op[0][x] == &setCandidates[1<<x] && op[0][y] == &setCandidates[1 + (1<<y)]) {
		/* x + 0 = \{x\} && y + 0 = \{0, y\} => (x + y) + 0 = x + y = (x + y) union \{x\} */
		std::set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		if (resA0 != setCandidates[i]) 
			return false;

		std::set<int> resAx = setCandidates[i]; /* (x + y) union \{x\} */
		resAx.insert(x);
		if (resA0 != resAx) 
			return false;

		int idx01 = op[0][n - 1] - &setCandidates[0];
		if ((idx01 & i) != idx01) 
			return false;
	}
	else 
	if (op[0][x] == &setCandidates[1 + (1<<x)] && op[0][y] == &setCandidates[1<<y]) {
		/* x + 0 = \{0, x\} && y + 0 = \{y\} => (x + y) + 0 = x + y =  (x + y) union \{y\} */
		std::set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		if (resA0 != setCandidates[i]) 
			return false;

		std::set<int> resAy = setCandidates[i]; /* (x + y) union \{y\} */
		resAy.insert(y);
		if (resA0 != resAy) 
			return false;

		int idx01 = op[0][n - 1] - &setCandidates[0];
		if ((idx01 & i) != idx01) 
			return false;
	}
	else 
	if (op[0][x] == &setCandidates[1 + (1<<x)] && op[0][y] == &setCandidates[1 + (1<<y)]) {
		/* x + 0 = \{0, x\} && y + 0 = \{0, y\} => (x + y) + 0 = (x + y) union \{0, x\} = (x + y) union \{0, y\} */
		std::set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		std::set<int> resA0x = setCandidates[i]; /* (x + y) union \{0, x\} */
		resA0x.insert(0);
		resA0x.insert(x);

		if (resA0 != resA0x) 
			return false;

		std::set<int> resA0y = setCandidates[i]; /* (x + y) union \{0, y\} */
		resA0y.insert(0);
		resA0y.insert(y);

		if (resA0 != resA0y) 
			return false;

		int idx01 = op[0][n - 1] - &setCandidates[0];
		int idxA0 = i | 1;
		idxA0 |= (1<<x); // union 0, x should be same as union 0, y

		if ((idx01 & idxA0) != idx01) 
			return false;
	}

	return true;

}

}
