#include <iomanip>
#include <string.h>
#include <stdlib.h>
#include <fstream>
#include <vector>

#include "hyperEA3_common.h"
#include "hea.h"
#include "hyperEA3_parse.h"
	

using namespace std;



struct Permutation {
	int n;
	int permutation[MAX];

	Permutation(const int _n, const int (&_permutation)[MAX]);
};

Permutation::Permutation(const int _n, const int (&_permutation)[MAX]) {
	n = _n;
	for (int i = 0; i < n; i++)
		permutation[i] = _permutation[i];	
}




void disp_set(const int n, const set<int> &ds) {
/*
	for (set<int>::iterator it = ds.begin(); it != ds.end(); ++it) {
		cout << *it << ", ";
	}
//	cout << endl;
*/

	set<int>::iterator it = ds.begin();
	if (it != ds.end()) {
		//cout << *it;
		hea::disp_elem(cout, n, *it);
		++it;
		do {
			if (it != ds.end()) {
				cout << ", "; 
				hea::disp_elem(cout, n, *it);
				++it;
			}
		} while (it != ds.end());
	}

}

void generate_all_order_preserving_permutations(int level, const hea::Hea &hea, int (&permutation)[MAX], vector<Permutation> &vec_order_preserve) {

	if (level == hea.n - 2) {
		
		for (int i = 1; i < hea.n - 1; i++)
		for (int j = 1; j < hea.n - 1; j++) {
			if (hea::is_summable(hea.op, i, hea.complement[j]) && !hea::is_summable(hea.op, permutation[i], hea.complement[permutation[j]]))
				return ;
			if (hea::is_summable(hea.op, permutation[i], hea.complement[permutation[j]]) && !hea::is_summable(hea.op, i, hea.complement[j]))
				return ;
		}

		vec_order_preserve.push_back(Permutation(hea.n, permutation));


	}
	else {
		for (int i = level; i < hea.n - 1; i++) {
			//swap level i
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
			generate_all_order_preserving_permutations( level + 1, hea, permutation, vec_order_preserve);
			//swap level i
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
		}
		
	}

}

int main(int argc, char **argv) {

	if (argc < 2) {
		cout << "Usage: " << argv[0] << " filename [filename2 ...]" << endl;
		return 0;
	
	}

	vector<hea::Hea> non_iso_list;
	bool is_first = true;

	const int n = 6;

	set<int> setCandidates[1 << MAX];
	const set<int>* op[MAX][MAX];

	// init
	for (int i = 0; i < 1 << n; i++) {
		setCandidates[i] = set<int>();
		hea::init_set_i(setCandidates[i], i);
	}
	
	int complement[MAX];
	int idx = 1;
	while (idx < argc) {
		ifstream ifs;

		ifs.open(argv[idx], ifstream::in);
		
		if (ifs.good()) {
			cerr << "Loading file " << argv[idx] << "." << endl;

	/*
			// skip header and copy header
			
			int to_skip = 2;
			char buffer[1024];
			while (to_skip > 0) {
				ifs.getline(buffer, 1024);
				if (!strncmp(buffer, "-----", 5)) {
					//cout << "Skipped..." <<endl;
					to_skip--;
				}
				cout << buffer<<endl;
			}

	*/

			while(hea::parse_model(ifs, n, op, complement, setCandidates)) {

				/*
				cout << "Testing ..." << endl;
				for (int i = 0; i < n; i++) {
					disp_elem(n, complement[i]);
					cout << " ";
				}
				cout << endl;
				disp_op(n, op);
				*/

				hea::Hea tested_hea(n, op, complement);

				if (is_first) {
					is_first = false;
					hea::disp_hea(cout, tested_hea);
					non_iso_list.push_back(tested_hea);
				}
				else {

					bool is_iso = false;
					for (vector<hea::Hea>::const_iterator it_vec = non_iso_list.begin(); it_vec != non_iso_list.end(); ++it_vec) {
						if (are_isomorphic(*it_vec, tested_hea, setCandidates)) {
							is_iso = true;
							break;
						}
					}

					if (!is_iso) {
						hea::disp_hea(cout, tested_hea);
						non_iso_list.push_back(tested_hea);
					}
				}

				// skip one line containing -M-
				char buffer[1024];
				ifs.getline(buffer, 1024);
			}
				
		}
		else {
			cerr << "Cannot open or read from file '" << argv[idx] << "'." << endl;
		}
		ifs.close();

		idx++;
	}

	return 0;
}
