#include <iomanip>
#include <string.h>
#include <stdlib.h>
#include <fstream>
#include <vector>

#include "hyperEA3_common.h"
#include "hea.h"
#include "hyperEA3_parse.h"
	

using namespace std;

void disp_set(const int n, const set<int> &ds) {
/*
	for (set<int>::iterator it = ds.begin(); it != ds.end(); ++it) {
		cout << *it << ", ";
	}
//	cout << endl;
*/

	set<int>::iterator it = ds.begin();
	if (it != ds.end()) {
		//cout << *it;
		hea::disp_elem(cout, n, *it);
		++it;
		do {
			if (it != ds.end()) {
				cout << ", "; 
				hea::disp_elem(cout, n, *it);
				++it;
			}
		} while (it != ds.end());
	}

}

bool has_unique_orthosupplement_elem(const int n, const set<int>* (&op)[MAX][MAX], int a) {
	bool found = false;

	for (int i = 0; i < n ; i++)

		if (op[a][i]->find(n - 1) != op[a][i]->end()) {
			if (!found)
				found = true;
			else
				return false; /* not unique element */
			
		}

	return found;
}

bool has_unique_orthosupplement(const int n, const set<int>* (&op)[MAX][MAX]) {
	for (int i = 0; i < n; i++)
		if (!has_unique_orthosupplement_elem(n, op, i))
			return false;

	return true;
}


void disp_op(const int n, const set<int>* const (&op)[MAX][MAX]) {
	for (int i = 0; i < n; i++) {

		for (int j = 0; j < n; j++) {
			cout << "{";
			disp_set(n, *op[i][j]);
			cout << "}, ";
		}
		cout << endl;
			
	}
	
}

void compute_complement(const int n, const set<int>* (&op)[MAX][MAX], int (&complement)[MAX]) {
/*
	cout << "computing complement for: " << endl;
	disp_op(n, op);
	cout << "end complement" << endl;
*/

//cout << "HERE" << endl;
	complement[0] = n - 1;
	complement[n - 1] = 0;

	for (int i = 1; i < n - 1; i++) {
		//cout << "c: " << i;
		for (int j = n - 2; j>=1; j--) {
		//	cout << j << " ";
			//disp_set(*op[i][j]);
			if (op[i][j]->find(n-1) != op[i][j]->end()) {
				complement[i] = j;
				break;
			}
		}
		//cout << endl;

			
		
	}
}


bool test_subsets(const int n, const set<int> *(&op)[MAX][MAX]) {
	for (int t = 1 ; t < n; t++)
		for (set<int>::const_iterator it = op[0][t]->begin(); it != op[0][t]->end(); ++it) {
			if (*it != 0 || *it != t) {
				for (set<int>::const_iterator it2 = op[0][*it]->begin(); it2 != op[0][*it]->end(); ++it2)
					if (op[0][t]->find(*it2) == op[0][t]->end())
						return false;
			}
		}

	return true;

}



int main(int argc, char **argv) {

	if (argc < 3) {
		cout << "Usage: " << argv[0] << " filename filename2 [filename3 ...]" << endl;

		return 0;
	
	}

	vector<hea::Hea> non_iso_list;
	bool is_first = true;

	const int n = 6;

	set<int> setCandidates[1 << MAX];
	const set<int>* op[MAX][MAX];

	// init
	for (int i = 0; i < 1 << n; i++) {
		setCandidates[i] = set<int>();
		hea::init_set_i(setCandidates[i], i);
	}
	
	int complement[MAX];
	
	// loads all models from filename into non_iso_list
	{
		ifstream ifs;

		ifs.open(argv[1], ifstream::in);
		
		if (ifs.good()) {

			
			cerr << "Loading file " << argv[1] << "." << endl;

			while(hea::parse_model(ifs, n, op, complement, setCandidates)) {

				is_first = false;
				hea::Hea tested_hea(n, op, complement);

#ifndef JUST_ELIMINATE
				hea::disp_hea(cout, tested_hea);
#endif
				//memory allocation testing
				try {

					non_iso_list.push_back(tested_hea);
				}
				catch (std::bad_alloc &ba) {
					cerr << "Memory allocation error while loading: " << endl;
					hea::disp_hea(cerr, tested_hea);
					non_iso_list.clear();
					ifs.close();
					return 1;
				}

				// skip one line containing -M-
				char buffer[1024];
				ifs.getline(buffer, 1024);
			}
				
		}
		else {
			cerr << "Cannot open or read from file '" << argv[1] << "'." << endl;
		}
		ifs.close();
	}

	int idx = 2;
	while (idx < argc) {
		ifstream ifs;

		ifs.open(argv[idx], ifstream::in);
		
		if (ifs.good()) {

			cerr << "Loading file " << argv[idx] << "." << endl;

			while(hea::parse_model(ifs, n, op, complement, setCandidates)) {

				hea::Hea tested_hea(n, op, complement);

				if (is_first) {
					is_first = false;
					hea::disp_hea(cout, tested_hea);
					non_iso_list.push_back(tested_hea);
				}
				else {

					bool is_iso = false;
					for (vector<hea::Hea>::const_iterator it_vec = non_iso_list.begin(); it_vec != non_iso_list.end(); ++it_vec) {
						if (are_isomorphic(*it_vec, tested_hea, setCandidates)) {
							is_iso = true;
							break;
						}
					}

					if (!is_iso) {
						hea::disp_hea(cout, tested_hea);
						try {
							non_iso_list.push_back(tested_hea);
						}
						catch (std::bad_alloc &ba) {
							cerr << "Memory allocation error while storing:" <<endl;
							hea::disp_hea(cerr, tested_hea);
							ifs.close();
							non_iso_list.clear();
							return 1;
						}
					}
				}

				// skip one line containing -M-
				char buffer[1024];
				ifs.getline(buffer, 1024);
			}
				
		}
		else {
			cerr << "Cannot open or read from file '" << argv[idx] << "'." << endl;
		}
		ifs.close();

		idx++;
	}

	return 0;
}
