#include <iomanip>
#include <string.h>
#include <stdlib.h>
#include <fstream>
#include <vector>
#include <set>

#include "hyperEA3_common.h"
#include "hea.h"
#include "hyperEA3_parse.h"
	

using namespace std;

void disp_set(const int n, const set<int> &ds) {
/*
	for (set<int>::iterator it = ds.begin(); it != ds.end(); ++it) {
		cout << *it << ", ";
	}
//	cout << endl;
*/

	set<int>::iterator it = ds.begin();
	if (it != ds.end()) {
		//cout << *it;
		hea::disp_elem(cout, n, *it);
		++it;
		do {
			if (it != ds.end()) {
				cout << ", "; 
				hea::disp_elem(cout, n, *it);
				++it;
			}
		} while (it != ds.end());
	}

}

bool has_unique_orthosupplement_elem(const int n, const set<int>* (&op)[MAX][MAX], int a) {
	bool found = false;

	for (int i = 0; i < n ; i++)

		if (op[a][i]->find(n - 1) != op[a][i]->end()) {
			if (!found)
				found = true;
			else
				return false; /* not unique element */
			
		}

	return found;
}

bool has_unique_orthosupplement(const int n, const set<int>* (&op)[MAX][MAX]) {
	for (int i = 0; i < n; i++)
		if (!has_unique_orthosupplement_elem(n, op, i))
			return false;

	return true;
}


void disp_op(const int n, const set<int>* const (&op)[MAX][MAX]) {
	for (int i = 0; i < n; i++) {

		for (int j = 0; j < n; j++) {
			cout << "{";
			disp_set(n, *op[i][j]);
			cout << "}, ";
		}
		cout << endl;
			
	}
	
}

void compute_complement(const int n, const set<int>* (&op)[MAX][MAX], int (&complement)[MAX]) {
/*
	cout << "computing complement for: " << endl;
	disp_op(n, op);
	cout << "end complement" << endl;
*/

//cout << "HERE" << endl;
	complement[0] = n - 1;
	complement[n - 1] = 0;

	for (int i = 1; i < n - 1; i++) {
		//cout << "c: " << i;
		for (int j = n - 2; j>=1; j--) {
		//	cout << j << " ";
			//disp_set(*op[i][j]);
			if (op[i][j]->find(n-1) != op[i][j]->end()) {
				complement[i] = j;
				break;
			}
		}
		//cout << endl;

			
		
	}
}


bool test_subsets(const int n, const set<int> *(&op)[MAX][MAX]) {
	for (int t = 1 ; t < n; t++)
		for (set<int>::const_iterator it = op[0][t]->begin(); it != op[0][t]->end(); ++it) {
			if (*it != 0 || *it != t) {
				for (set<int>::const_iterator it2 = op[0][*it]->begin(); it2 != op[0][*it]->end(); ++it2)
					if (op[0][t]->find(*it2) == op[0][t]->end())
						return false;
			}
		}

	return true;

}

class heaCompare {

public:
	bool operator() (const hea::Hea &hea1, const hea::Hea &hea2) const {
		// assert hea1.n == hea2.n
		if (hea1.n != hea2.n)
			cerr << "Warning: structures are not of same size";
		// assert commutativity
		int i = 0, j = 0;

		while (i < hea1.n && hea1.op[i][j] == hea2.op[i][j]) {
			if (j < hea1.n - 1)
				j++;
			else  {
				i++;
				j = i;
			}
		}
		
		if (i == hea1.n)
			// the same
			return false;

		// i < hea1.n
		if (hea1.op[i][j] < hea2.op[i][j])
			return true;
		else
			return false;
			
	}
};



int main(int argc, char **argv) {
#if 0
// ** test part
	
	set<hea::Hea, heaCompare> test_set;
	const set<int> *test_op[MAX][MAX]={{NULL, NULL},{NULL, NULL}};
	int test_complement[MAX]={1,0};

	test_set.insert(hea::Hea(2, test_op, test_complement));



	return 0;
// -- test part
#endif

	if (argc < 2) {
		cout << "Usage: " << argv[0] << " filename " << endl;

		return 0;
	
	}

//	vector<hea::Hea> non_iso_list;
	set<hea::Hea, heaCompare> iso_images_set;
	bool is_first = true;

	const int n = 6;

	set<int> setCandidates[1 << MAX];
	const set<int>* op[MAX][MAX];

	// init
	for (int i = 0; i < 1 << n; i++) {
		setCandidates[i] = set<int>();
		hea::init_set_i(setCandidates[i], i);
	}
	
	int complement[MAX];
	
	vector<hea::Permutation> permutation_candidates;

	// load all isomorphic images to iso_images_set
	{
		ifstream ifs;

		ifs.open(argv[1], ifstream::in);
		
		if (ifs.good()) {

			
			cerr << "Loading file " << argv[1] << "." << endl;

			while(hea::parse_model(ifs, n, op, complement, setCandidates)) {

				hea::Hea tested_hea(n, op, complement);

				if (is_first) {
					is_first = false;

					generate_all_order_preserving_permutations(tested_hea, false, permutation_candidates);

					select_plan_preserving_permutations_old(tested_hea, permutation_candidates);

					cerr << "Order preserving permutations: " << endl;
					cerr << "----------------------------------------" << endl;
					for (vector<hea::Permutation>::const_iterator it_p_vec = permutation_candidates.begin() ; it_p_vec != permutation_candidates.end() ; ++it_p_vec) {
						for (int i = 0; i < it_p_vec->n; i++)
							cerr << it_p_vec->permutation[i] << " ";
						cerr << endl;
					}
					cerr << "------------------ C ------------------" << endl;

					



				}

				if (iso_images_set.find(tested_hea) == iso_images_set.end()) {

					hea::disp_hea(cout, tested_hea);
					for (vector<hea::Permutation>::const_iterator it_p_vec = permutation_candidates.begin() ; it_p_vec != permutation_candidates.end() ; ++it_p_vec) {
						const set<int> *op_iso[MAX][MAX];
						int complement_iso[MAX];

						/* compute iso_image */
						hea::iso_image_full(n, op, complement, it_p_vec->permutation, setCandidates, op_iso, complement_iso);
						hea::Hea hea_iso(n, op_iso, complement_iso);

	#if 0
						/* test debug */

						for (int i = 0; i < n; i++)
							cerr << i << "  ";
						cerr << endl;

						for (int i = 0; i < n; i++)
							cerr << complement[i] << "  ";
						cerr << endl;

						for (int i = 0; i < n; i++)
							cerr << complement_iso[i] << "  ";
						cerr << endl;
						/* end test debug */
	#endif

						//memory allocation testing
						try {

							iso_images_set.insert(hea_iso);
						}
						catch (std::bad_alloc &ba) {
							cerr << "Memory allocation error while loading: " << endl;
							hea::disp_hea(cerr, hea_iso);
							iso_images_set.erase(iso_images_set.begin(), iso_images_set.end());
							ifs.close();
							return 1;
						}

					}
				}

				// skip one line containing -M-
				char buffer[1024];
				ifs.getline(buffer, 1024);
			}
				
		}
		else {
			cerr << "Cannot open or read from file '" << argv[1] << "'." << endl;
		}
		ifs.close();
	}

	return 0;
}
