#include <iostream>
#include <set>
#include <vector>
#include <map>
#include <stdlib.h>
#include <ctime>
#include <algorithm>
#include <numeric>

#include "hyperEA3_common.h"
#include "permutation.h"
#include "candidate.h"

using namespace std;


struct fromToState {
	bool bFromState;
	int from_state[MAX*MAX];
	bool used_from_state[MAX*MAX];	

	bool bToState;
	int to_state[MAX*MAX];
	bool used_to_state[MAX*MAX];	
	
	fromToState() {
		bFromState = false;
		bToState = false;
		for (int i = 0; i < MAX*MAX; i++) {
			from_state[i] = 0;
			to_state[i] = 0;
			used_from_state[i] = false;
			used_to_state[i] = false;
		}
				
	}
};

struct Relation {
	int n;
	int relation[MAX][MAX];

	Relation(const int _n, const int (&_relation)[MAX][MAX]) {
		n = _n;

		for (int i = 0; i < _n; i++)	
		for (int j = 0; j < _n; j++)	

			relation[i][j] = _relation[i][j];
	}
};

struct Zoznamy {
	set<pair<int, int> > one_pairs;
	set<pair<int, int> > other_pairs;

	Zoznamy() {
	}

	Zoznamy( const set<pair<int, int> > &_one_pairs, const set<pair<int, int> > &_other_pairs ) {
		for (set<pair<int, int> >::const_iterator it = _one_pairs.begin(); it != _one_pairs.end(); ++it) {
			one_pairs.insert(*it);
		}

		for (set<pair<int, int> >::const_iterator it = _other_pairs.begin(); it != _other_pairs.end(); ++it) {
			other_pairs.insert(*it);
		}
	}
};



bool are_rel_isomorphic_via_perm(const int n, const int (&permutation)[MAX], const int (&relation)[MAX][MAX], const Relation &relation2) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if (relation[i][j] != relation2.relation[ permutation[i] ][ permutation[j] ])
			return false;

	return true;

}

void are_isomorphic_recursive(int level, int (&permutation)[MAX], const int n, const int (&relation)[MAX][MAX], const Relation &relation2, bool &are_iso) {
	if (!are_iso) {
		if (level == n - 1) {

			if (are_rel_isomorphic_via_perm(n, permutation, relation, relation2))
				are_iso = true;

		}
		else {
			for (int i=level; i < n; i++) {
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
				
				are_isomorphic_recursive(level + 1, permutation, n, relation, relation2, are_iso);
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}

			}
		}
	}
}

/* global */
unsigned long int number_of_tested_configurations=0UL, number_of_cycles=0UL;
const long int dump_state_count=10000000UL;
/* */
int compute_set_value(int n, const set<int> *setc) {
	int val = 0;	

	for (set<int>::const_iterator itset = setc->begin(); itset != setc->end(); ++itset)
		val+= (1 << *itset);
	
	return val;
}

bool are_isomorphic(const int n, const int (&relation)[MAX][MAX], const Relation &relation2) {
	bool are_iso = false;
	int permutation[MAX];
	for (int i = 0; i < n; i++)
		permutation[i] = i;

	are_isomorphic_recursive(0, permutation, n, relation, relation2, are_iso);

	return are_iso;
}

bool is_isomorphic(const int n, const int (&relation)[MAX][MAX], const vector<Relation> &relations) {
	/* first relation */
	if (relations.empty())
		return false;


	for (vector<Relation>::const_iterator it = relations.begin(); it != relations.end(); ++it) {
		if (are_isomorphic(n, relation, *it))
			return true;
	}

	return false;
}

bool is_double_negation_permutation(const int n, const int (&permutation)[MAX]) {
	for (int i = 0; i < n; i++)
		if (i != permutation [ permutation [ i ] ])
			return false;

	return true;
}



void admits_reverse_of_relation_recursive(int level, int (&permutation)[MAX], const int n, const int (&relation)[MAX][MAX], bool &found) {
	if (!found) {
		if (level == n - 1) {
			bool is_valid  = true;
			for (int i = 0; i < n && is_valid; i++)
			for (int j = 0; j < n && is_valid; j++)
				if (i != j)
					if (relation[i][j] && !relation[permutation[j]][permutation[i]]) {
						/* wrong permutation */
						is_valid = false;
						break;
					}

			if (is_valid) {
				found = true;


#if 0
				/* test permutation */	
				for (int i = 0; i < n; i++)
					cout << permutation[i] << " ";
				cout << endl;
				cout << "-P-" << endl;
#endif
			}
		}
		else {
			for (int i=level; i < n; i++) {
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
				
				admits_reverse_of_relation_recursive(level + 1, permutation, n, relation, found);
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
					
			}
				
		}
	}
}


bool admits_reverse_of_relation(const int n, const int (&relation)[MAX][MAX]) {
	
	bool found = false;
	int permutation[MAX];
	for (int i = 0; i < n; i++)
		permutation[i] = i;

	admits_reverse_of_relation_recursive(0, permutation, n, relation, found); 

	return found;
	
}

void get_vector_of_orthosupplement_candidates(int level, int (&permutation)[MAX], const int n, const int (&relation)[MAX][MAX], vector<hea::Permutation> &vector_ortho) {

	if (level == n - 1) {
		bool is_valid  = true;
		if (!is_double_negation_permutation(n, permutation))
			is_valid = false;

		for (int i = 0; i < n && is_valid; i++)
		for (int j = 0; j < n && is_valid; j++)
			if (i != j)
				if (relation[i][j] && !relation[permutation[j]][permutation[i]]) {
					/* wrong permutation */
					is_valid = false;
					break;
				}

		if (is_valid) {
			vector_ortho.push_back(hea::Permutation(n, permutation));


#if 0
			/* test permutation */	
			for (int i = 0; i < n; i++)
				cout << permutation[i] << " ";
			cout << endl;
			cout << "-P-" << endl;
#endif
		}
	}
	else {
		for (int i=level; i < n; i++) {
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
			
			get_vector_of_orthosupplement_candidates(level + 1, permutation, n, relation, vector_ortho);
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
				
		}
			
	}
}

/* mapping x-> x', such that x<=y, iff y'<=x' 
 *  returns true, if there is at least one such mapping
 */
bool get_vector_of_orthosupplement_candidates(const int n, const int (&relation)[MAX][MAX], vector<hea::Permutation> &vector_ortho) {
	
	int permutation[MAX];
	for (int i = 0; i < n; i++)
		permutation[i] = i;

	get_vector_of_orthosupplement_candidates(0, permutation, n, relation, vector_ortho); 

	return !vector_ortho.empty();
}


bool is_similar_pairs_via_perm(const int n, const Zoznamy &zoznamy, const set<pair<int, int> > &one_pairs, const set<pair<int, int> > &other_pairs, const int (&permutation)[MAX]) {
	//set<pair<int, int> > one_iso_pairs, other_iso_pairs;

	for (set<pair<int, int> >::const_iterator it_set = one_pairs.begin(); it_set != one_pairs.end(); ++it_set) {
		pair<int, int> iso_pair(permutation[it_set->first - 1] + 1, permutation[it_set->second - 1] + 1);
		if (zoznamy.one_pairs.find(iso_pair) == zoznamy.one_pairs.end())
			/* not found */
			return false;


	//	one_iso_pairs.insert(pair<int, int>(permutation[it_set->first], permutation[it_set->second]));	

	}

	for (set<pair<int, int> >::const_iterator it_set = other_pairs.begin(); it_set != other_pairs.end(); ++it_set) {
		pair<int, int> iso_pair(permutation[it_set->first - 1] + 1, permutation[it_set->second - 1] + 1);
		if (zoznamy.other_pairs.find(iso_pair) == zoznamy.other_pairs.end())
			/* not found */
			return false;


	//	one_iso_pairs.insert(pair<int, int>(permutation[it_set->first], permutation[it_set->second]));	

	}

	return true;
}

/* is  zoznamy isomorphic to one_pairs, other_pairs*/
void is_similar_pairs_recursive(int level, const int n, const Zoznamy &zoznamy, const set<pair<int, int> > &one_pairs, const set<pair<int, int> > &other_pairs, int (&permutation)[MAX], bool &is_similar) {
	if (!is_similar) {
		if (level == n - 1) {
/*
			for (int i = 0; i < n; i++)
				cout << permutation[i] << " ";
			cout << endl;
*/

			if (is_similar_pairs_via_perm(n, zoznamy, one_pairs, other_pairs, permutation)) {
/*
				cout << "Permutation: ";
				for (int i = 0; i < n; i++)
					cout << permutation[i] ;
				cout << endl;
*/
				is_similar = true;
			}

		}
		else {

			for (int i = level; i < n; i++) {
				if (i!=level) {
					int tmp = permutation[i];
					permutation[i] = permutation[level];
					permutation[level] = tmp;
				}

				is_similar_pairs_recursive(level + 1, n, zoznamy, one_pairs, other_pairs, permutation, is_similar);

				if (i!=level) {
					int tmp = permutation[i];
					permutation[i] = permutation[level];
					permutation[level] = tmp;
				}

			}

		}

	}


} 

/* is  zoznamy isomorphic to one_pairs, other_pairs*/
bool is_similar_pairs(const int n, const Zoznamy &zoznamy, const set<pair<int, int> > &one_pairs, const set<pair<int, int> > &other_pairs) {
	int permutation[MAX];
	for (int i = 0; i < n; i++)
		permutation[i] = i;

	bool is_similar = false;

	is_similar_pairs_recursive(0, n, zoznamy, one_pairs, other_pairs, permutation, is_similar);

	return is_similar;
} 

/* search for vec_zoznamy wheater it contains isomorphic (one, other) pairs to one_pairs and other_pairs */
bool is_similar_vec_pairs(const int n, const vector<Zoznamy> &vec_zoznamy, const set<pair<int, int> > &one_pairs, const set<pair<int, int> > &other_pairs) {
	for (vector<Zoznamy>::const_iterator it_vec = vec_zoznamy.begin(); it_vec != vec_zoznamy.end(); ++it_vec) {

		if (is_similar_pairs(n, *it_vec, one_pairs, other_pairs))
			return true;
		
	}

	return false;

} 

/* reduces summable pairs with respect to commutativity property 
 * i.e. it removes (i, j), (j, i) duplicates and replaces it with 
 * just (i, j), i<=j.
 */
void reduce_summable_pairs(const Zoznamy &zoznamy, Zoznamy &reduced) {
	for (set<pair<int, int> >::const_iterator it_set = zoznamy.one_pairs.begin(); it_set != zoznamy.one_pairs.end(); ++it_set)
		if (it_set->first <= it_set->second)
			reduced.one_pairs.insert(*it_set);	

	for (set<pair<int, int> >::const_iterator it_set = zoznamy.other_pairs.begin(); it_set != zoznamy.other_pairs.end(); ++it_set)
		if (it_set->first <= it_set->second)
			reduced.other_pairs.insert(*it_set);	

}

#if 0
struct Candidate {
	pair<int, int> candidate_pair;
	bool	is_one; // true one_pairs, false other_pairs

	Candidate(const pair<int, int> &_candidate_pair, bool _is_one) {
		is_one = _is_one;
		candidate_pair.first = _candidate_pair.first;
		candidate_pair.second = _candidate_pair.second;
	}

	bool operator<(const Candidate &cand2) const {
		return (candidate_pair.first < cand2.candidate_pair.first || 
				(candidate_pair.first == cand2.candidate_pair.first && candidate_pair.second < cand2.candidate_pair.second));
	}
};
#endif


#if 0
// C++11 :(
void init_last_cand_for_x_set11(const int n, const set<Candidate> &candidates, set<pair<int, int> > &last_cand_for_x) {
	int last_x = n - 1;
	bool was_inserted = false;

	for (set<Candidate>::const_iterator it_cand = candidates.crbegin(); it_cand != candidates.crend(); ++it_cand) {
		if (!was_inserted) {
			last_cand_for_x.insert(it_cand->candidate_pair);
			was_inserted = true;
		}
		else {
			if (last_x > it->candidate_pair.first) {
				last_x = it->candidate_pair.first;
				last_cand_for_x.insert(it_cand->candidate_pair);
			}
		}
	}
}
#endif

/* check associativity at 0 

x + y candidate A:

0 + (x + y) = (0 + x) + y = x + (0 + y)

x + 0 = x || y + 0 = y 
	-> A + 0 = A
x + 0 = {0, x}
	-> A + 0 = A union (0 + y)
y + 0 = {0, y)
	-> A + 0 = A union (0 + x)
*/

/* set candidate: i */
bool is_associative_at_0(int i, int x, int y, int n, const set<int> *(&op)[MAX][MAX], set<int> *setCandidates) {
	set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
	
	if (op[0][x] == &setCandidates[1<<x]) {

		if (resA0 != setCandidates[i])
			return false;


	}
	else
	if (op[0][x] == &setCandidates[1 + (1<<x)]) {
		set<int> res_0x_y = setCandidates[i];
		for (set<int>::const_iterator it_0y = op[0][y]->begin();it_0y != op[0][y]->end(); ++it_0y) 
			res_0x_y.insert(*it_0y);

		if (resA0 != res_0x_y)
			return false;
	}

	/* can be summed, i.e. all necessary subresults for computing  x + (0 + y) are already defined (x  + y = setCandidate[i]) */
	bool can_be_summed_x = true;
	for (set<int>::const_iterator it_0y = op[0][y]->begin();it_0y != op[0][y]->end(); ++it_0y) {
		if ( *it_0y > y) {
			can_be_summed_x = false;
			break;
		}

	}
	
	if (can_be_summed_x) {
		set<int> res_x_0y = setCandidates[i];
		for (set<int>::const_iterator it_0y = op[0][y]->begin();it_0y != op[0][y]->end(); ++it_0y) {
			if (*it_0y < y) {
				if (op[x][*it_0y] == &setCandidates[0])
					cerr << "Problem summed_x...?" << endl;	
					
				for (set<int>::const_iterator it_xy = op[x][*it_0y]->begin(); it_xy != op[x][*it_0y]->end(); ++it_xy)
					res_x_0y.insert(*it_xy);
			}
		}

		if (res_x_0y != resA0)
			return false;
		
	}

	if (op[0][y] == &setCandidates[1<<y]) {

		if (resA0 != setCandidates[i])
			return false;


	}
	else
	if (op[0][y] == &setCandidates[1 + (1<<y)]) {
		set<int> res_x_0y = setCandidates[i];
		for (set<int>::const_iterator it_0x = op[0][x]->begin();it_0x != op[0][x]->end(); ++it_0x) 
			res_x_0y.insert(*it_0x);

		if (resA0 != res_x_0y)
			return false;
	}


	/* can be summed, i.e. all necessary subresults for computing  (0 + x) + y are already defined (x  + y = setCandidate[i]) */
	bool can_be_summed_y = true;
	for (set<int>::const_iterator it_0x = op[0][x]->begin();it_0x != op[0][x]->end(); ++it_0x) {
		if ( *it_0x > x) {
			can_be_summed_y = false;
			break;
		}

	}
	
	if (can_be_summed_y) {
		set<int> res_0x_y = setCandidates[i];
		for (set<int>::const_iterator it_0x = op[0][x]->begin();it_0x != op[0][x]->end(); ++it_0x) {
			if (op[*it_0x][y] == &setCandidates[0])
				cerr << "Problem summed_y...?" << endl;	

			if (*it_0x < x) {
				for (set<int>::const_iterator it_xy = op[*it_0x][y]->begin(); it_xy != op[*it_0x][y]->end(); ++it_xy)
					res_0x_y.insert(*it_xy);
			}
		}

		if (res_0x_y != resA0)
			return false;
		
	}

	return true;
	
}



void clear_used_from(fromToState &current_setup, int level) {
	int j = level + 1;
	while (j < MAX * MAX)  {
		current_setup.used_from_state[j] = false;
		j++;
	}

}

int sumpower2(int sumsofar, int val) {
	return sumsofar + (1<<val);
}


void solve_candidates_state_setup(set<Candidate>::const_iterator it_cand, int level, const set<Candidate> &candidates, const set<pair<int, int> > &last_cand_for_x, const int n, const set<int> *(&op)[MAX][MAX], const int (&complement)[MAX], set<int> *setCandidates, fromToState &current_setup, const bool only0x[MAX], map<pair<int, int>, set<int> > not_subset_of_map) {

	if (it_cand == candidates.end()) {
		

		number_of_tested_configurations++;
		if (number_of_tested_configurations == dump_state_count) {
			number_of_cycles++;
			number_of_tested_configurations = 0UL;

			/* dump state */

			cerr << "State: ";
			for (set<Candidate>::const_iterator itcan = candidates.begin(); itcan != candidates.end(); ++itcan) {

				int val = compute_set_value(n, op[itcan->candidate_pair.first][itcan->candidate_pair.second]);
				cerr << val << "," ; 
			}
			cerr << endl;
		}


/*
				for (int i =0 ; i < n; i++)
					cout << complement[i] << "  ";
				cout << endl;
*/


		if (hea::is_associative(n, op) /*&& has_unique_orthosupplement(n, op)*/) {
			/*int complement[MAX];*/
			/*for (int i = 0; i < n; i++)
				complement[n-1-i] = i;*/

			/*compute_complement(n, op, complement);*/

			if (hea::is_antisymmetric(n, op, complement) /*&& 
				(!transitive || hea::is_transitive(n, op, complement))*/
			) 
			{
				/* display complement */
#if 0
				for (int i =0 ; i < n; i++) {
					hea::disp_elem(cout, n, i);
					cout << "  ";
				}
				cout << endl;

				for (int i =0 ; i < n; i++) {
					hea::disp_elem(cout, n, complement[i]);
					cout << "  ";
				}
				cout << endl;
#endif

				hea::disp_complement(cout, n, complement);
				hea::disp_op(cout, n, op);
				cout << "\t-M-" << endl;
			}
		}

		//disp_op(n, op);
	}
	else {
		if (it_cand->candidate_pair.first == 0) {
			if (it_cand->is_one) {
				/* contains 1 == n - 1 */

				int from;
				bool was_cleared = true; 
				if (current_setup.bFromState && current_setup.used_from_state[level]) {
					current_setup.used_from_state[level] = false;
					from = current_setup.from_state[level];
					was_cleared = false;
				}
				else
					from = 1 << (n - 1);

				int to;
				if (current_setup.bToState && current_setup.used_to_state[level]) {
					current_setup.used_to_state[level] = false;
					to = current_setup.to_state[level];
				}
				else
					to = (1 << n) - 1;

				for (int i = from; i <= to; i++) {

#ifdef DEBUG
					for (int q = 0; q < level; q++)
						cout << " ";
					cout << i << endl;
#endif

					/* for t = y - 1, y - 2, ... 0, if candidate for [0][y] contains t
					 * then it needs to contain all elements of [0][t] 
					 */

					if (hea::test_subsets_new(n, i, op, it_cand->candidate_pair.second, setCandidates)) {

						/* b \in 0 + y => b + y' are defined */
						bool correct = true;
						for (set<int>::const_iterator it_summable = setCandidates[i].begin(); it_summable != setCandidates[i].end(); ++it_summable) {
							// skip trivial checks 0, y
							if (*it_summable != 0 && *it_summable != it_cand->candidate_pair.second) {
								pair<int, int> my_pair;
								int tmp_compl = complement[ it_cand->candidate_pair.second ];

								if (*it_summable < tmp_compl) {
									my_pair.first = *it_summable;
									my_pair.second = tmp_compl;
								}
								else {
									my_pair.first = tmp_compl;
									my_pair.second = *it_summable;
								}
								
								if (candidates.find(Candidate(my_pair, false)) == candidates.end()) {
									correct = false;	
									break;
								}
										
							}
} 
						if (correct) {

							op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
							op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

							set<Candidate>::const_iterator it_new = it_cand;
							it_new++;
							solve_candidates_state_setup(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, current_setup, only0x, not_subset_of_map);		

							// unset filled elements - empty set
							op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
							op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
						}
						else {
							/* we are moving to another i (from + 1), thus we need to set all successive bFromState to false */
							if (!was_cleared) {
								clear_used_from(current_setup, level);

								was_cleared = true;
							}

						}
					}
				}
			}
			else {
				/* does not contain one */
				/* it must be defined - excludes i = 0 */
				/* {0} cannot be used any more  - excludes i = 1 */

				int from;
				bool was_cleared = true; 
				if (current_setup.bFromState && current_setup.used_from_state[level]) {
					current_setup.used_from_state[level] = false;
					from = current_setup.from_state[level];

					was_cleared = false; 
				}
				else
					from = /*2*/(1 << it_cand->candidate_pair.second);

				int to;
				if (current_setup.bToState && current_setup.used_to_state[level]) {
					current_setup.used_to_state[level] = false;
					to = current_setup.to_state[level];
				}
				else
					to = (1 << (n - 1)) - 1;
				
				for (int i = from; i <= to; i++) {

					/* x\in 0 + x */
					if (setCandidates[i].find(it_cand->candidate_pair.second) != setCandidates[i].end()) {
#ifdef DEBUG
						for (int q = 0; q < level; q++)
							cout << " ";
						cout << i << endl;
#endif

						bool was_used = false;

						for (int k = it_cand->candidate_pair.second - 1; k >= 1; k--) {
							if (op[0][k] == &setCandidates[i]) {
								was_used = true;	
								break;
							}
						}
						
						if (was_used) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						// Candidate for 0+b: b<=c, then (0+b) is summable with c'
						//cout << "Level: " << level << ", candidate: " << i << endl;
						int b = it_cand->candidate_pair.second;
						bool is_satisfied = true;

						// 1==n-1 cannot be contained and 0 is trivial, skipping
						for (int ii = 1; is_satisfied && ii < n - 1; ii++) {
							// skipping also b as trivial
							if (ii == b)
								continue;

							// b <= ii, i.e. b+ii' is summable
							if (is_summable_candidate(candidates, complement[ii], b)) {
								//cout << "ii: " << ii << ", ii' = " << complement[ii] << endl;
								for (set<int>::const_iterator it_sat = setCandidates[i].begin(); it_sat != setCandidates[i].end(); ++it_sat) {
									// skip trivial checks
									if (*it_sat == 0 || *it_sat == b)
										continue;
									
									if (!is_summable_candidate(candidates, *it_sat, complement[ii])) {
										
										is_satisfied = false;
										break;
									}
								} 
								
							}
						}
						
						if (!is_satisfied) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						// --


						/* for t = y - 1, y - 2, ... 0, if candidate for [0][y] contains t
						 * then it needs to contain all elements of [0][t] 
						 */

						if (hea::test_subsets_new(n, i, op, it_cand->candidate_pair.second, setCandidates)) {


							/* b \in 0 + y => b + y' are defined */
							bool correct = true;
							for (set<int>::const_iterator it_summable = setCandidates[i].begin(); it_summable != setCandidates[i].end(); ++it_summable) {
								// skip trivial checks 0, y
								if (*it_summable != 0 && *it_summable != it_cand->candidate_pair.second) {
									pair<int, int> my_pair;
									int tmp_compl = complement[ it_cand->candidate_pair.second ];

									if (*it_summable < tmp_compl) {
										my_pair.first = *it_summable;
										my_pair.second = tmp_compl;
									}
									else {
										my_pair.first = tmp_compl;
										my_pair.second = *it_summable;
									}
									
									if (candidates.find(Candidate(my_pair, false)) == candidates.end()) {
										correct = false;	
										break;
									}
											
								}
	} 
							if (correct) {

								op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
								op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

								set<Candidate>::const_iterator it_new = it_cand;
								it_new++;
								solve_candidates_state_setup(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, current_setup, only0x, not_subset_of_map);		

								// unset filled elements - empty set
								op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
								op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
							}
							else {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
							}
						}
						
					}
				}
			}


		}
		else {
			if (it_cand->is_one) {
/*
				complement[it_cand->candidate_pair.first] = it_cand->candidate_pair.second;
				complement[it_cand->candidate_pair.second] = it_cand->candidate_pair.first;
*/
				/* contains 1 */

				int from;
				bool was_cleared = true; 

				if (current_setup.bFromState && current_setup.used_from_state[level]) {
					current_setup.used_from_state[level] = false;
					from = current_setup.from_state[level];
					was_cleared = false;
				}
				else
					from = 1 << (n - 1);

				int to;
				if (current_setup.bToState && current_setup.used_to_state[level]) {
					current_setup.used_to_state[level] = false;
					to = current_setup.to_state[level];
				}
				else
					to = (1 << n) - 1;

#ifdef LINEAR7incomplete
				set<int> res_a_ad;
				/* b + d */
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 4) {
					/* a + a = \{b\}, a + d = \{e\} = > a + e = a + (a + d) = (a + a) + d = b + d */
					if (op[1][1] == &setCandidates[4] && op[1][4] == &setCandidates[32]) {
						int num = (op[1][5] - &setCandidates[0]);

						if (from > num || to < num) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							return;
						}
						else 
							from = to = num;
					}
					else
					/* a + a = \{b\} => b + d = (a + a) + d = a + (a + d) */
					if (op[1][1] == &setCandidates[4]) {
						res_a_ad = hea::compute_sum(op, 1, *(op[1][4]));
						int num = accumulate(res_a_ad.begin(), res_a_ad.end(), 0, sumpower2);

						if (from > num || to < num) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							return;
						}
						else 
							from = to = num;

					}
					else
					/* a + a = \{0, b\} => a + (a + d) = (0 + d) union (b + d), 
					   i.e. it is possible test (b + d) results */

					/* a + a = \{a, b\} => a + (a + d) = (a + d) union (b + d) */

					/* a + a = \{0, a, b\} => a + (a + d) = (0 + d) union (a + d) union (b + d) */
					if (op[1][1] == &setCandidates[5] || op[1][1] == &setCandidates[6] || op[1][1] == &setCandidates[7]) {
						res_a_ad = hea::compute_sum(op, 1, *(op[1][4]));
					}
				}
				

				/* c + c */
				if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 3) {
					/* a + b = \{c\}, a + c = \{d\} => c + c = (a + b) + c = b + (a + c) = b + d */
					if (op[1][2] == &setCandidates[8] && op[1][3] == &setCandidates[16]) {
						int num = (op[2][4] - &setCandidates[0]);

						if (from > num || to < num) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							return;
						}
						else 
							from = to = num;
					}
				}
#endif				


#if (modelno == 24)
	#if (resaa == 2)
				/* only for a + a and model number 24 and a + 0 = {a} ! */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
					int from_new = op[0][n - 1] - &setCandidates[0];

						//std::cerr << "From new: " << from_new << endl;

					if (from_new > from) {
						from = from_new;
						if (!was_cleared) {
							clear_used_from(current_setup, level);
							was_cleared = true;
						}
					}
				}
	#else
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
					int from_new = op[0][n - 1] - &setCandidates[0];
					from_new &= ~3;

						//std::cerr << "From new: " << from_new << endl;

					if (from_new > from) {
						from = from_new;
						if (!was_cleared) {
							clear_used_from(current_setup, level);
							was_cleared = true;
						}

					}
				}
				
	#endif
#endif

				for (int i = from; i <= to; i++) {
#if (modelno == 24) && (resaa == 3)
				/* a + 0 = {0, a} */
				/* a + a = A, potom 
						1) {0, a} union A = A + 0 
						2) 1 + 0 \subseteq A + 0 = {0, a} union A
				*/
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
					/* compute setCandidates[i] union {0, a} */
					set<int> resAa0 = setCandidates[i];
					resAa0.insert(0); // insert element 0
					resAa0.insert(1); // insert element a

					/* compute (setCandidates[i]) + 0 */
					set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);

					if (resAa0 != resA0) {
						if (!was_cleared) {
							clear_used_from(current_setup, level);
							was_cleared = true;
						}
						continue;
					}
				}
#endif

				if (it_cand->candidate_pair.first == it_cand->candidate_pair.second) {
					if (only0x[it_cand->candidate_pair.first])
						if (!hea::testCandidateSetForXPlusXContainingOne(i, it_cand->candidate_pair.first, n, op, setCandidates)) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;

}
						
				}
				else {
					if (only0x[it_cand->candidate_pair.first]) {
						if (only0x[it_cand->candidate_pair.second]) {
							if (!hea::testCandidateSetForXPlusYBothContainingOne(i, it_cand->candidate_pair.first, it_cand->candidate_pair.second, n, op, setCandidates)) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;

							}
						}
						else
							if (!hea::testCandidateSetForXPlusYJustOneContainingOne(i, it_cand->candidate_pair.first, true, it_cand->candidate_pair.second, n, op, setCandidates)) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
	
					}
					else {
						if (only0x[it_cand->candidate_pair.second]) {
							if (!hea::testCandidateSetForXPlusYJustOneContainingOne(i, it_cand->candidate_pair.first, false, it_cand->candidate_pair.second, n, op, setCandidates)) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;

							}
						}
					}

				}
					/* x = y', then 0 + 1 \subseteq x + y, if 0+x = {x}, or 0+y = {y} ! */


					if (	(/*!only0x[it_cand->candidate_pair.first] && */op[0][it_cand->candidate_pair.first]->size() == 1) || 
			(it_cand->candidate_pair.first != it_cand->candidate_pair.second && op[0][it_cand->candidate_pair.second]->size() == 1)
					) {

						bool is_ok = true;
						for (set<int>::const_iterator it_01set = op[0][n - 1]->begin(); it_01set != op[0][n - 1]->end(); ++it_01set)
							if ( !(i & (1 << (*it_01set)))) {
								is_ok = false;
								break;
							}
						/* -- */

						if (!is_ok) {
							//cerr << "Reducing ..." << endl;
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}

#ifdef LINEAR7incomplete
					/* b + d */
					if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 4) {
						/* a + a = \{0, b\} => as before (a + a) + d = (0 + d) union (b + d), where 
						   candidate for b + d is setCandidates[i] */
						if (op[1][1] == &setCandidates[5]) {
							set<int> res_aa_d;
							set_union(setCandidates[i].begin(), setCandidates[i].end(), op[0][4]->begin(), op[0][4]->end(), inserter(res_aa_d, res_aa_d.end()));
							if (res_aa_d != res_a_ad) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;

							}
						}
						else
						/* a + a = \{a, b\} => as before (a + a) + d = (a + d) union (b + d), where 
						   candidate for b + b is setCandidates[i] */
						if (op[1][1] == &setCandidates[6]) {
							set<int> res_aa_d;
							set_union(setCandidates[i].begin(), setCandidates[i].end(), op[1][4]->begin(), op[1][4]->end(), inserter(res_aa_d, res_aa_d.end()));
							if (res_aa_d != res_a_ad) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;

							}
						
						}
						else
						/* a + a = \{0, a, b\} => as before (a + a) + d = (0 + d) union (a + d) union (b + d), where 
						   candidate for b + b is setCandidates[i] */
						if (op[1][1] == &setCandidates[7]) {
							/*
							set<int> res_aa_d_tmp, res_aa_d;
							set_union(setCandidates[i].begin(), setCandidates[i].end(), op[0][4]->begin(), op[0][4]->end(), inserter(res_aa_d_tmp, res_aa_d_tmp.end()));
							set_union(res_aa_d_tmp.begin(), res_aa_d_tmp.end(),  op[1][4]->begin(), op[1][4]->end(), inserter(res_aa_d, res_aa_d.end()));
							*/
							set<int> res_aa_d = setCandidates[i];
							res_aa_d.insert(op[0][4]->begin(), op[0][4]->end());
							res_aa_d.insert(op[1][4]->begin(), op[1][4]->end());
							if (res_aa_d != res_a_ad) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;

							}
						
						}
					}
#endif



					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

					if (!check_summability(candidates, it_cand->candidate_pair, op)) {
						if (!was_cleared) {
							clear_used_from(current_setup, level);
							was_cleared = true;
						}

						continue;
					}
#if 0
					bool valid = true;
					if (last_cand_for_x.find(it_cand->candidate_pair) != last_cand_for_x.end()) {

						// x is summable with all elements in y + 0, if x + y is summable
						for (int i = it_cand->candidate_pair.first; i < n - 1 && valid; i++)
							if (hea::is_summable(op, it_cand->candidate_pair.first, i))
								for (set<int>::const_iterator it = op[0][i]->begin(); it != op[0][i]->end(); ++it)
									if (!hea::is_summable(op, it_cand->candidate_pair.first, *it)) {
										valid = false;
										break;
									}
					}

					if (!valid)
						continue;
#endif

					set<Candidate>::const_iterator it_new = it_cand;
					it_new++;
					solve_candidates_state_setup(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, current_setup, only0x, not_subset_of_map);

					// unset
					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
				}

			}
			else {
				/* cannot contain 1, or be equal to {0}, or be empty */

				int from;
				bool was_cleared = true;

				if (current_setup.bFromState && current_setup.used_from_state[level]) {
					current_setup.used_from_state[level] = false;
					from = current_setup.from_state[level];

					was_cleared = false;
				}
				else
					from = 2;

				int to;
				if (current_setup.bToState && current_setup.used_to_state[level]) {
					current_setup.used_to_state[level] = false;
					to = current_setup.to_state[level];
				}
				else
					to = (1 << (n - 1)) - 1;

#ifdef LINEAR7incomplete
				/* b + b */
				set<int> res_a_ab;
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
					// a + a = \{b\}, a + b = \{c} => b + b = a + c
					if (op[1][1] == &setCandidates[4] && op[1][2] == &setCandidates[8]) {
						int num = (op[1][3] - &setCandidates[0]);

						if (from > num || to < num) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							return;

						}
						else {
							from = to = num;
						}
							

					}
					else
					/* a + a = \{b\} => a + (a + b) = b + b */
					if (op[1][1] == &setCandidates[4]) {
						set<int> res_a_ab = hea::compute_sum(op, 1, *(op[1][2]));
						int num =  accumulate(res_a_ab.begin(), res_a_ab.end(), 0, sumpower2);

						if (from > num || to < num) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							return;
						}
						else {
							from = to = num;
						}

					}
					else
					/* a + a = \{0, b\} => a + (a + b) = (a + a) + b = (0 + b) union (b + b) */

					/* a + a = \{a, b\} => a + (a + b) = (a + a) + b = (a + b) union (b + b) */

					/* a + a = \{0, a, b\} => a + (a + b) = (a + a) + b = (0 + b) union (a + b) union (b + b) */
					if (op[1][1] == &setCandidates[5] || op[1][1] == &setCandidates[6] || op[1][1] == &setCandidates[7]) {
						res_a_ab = hea::compute_sum(op, 1, *(op[1][2]));
					}
				}

				/* b + c */
				set<int> res_a_ac;
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 3) {
					// a + a = \{b\}, a + b = \{c}, a + c = \{d\} => b + c = a + d
					if (op[1][1] == &setCandidates[4] && op[1][3] == &setCandidates[16]) {
						int num = (op[1][4] - &setCandidates[0]);

						if (from > num || to < num) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							return;
						}
						else {
							from = to = num;
						}
							

					}
					else	
					/* a + a = \{b\}, a + b = \{c\} => b + c = (a + b) + b = a + (b + b) = c + (a + a) = a + (a + c) */
					if (op[1][1] == &setCandidates[4] && op[1][2] == &setCandidates[8]) {
						set<int> res_a_bb = hea::compute_sum(op, 1, *(op[2][2]));
						set<int> res_a_ac = hea::compute_sum(op, 1, *(op[1][3]));

						if (res_a_bb != res_a_ac) {
							
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							return ;
						}
						else {
							int num =  accumulate(res_a_bb.begin(), res_a_bb.end(), 0, sumpower2);
/*
							int num = 0;
							for (set<int>::const_iterator it_set = res_a_bb.begin(); it_set != res_a_bb.end(); ++it_set)
								num += (1 << *it_set);
*/
							from = to = num;
						}
					}	
					else
					/* a + a = \{b\} => b + c = (a + a) + c = a + (a + c)*/
					if (op[1][1] == &setCandidates[4]) {
						set<int> res_a_ac = hea::compute_sum(op, 1, *(op[1][3]));
						int num =  accumulate(res_a_ac.begin(), res_a_ac.end(), 0, sumpower2);

						if (from > num || to < num) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							return;
						}
						else {
							from = to = num;
						}
					}
					else
					/* a + a = \{0, b\} => a + (a + c) = (0 + c) union (b + c)  */

					/* a + a = \{a, b\} => a + (a + c) = (a + c) union (b + c)  */

					/* a + a = \{0, a, b\} => a + (a + c) = (0 + c) union (a + c) union (b + c)  */
					//if (op[1][1] == &setCandidates[5]) {
						res_a_ac = hea::compute_sum(op, 1, *(op[1][3]));
					//}
				}
				
#endif

				for (int i = from; i <= to; i++) {

					// skip x + y = {x}, {0, x}, {y}, {0, y}
					if (i == (1<<it_cand->candidate_pair.first) || i == 1 + (1<<it_cand->candidate_pair.first) || i == (1<<it_cand->candidate_pair.second) || i == 1 + (1<<it_cand->candidate_pair.second)) {

						if (!was_cleared) {
							clear_used_from(current_setup, level);
							was_cleared = true;
						}

						continue;
					}

#ifdef LINEAR7incomplete
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
						/* b \in a + a and a + a is subset of \{0, a, b\}*/
						if (i > 7 || i < 4) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}			
					else 
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
						/* c \in a + b, a + b subset of \{0, a, b, c\} */
						if (i > 15 || i < 8) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}
					else
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 3) {
						/* d \in a + c, a + c subset of \{0, a, b, c, d\} */
						if (i > 31 || i < 16) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}
					else
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 4) {
						/* e \in a + d */
						if (i < 32) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}
					else
					if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
						/* {0, a, b, c} not subset of b + b */
						if (i < 16) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}

					}
					else
					if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 3) {
						/* {0, a, b, c, d} not subset of b + c */
						if (i < 32) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}

					/* b + b */
					if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
						/* a + a = \{0, b\} (a + a) + b = (0 + b) union (b + b), where 
						   b + b = setCandidates[i] */
						if (op[1][1] == &setCandidates[5]) {
							set<int> res_aa_b;
				
							/*res_aa_b = */set_union(setCandidates[i].begin(), setCandidates[i].end(), op[0][2]->begin(), op[0][2]->end(), inserter(res_aa_b, res_aa_b.end()));

							if (res_aa_b != res_a_ab) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}

								continue;
							}
						}
						else
						/* a + a = \{a, b\} (a + a) + b = (a + b) union (b + b), where 
						   b + b = setCandidates[i] */
						if (op[1][1] == &setCandidates[6]) {
							set<int> res_aa_b;
				
							set_union(setCandidates[i].begin(), setCandidates[i].end(), op[1][2]->begin(), op[1][2]->end(), inserter(res_aa_b, res_aa_b.end()));

							if (res_aa_b != res_a_ab) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}

								continue;
							}
						}
						else
						/* a + a = \{0, a, b\} (a + a) + b = (0 + b) union (a + b) union (b + b), where 
						   b + b = setCandidates[i] */
						if (op[1][1] == &setCandidates[7]) {
							set<int> res_aa_b = setCandidates[i];
				
							res_aa_b.insert(op[0][2]->begin(), op[0][2]->end());
							res_aa_b.insert(op[1][2]->begin(), op[1][2]->end());

							if (res_aa_b != res_a_ab) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}

								continue;
							}
						}

					}
					else
					/* b + c */
					if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 3) {
						/* a + a = \{0, b\} => (a + a) + c = (0 + c) union (b + c), where
						   b + c = setCandidates[i] */
						if (op[1][1] == &setCandidates[5]) {
							set<int> res_aa_c;
				
							/*res_aa_c = */set_union(setCandidates[i].begin(), setCandidates[i].end(), op[0][3]->begin(), op[0][3]->end(), inserter(res_aa_c, res_aa_c.end()));

							if (res_aa_c != res_a_ac) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}

								continue;
							}
						}
						else
						/* a + a = \{a, b\} => (a + a) + c = (a + c) union (b + c), where
						   b + c = setCandidates[i] */
						if (op[1][1] == &setCandidates[6]) {
							set<int> res_aa_c;
				
							/*res_aa_c = */set_union(setCandidates[i].begin(), setCandidates[i].end(), op[1][3]->begin(), op[1][3]->end(), inserter(res_aa_c, res_aa_c.end()));

							if (res_aa_c != res_a_ac) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}

								continue;
							}
						}
						else
						/* a + a = \{0, a, b\} => (a + a) + c = (0 + c) union (a + c) union (b + c), where
						   b + c = setCandidates[i] */
						if (op[1][1] == &setCandidates[7]) {
							set<int> res_aa_c = setCandidates[i];
				
							res_aa_c.insert(op[0][3]->begin(), op[0][3]->end());
							res_aa_c.insert(op[1][3]->begin(), op[1][3]->end());

							if (res_aa_c != res_a_ac) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}

								continue;
							}
						}
					}
					

#endif

#ifdef LINEAR6
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
						/* b \in a + a and a + a is subset of \{0, a, b\}*/
						if (i > 7 || i < 4) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}			
					else 
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
						/* c \in a + b, a + b subset of \{0, a, b, c\} */
						if (i > 15 || i < 8) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}
					else
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 3) {
						if (i < 16) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}
					else
					if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
						if (i < 16) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}

#endif

					map<pair<int,int>, set<int> >::iterator itmap = not_subset_of_map.find(it_cand->candidate_pair);
					if (itmap != not_subset_of_map.end()) {
						bool is_OK = true;
						for (set<int>::const_iterator it_num_sets = itmap->second.begin(); it_num_sets != itmap->second.end(); ++it_num_sets)
							if ((i & *it_num_sets) == i) {
								is_OK = false;
								break;
							}

						if (!is_OK) {

							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}



					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

					if (!check_summability(candidates, it_cand->candidate_pair, op)) {

						if (!was_cleared) {
							clear_used_from(current_setup, level);
							was_cleared = true;
						}

						continue;
					}

					if (!is_associative_at_0(i, it_cand->candidate_pair.first, it_cand->candidate_pair.second, n, op, setCandidates)) {

							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
					} 




#if 0
					bool valid = true;
					if (last_cand_for_x.find(it_cand->candidate_pair) != last_cand_for_x.end()) {
						// x is summable with all elements in y + 0, if x + y is summable
						for (int i = it_cand->candidate_pair.first; i < n - 1 && valid; i++)
							if (hea::is_summable(op, it_cand->candidate_pair.first, i))
								for (set<int>::const_iterator it = op[0][i]->begin(); it != op[0][i]->end(); ++it)
									if (!hea::is_summable(op, it_cand->candidate_pair.first, *it)) {
										valid = false;
										break;
									}

					}
					
					if (!valid)
						continue;

#endif

					set<Candidate>::const_iterator it_new = it_cand;
					it_new++;
					solve_candidates_state_setup(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, current_setup,only0x, not_subset_of_map);

					// unset
					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
				}
			}

		}
	}
}



void init_last_cand_for_x_set(const int n, const set<Candidate> &candidates, set<pair<int, int> > &last_cand_for_x) {
	set<Candidate>::const_iterator it_next = candidates.begin();
	for (set<Candidate>::const_iterator it_cand = candidates.begin(); it_cand != candidates.end(); ++it_cand) {
		++it_next;
		
		if (it_next == candidates.end() ||
			it_cand->candidate_pair.first < it_next->candidate_pair.first)
			last_cand_for_x.insert(it_cand->candidate_pair);	

	}
}

void gen_rel_ith_state_setup(int x, int y, const int n, int (&relation)[MAX][MAX], vector<Relation> &relations, int ith, int &current_count, fromToState &current_setup) {
	if (x == n - 1) {

		if (admits_reverse_of_relation(n, relation)) {

			if (!is_isomorphic(n, relation, relations)) {

				/* disp relation */

/*
				for (int i = 0; i < n; i++) {
					for (int j = 0; j < n; j++)
						cerr << relation[i][j];
					cerr << endl;
				}
				cerr << "-R-"<<endl;
*/

				relations.push_back(Relation(n, relation));

				{

//					vector< pair< set< pair<int,int> >, set< pair<int,int> > > > HHHH;
					vector<Zoznamy> HHHH;
					vector<hea::Permutation> IIII;


					vector<hea::Permutation> vec_ortho;

					get_vector_of_orthosupplement_candidates(n, relation, vec_ortho);

					for (vector<hea::Permutation>::const_iterator it = vec_ortho.begin(); it != vec_ortho.end(); ++it) {

					set<pair<int, int> > one_pairs;

					for (int i = 0; i < n; i++) {
						one_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[i]));
						if (i != it->permutation[i])
							one_pairs.insert(pair<int, int>(it->permutation[i] + 1, 1 + i));

					}

/*
					if (one_pairs.size() == (unsigned int) n) {
*/

					set<pair<int, int> > other_pairs;
					
					for (int i = 0; i < n; i++)
					for (int j = 0; j < n; j++)
						if (i != j && relation[i][j]) {

							other_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[j]));
							if (i != it->permutation[j])

								other_pairs.insert(pair<int, int>(1 + it->permutation[j], i + 1));
						}




					if (!is_similar_vec_pairs(n, HHHH, one_pairs, other_pairs)) {
					
						IIII.push_back(hea::Permutation(*it));
						HHHH.push_back(Zoznamy(one_pairs, other_pairs));

					}

				}
/*
			cout << "IIII size: " << IIII.size() << endl;
			cout << "HHHH size: " << HHHH.size() << endl;
*/
			/* display candidates */
//			cerr << "----------------------------" << endl;
			vector<hea::Permutation>::const_iterator it_perm = IIII.begin();
			vector<Zoznamy>::const_iterator it_zozn = HHHH.begin();
			while (it_perm != IIII.end()) {
				current_count++;
				if (ith == current_count)
				{

					cerr << "Candidate: ";
					for (int i = 0; i < n; i++)
						cerr << 1 + it_perm->permutation[i]; 
					cerr << endl;

					cerr << "Candidates for sets containing 1: " ;


					for (set<pair<int, int> >::const_iterator it_set = it_zozn->one_pairs.begin(); it_set != it_zozn->one_pairs.end(); ++it_set) 
						cerr << "(" << it_set->first << ", " << it_set->second << "), ";

					cerr << endl;

					cerr << "Candidates for other summable elements: " ;
					for (set<pair<int, int> >::const_iterator it_set = it_zozn->other_pairs.begin(); it_set != it_zozn->other_pairs.end(); ++it_set) 
						cerr << "(" << it_set->first << ", " << it_set->second << "), ";

					cerr << endl;



					Zoznamy reduced;
					reduce_summable_pairs(*it_zozn, reduced);

					set<Candidate> candidates;

					for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
						candidates.insert(Candidate(*it_set, true));

					for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 

						candidates.insert(Candidate(*it_set, false));

					/* insert pairs (0, x) into candidates */
					for (int i = 1; i <=n; i++)
						candidates.insert(Candidate(pair<int,int>(0, i), false));
					candidates.insert(Candidate(pair<int, int>(0, n + 1), true));


					/* display candidates */

					cerr << "Candidates: ";

					for (set<Candidate>::const_iterator it_cand = candidates.begin(); it_cand != candidates.end(); ++it_cand) {
						
						cerr << "(" << it_cand->candidate_pair.first << ", " << it_cand->candidate_pair.second << ")";
						if (it_cand->is_one)
							cerr << "*";

						cerr << ", ";

					}
					cerr << endl;

				// only0x[i] is true, iff i + 0 is subset of \{0, i\}
				bool only0x[MAX];
				for (int i = 0; i <= n ; i++)
					only0x[i] = false;

				/* candidates for 0 + x */
				{
					for (int i = 0; i <= n - 1; i++) {
						cerr << "0 + " ;
						hea::disp_elem(cerr, n + 2, i + 1);
						cerr << " is subset of {";
						set<int> set0xcand;
						set0xcand.insert(0);

						for (int j = 0; j < n; j++)
							if (relation[j][i]) {
#ifdef DEBUG
								cerr << j + 1 << " <= " << i + 1 << endl;
#endif
								if (j != i) {
									/* if i + k exists then j + k must exist */
									bool is_OK = true;
									for (int k = 1; k <= n; k++)
										if (candidates.find(Candidate(i + 1, k)) != candidates.end()  && 
											candidates.find(Candidate(j + 1, k)) == candidates.end()) {
#ifdef DEBUG
											cerr << "Failed: " << i + 1 << " + " << k << " exists, but " << j + 1 << " + " << k << " don't." << endl;
#endif
											is_OK = false;
											break;
										}
									if (!is_OK) continue;
								}
								set0xcand.insert(j + 1);
							}


						hea::disp_set(cerr, n + 2, set0xcand);
						cerr << "}";
						cerr << endl;
						if (set0xcand.size() == 2)
							only0x[i + 1] = true;
					}
					for (int i = 1; i <= n; i++)
						cerr << (only0x[i]? "T":"F");
					cerr << endl;	

				}
/* is_summable array */
				
				int arr_is_summable[MAX][MAX];
				{
					for (int i = 0; i < n; i++)
					for (int j = 0; j < n; j++)
						arr_is_summable[i][j] = 0;

					for (set<pair<int, int> >::const_iterator it_set = it_zozn->one_pairs.begin(); it_set != it_zozn->one_pairs.end(); ++it_set) {
						arr_is_summable[it_set->first - 1][it_set->second - 1] = 1;
						arr_is_summable[it_set->second - 1][it_set->first - 1] = 1;
						
					}

					for (set<pair<int, int> >::const_iterator it_set = it_zozn->other_pairs.begin(); it_set != it_zozn->other_pairs.end(); ++it_set) {
						arr_is_summable[it_set->first - 1][it_set->second - 1] = 1;
						arr_is_summable[it_set->second - 1][it_set->first - 1] = 1;
						
					}
					cerr << "Is summable array: " << endl;
					for (int i = 0; i < n; i++) {
						for (int j = 0; j < n; j++)
							cerr << arr_is_summable[i][j];
						cerr << endl;
					}
				}

/* negative consequences of associativity */
				map<pair<int,int>, set<int> > not_subset_of_map;
				{
					  for (set<pair<int, int> >::const_iterator it_set = it_zozn->other_pairs.begin(); it_set != it_zozn->other_pairs.end(); ++it_set) {
						if (it_set->first <= it_set->second) {
							cerr << "(";
							hea::disp_elem(cerr, n + 2, it_set->first);
							cerr << ", ";
							hea::disp_elem(cerr, n + 2, it_set->second);
							cerr  << ") is not subset of: ";
#if 0
							if (it_set->first == it_set->second) {
								set<int> num_sets;
								for (int i = 0; i < n; i++)
									if (i == it_perm->permutation[it_set->first - 1] || !arr_is_summable[i][it_set->first - 1]) {
/*
										cout << "\t";
										hea::disp_elem(cout, n + 2, i + 1);
										cout << " is not summable with "; 
										hea::disp_elem(cout, n + 2, it_set->first); 
										cout << endl;
*/
										int elem_num_set = 1; /* 0 is in set */
										cout << "{0";
										for (int j = 0; j < n; j++)
											if (arr_is_summable[i][j]) {
												elem_num_set |= (1 << (j + 1));
												cout << ", ";
												hea::disp_elem(cout, n + 2, j + 1);						
											}
										num_sets.insert(elem_num_set);
										cout << "}, ";
										//cout << elem_num_set << endl;
									}
								not_subset_of_map.insert(pair<pair<int, int>, set<int> >(*it_set, num_sets));
								cout << endl;

							}
							else {
#endif
							set<int> num_sets;
							for (int i = 0; i < n; i++)
								if (i == it_perm->permutation[it_set->first - 1] || i == it_perm->permutation[it_set->second - 1] || !arr_is_summable[i][it_set->first - 1] || !arr_is_summable[i][it_set->second - 1]) {
/*
									cout << "\t";
									hea::disp_elem(cout, n + 2, i + 1);
									cout << " is not summable with "; 
									hea::disp_elem(cout, n + 2, it_set->first); 
									cout << endl;
*/
									int elem_num_set = 1; /* 0 is in set */
									cerr << "{0";
									for (int j = 0; j < n; j++)
										if (arr_is_summable[i][j]) {
											elem_num_set |= (1 << (j + 1));
											cerr << ", ";
											hea::disp_elem(cerr, n + 2, j + 1);						
										}
									num_sets.insert(elem_num_set);
									cerr << "}, ";
									//cout << elem_num_set << endl;
								}
							not_subset_of_map.insert(pair<pair<int, int>, set<int> >(*it_set, num_sets));
							cerr << endl;
							set<int> reduced_num_sets;
							for (set<int>::const_iterator it_num_sets = num_sets.begin(); it_num_sets != num_sets.end(); ++it_num_sets) {
								bool insert = true;
								set<int>::const_iterator it_num_sets_pom = it_num_sets;
								for (; it_num_sets_pom != num_sets.end(); ++it_num_sets_pom) {
									if (

										(it_num_sets != it_num_sets_pom && (*it_num_sets & *it_num_sets_pom) == *it_num_sets) || *it_num_sets == 1 + (1 << it_set->first) || *it_num_sets == 1 + (1 << it_set->second)) {
										insert = false;
										break;
									}
								}
								if (insert)
									reduced_num_sets.insert(*it_num_sets);
							}
							not_subset_of_map[*it_set] = reduced_num_sets;


							/* disp reduced */
							cerr << "Reduced list: ";
							for (set<int>::const_iterator it_reduced = reduced_num_sets.begin(); it_reduced != reduced_num_sets.end(); ++it_reduced) {
								cerr << "{0";
								for (int j = 1; j <=n; j++)
									if (*it_reduced & (1 << j)) {
										cerr << ", ";
										hea::disp_elem(cerr, n + 2, j);
									}
								cerr << "}, ";
								
							}	
							cerr << endl;

#if 0
							}	
#endif

//							cout << endl;					
					}
				  }
				
				}

					
					set<pair<int, int> > last_of_x;
					init_last_cand_for_x_set(n, candidates, last_of_x); 
					cerr << "Last of x pairs: " ;
					
					for (set<pair<int, int> >::const_iterator it_x = last_of_x.begin(); it_x != last_of_x.end(); ++it_x) {
						
						cerr << "(" << it_x->first << ", " << it_x->second << "), ";

					}
					cerr << endl;
		


/*
				cout << "Reduced candidates for sets containing 1: " ;


				for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;

				cout << "Reduced candidates for other summable elements: " ;
				for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;
*/

					//cerr << "----------------------------" << endl;

					cerr << "\t ***" << endl;
					set<int> setCandidates[1 << MAX];
					const set<int>* op[MAX][MAX];
					int complement[MAX];

					// init
					for (int i = 0; i < 1 << (n + 2); i++) {
						setCandidates[i] = set<int>();
						hea::init_set_i(setCandidates[i], i);
					}
					

			/*
					for (int i = 0 ; i < 1 << n; i++) {
						cout << "Set " << i << ": ";
						disp_set(setCandidates[i]);		
						cout << endl;
					}
			*/
					// just to have initialize op
					for (int i = 0; i < n + 2; i++)
					for (int j = 0; j < n + 2; j++)
						op[i][j] = &setCandidates[0];
					

					// 0 + 0  = {0}
					op[0][0] = &setCandidates[1];

					complement[0] = n + 1;
					complement[n + 1] = 0;

					/* init complement */
					for (set<Candidate>::const_iterator it_compl = candidates.begin(); it_compl != candidates.end(); ++it_compl)
						if (it_compl->is_one) {
							complement[it_compl->candidate_pair.first] = it_compl->candidate_pair.second;
							if (it_compl->candidate_pair.first != it_compl->candidate_pair.second)
								complement[it_compl->candidate_pair.second] = it_compl->candidate_pair.first;
						}
					/**/
				
					solve_candidates_state_setup(candidates.begin(), 0, candidates, last_of_x, n + 2, op, complement, setCandidates, current_setup, only0x, not_subset_of_map);
				}

				++it_perm;
				++it_zozn;
			}

			}


		}

		}
	}
	else {
		for (int i = 0; i < 3; i++) {
			switch (i) {
				case 0:
					relation[x][y] = 0;
					relation[y][x] = 0;
					break;
				case 1:
					relation[x][y] = 1;
					relation[y][x] = 0;
					break;
				case 2:
					relation[x][y] = 0;
					relation[y][x] = 1;
					break;
			}

			if (y < n - 1)
				gen_rel_ith_state_setup(x, y + 1, n, relation, relations, ith, current_count, current_setup);
			else
				gen_rel_ith_state_setup(x + 1, x + 2, n, relation, relations, ith, current_count, current_setup);
		}
	}

}




int main(int argc, char **argv) {

	fromToState current_state_setup;

	//bool bFromState = false;
	//bool bToState = false;
	if (argc < 3) {
		cout << "Usage: " << argv[0] << " <size> <n> [<from_state> [<to_state>]]" << endl
		     << "\t<size> - computes models of size <size> where " << MAX << " <= size < 2" << endl
		     << "\t<n> - computes models corresponding to the <n>-th model, (n = 1-...)" << endl
		     << "\t<from_state> - comma separated list of numbers" << endl 
		     << "\t<to_state> - comma separated list of numbers" << endl;

		cout << endl << "Note: If neceserry to generate larger models, increase the MAX constant in consts.h file." << endl;
		return 0;
	}

	const int n = atoi(argv[1]) - 2;

	if (n <= 0) {
		cout << "Size must be > 2." << endl;
		return 0;
	}

	if (n > MAX - 2) {
		cout << "Size must be <= " << MAX << "." << endl;
		return 0 ;
	}
	//int from_state[MAX*MAX];
	//bool used_from_state[MAX*MAX];

	//int to_state[MAX*MAX];
	//bool used_to_state[MAX*MAX];

	//for (int i = 0; i < MAX*MAX ; i++)
		//used_from_state[i] = false;

	//for (int i = 0; i < MAX*MAX; i++)
		//used_to_state[i] = false;

	if (argc >= 4) {
		current_state_setup.bFromState = true;
		char *pom = argv[3];
		int idx = 0;

		while (*pom) {
			int val = 0;
			while (*pom>='0' && *pom<='9') {
				val=10*val + (int)(*pom-'0');
				pom++;
			}
			if (*pom==',') 
				pom++;
			else 
				if (*pom != '\0') {
					cerr << "Invalid character '"<< *pom << "'." << endl;
					return -1;
				}
			
			current_state_setup.from_state[idx] = val;
			current_state_setup.used_from_state[idx] = true;
			idx++;
		}
	}

	if (argc == 5) {
		current_state_setup.bToState = true;
		char *pom = argv[4];
		int idx = 0;

		while (*pom) {
			int val = 0;
			while (*pom>='0' && *pom<='9') {
				val=10*val + (int)(*pom-'0');
				pom++;
			}
			if (*pom==',') 
				pom++;
			else 
				if (*pom != '\0') {
					cerr << "Invalid character '"<< *pom << "'." << endl;
					return -1;
				}
			
			current_state_setup.to_state[idx] = val;
			current_state_setup.used_to_state[idx] = true;
			idx++;
		}
	}
		

	int relation[MAX][MAX];
	vector<Relation> relations;

	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		relation[i][j] = 0;

	for (int i = 0; i < n; i++)
		relation[i][i] = 1;

	int current_count = 0;

	if (current_state_setup.bFromState) {
		cerr << "Computing " << argv[2] << "th model\n\t from state: " << endl;
		int idx = 0;
		while (idx < MAX*MAX && current_state_setup.used_from_state[idx]) {
			cerr << current_state_setup.from_state[idx] << ", ";
			idx++;
		}
		cerr << endl;
		
		if (current_state_setup.bToState) {
			cerr << "\t to state: " << endl;
			idx = 0;
			while (idx < MAX*MAX && current_state_setup.used_to_state[idx]) {
				cerr << current_state_setup.to_state[idx] << ", ";
				idx++;
			}
			cerr << endl;

			//gen_rel_ith_from_state_to_state(0, 1, n, relation, relations, atoi(argv[1]), current_count, from_state, used_from_state, to_state, used_to_state);

		}
		//else 
			//gen_rel_ith_from_state(0, 1, n, relation, relations, atoi(argv[1]), current_count, from_state, used_from_state);

	}
	//else
		//gen_rel_ith(0, 1, n, relation, relations, atoi(argv[1]), current_count);

	time_t time_start;
	time_t time_end;
	time(&time_start);
	gen_rel_ith_state_setup(0, 1, n, relation, relations, atoi(argv[2]), current_count, current_state_setup);
	time(&time_end);

	cerr << "Number of cycles: " << number_of_cycles << endl;
	cerr << "Last possibilities: " << number_of_tested_configurations << endl;
	cerr << "Start time: " << time_start << endl;
	cerr << "End   time: " << time_end << endl;

/*
	map<pair<int, int>, bool> is_last_of_x;
	is_last_of_x.insert(pair<pair<int, int>, bool>(pair<int, int>(1,2), true));
*/

}


