#include <iomanip>
#include <string.h>
#include <stdlib.h>
#include <fstream>
#include <vector>

#include "hyperEA3_common.h"
#include "hea.h"
#include "hyperEA3_parse.h"
#include <set>
#include <algorithm>
	

using namespace std;



struct Permutation {
	int n;
	int permutation[MAX];

	Permutation(const int _n, const int (&_permutation)[MAX]);
};

Permutation::Permutation(const int _n, const int (&_permutation)[MAX]) {
	n = _n;
	for (int i = 0; i < n; i++)
		permutation[i] = _permutation[i];	
}




void disp_set(const int n, const set<int> &ds) {
/*
	for (set<int>::iterator it = ds.begin(); it != ds.end(); ++it) {
		cout << *it << ", ";
	}
//	cout << endl;
*/

	set<int>::iterator it = ds.begin();
	if (it != ds.end()) {
		//cout << *it;
		hea::disp_elem(cout, n, *it);
		++it;
		do {
			if (it != ds.end()) {
				cout << ", "; 
				hea::disp_elem(cout, n, *it);
				++it;
			}
		} while (it != ds.end());
	}

}

void generate_all_order_preserving_permutations(int level, const hea::Hea &hea, int (&permutation)[MAX], vector<Permutation> &vec_order_preserve) {

	if (level == hea.n - 2) {
		
		for (int i = 1; i < hea.n - 1; i++)
		for (int j = 1; j < hea.n - 1; j++) {
			if (hea::is_summable(hea.op, i, hea.complement[j]) && !hea::is_summable(hea.op, permutation[i], hea.complement[permutation[j]]))
				return ;
			if (hea::is_summable(hea.op, permutation[i], hea.complement[permutation[j]]) && !hea::is_summable(hea.op, i, hea.complement[j]))
				return ;
		}

		vec_order_preserve.push_back(Permutation(hea.n, permutation));


	}
	else {
		for (int i = level; i < hea.n - 1; i++) {
			//swap level i
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
			generate_all_order_preserving_permutations( level + 1, hea, permutation, vec_order_preserve);
			//swap level i
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
		}
		
	}

}

void getComplement(int n, const set<int>* (&op)[MAX][MAX], int (&complement)[MAX]) {
	complement[0] = n - 1;
	complement[n - 1] = 0;

	for (int i = 1; i < n - 1; i++) {
		for (int  k = 1; k < n - 1; k++) {
			if (op[i][k]->find(n - 1) != op[i][k]->end()) {
				complement[i] = k;
				break;
			}
		}
				
	}

}

void getOrder(int n, const set<int>* (&op)[MAX][MAX], int (&complement)[MAX], int (&order)[MAX][MAX]) {

	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		order[i][j] = 0;

	for (int i = 0; i < n; i++) {
		order[0][i] = 1;
		order[i][n-1] = 1;
		order[i][i] = 1;
	}
	
	// ``order''
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if (!op[i][complement[j]]->empty())
			order[i][j] = 1;

}


bool isIdeal(int n, const set<int>* (&op)[MAX][MAX], int (&complement)[MAX], int (&order)[MAX][MAX], const set<int> &ideal) {
	// non-empty subset
	if (ideal.empty())
		return false;
	
	// 0\in ideal
	if (ideal.find(0) == ideal.end())
		return false;

/*
	int order[MAX][MAX];
	int complement[MAX];
	getComplement(n, op, complement);
	getOrder(n, op, complement,  order);
*/
	// x\in I, y\leq x => y \in I
	for (set<int>::const_iterator it = ideal.begin(); it != ideal.end(); ++it) {
		for (int i = 0; i < n; i++)
			if (order[i][*it] && ideal.find(i) == ideal.end()) 	
				return false;
	}

	// x, y \in I => x + y \subseteq I 

	for (set<int>::const_iterator it_x = ideal.begin(); it_x != ideal.end(); ++it_x) 
		for (set<int>::const_iterator it_y = ideal.begin(); it_y != ideal.end(); ++it_y) {

			if (!op[*it_x][*it_y]->empty()) {
				set<int> result;

				set_intersection(op[*it_x][*it_y]->begin(), op[*it_x][*it_y]->end(), ideal.begin(), ideal.end(), inserter(result, result.begin()));
				
				if (*(op[*it_x][*it_y]) != result)
					return false;
			}
		}
		

	return true;
}


void generate_all_ideals(int n, const set<int>* (&op)[MAX][MAX]) {

	hea::disp_op(cout, n, op);
	for (int i = 1; i < (1 << n) ; i++) {
		set<int> ideal;

		hea::init_set_i(ideal, i);

#if 0
		cout << endl << "Ideal candidate: " ;
		for (set<int>::const_iterator it = ideal.begin(); it != ideal.end(); ++it)
			cout << *it;
		cout << endl;
#endif
		int complement[MAX];
		int order[MAX][MAX];
		getComplement(n, op, complement);
#if 0
		for (int i = 0; i < n; i++)
			cout << complement[i];
		cout << endl;
#endif

		getOrder(n, op, complement, order);
		if (isIdeal(n, op, complement, order, ideal)) {
			cout << "\tIdeal: {";
			bool first = true;
			for (set<int>::const_iterator it = ideal.begin(); it != ideal.end(); ++it) {
				if (!first)
					cout << ", ";
				else {
					first = false;
				}

				hea::disp_elem(cout, n, *it);
			}
			cout << "}\n";
		}
	
	}
}

int set_to_int(const set<int> *mset) {
	int result = 0;

	for (set<int>::const_iterator it = mset->begin(); it != mset->end(); ++it)
		result |= (1 << *it);

	return result;
}

void compute_h_set(int n, const set<int>* (&op)[MAX][MAX], int (&complement)[MAX]) {
	set<int> set_of_outcomes_int;
	int size, old_size;

	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if (hea::is_summable(op, i, j))
			set_of_outcomes_int.insert(set_to_int(op[i][j]));	

	size = set_of_outcomes_int.size();

	do {
		cout << "Size: " << size << endl;
		old_size = size;	
		set<int> set_of_outcomes_pom = set_of_outcomes_int;

		for (set<int>::const_iterator it1 = set_of_outcomes_pom.begin(); it1 != set_of_outcomes_pom.end(); ++it1)
		for (set<int>::const_iterator it2 = set_of_outcomes_pom.begin(); it2 != set_of_outcomes_pom.end(); ++it2) {
			set<int> arg1, arg2;
			hea::init_set_i(arg1, *it1);
			hea::init_set_i(arg2, *it2);

			if (hea::is_summable(op, arg1, arg2)) {
				set<int> result = hea::compute_sum(op, arg1, arg2);

				set_of_outcomes_int.insert(set_to_int(&result));
			}
		
		}


		size = set_of_outcomes_int.size();
	}	
	while (old_size != size);

	cout << "\t--" << endl;

}

/* H_0 = { {0} }, H_{i + 1} = { x + y; x in H_i, y in E } */
void compute_h_set2(int n, const set<int>* (&op)[MAX][MAX], int (&complement)[MAX]) {
	set<int> set_of_outcomes_int;
	int size, old_size;

	set_of_outcomes_int.insert(1); /* H_0 = \{0\} */

	size = set_of_outcomes_int.size();

	do {
		cout << "Size: " << size << endl;
		old_size = size;	
		set<int> set_of_outcomes_pom = set_of_outcomes_int;

		for (set<int>::const_iterator it1 = set_of_outcomes_pom.begin(); it1 != set_of_outcomes_pom.end(); ++it1)
			for (int i = 0; i < n; i++) {
				set<int> arg1, arg2;
				hea::init_set_i(arg1, *it1);
				hea::init_set_i(arg2, 1 << i);

				if (hea::is_summable(op, arg1, arg2)) {
					set<int> result = hea::compute_sum(op, arg1, arg2);

					set_of_outcomes_int.insert(set_to_int(&result));
				}
			
			}


		size = set_of_outcomes_int.size();
	}	
	while (old_size != size);
	
	cout << "H+(E) = {";
	for (set<int>::const_iterator it=set_of_outcomes_int.begin(); it != set_of_outcomes_int.end(); ++it) {
		cout << " {";
		set<int> result;
  		hea::init_set_i(result, *it);
		hea::disp_set(cout, n, result);
		cout << "},";
	}
	cout << "}" << endl;

	cout << "\t--" << endl;

}



int main(int argc, char **argv) {

	const int n = 5;

	if (argc < 2) {
		cout << "Usage: " << argv[0] << " filename" << endl;
		return 0;
	
	}

	set<int> setCandidates[1 << MAX];
	const set<int>* op[MAX][MAX];

	// init
	for (int i = 0; i < 1 << n; i++) {
		setCandidates[i] = set<int>();
		hea::init_set_i(setCandidates[i], i);
	}
	
	int complement[MAX];
	{
		ifstream ifs;

		ifs.open(argv[1], ifstream::in);
		
		if (ifs.good()) {
			cerr << "Loading file " << argv[1] << "." << endl;

			while(hea::parse_model(ifs, n, op, complement, setCandidates)) {

				hea::Hea myhea(n, op, complement);
				hea::disp_hea(cout, myhea);
				compute_h_set2(n, op, complement);
				

				// skip one line containing -M-
				char buffer[1024];
				ifs.getline(buffer, 1024);
			}
				
		}
		else {
			cerr << "Cannot open or read from file '" << argv[1] << "'." << endl;
		}
		ifs.close();
	}

	return 0;
}
