#include <iostream>
#include <set>
#include <vector>
#include <map>
#include <stdlib.h>
#include <ctime>

#include "hyperEA3_common.h"
#include "permutation.h"
#include "candidate.h"

//#define MAX	6

using namespace std;


struct fromToState {
	bool bFromState;
	int from_state[MAX*MAX];
	bool used_from_state[MAX*MAX];	

	bool bToState;
	int to_state[MAX*MAX];
	bool used_to_state[MAX*MAX];	
	
	fromToState() {
		bFromState = false;
		bToState = false;
		for (int i = 0; i < MAX*MAX; i++) {
			from_state[i] = 0;
			to_state[i] = 0;
			used_from_state[i] = false;
			used_to_state[i] = false;
		}
				
	}
};

struct Relation {
	int n;
	int relation[MAX][MAX];

	Relation(const int _n, const int (&_relation)[MAX][MAX]) {
		n = _n;

		for (int i = 0; i < _n; i++)	
		for (int j = 0; j < _n; j++)	

			relation[i][j] = _relation[i][j];
	}
};

struct Zoznamy {
	set<pair<int, int> > one_pairs;
	set<pair<int, int> > other_pairs;

	Zoznamy() {
	}

	Zoznamy( const set<pair<int, int> > &_one_pairs, const set<pair<int, int> > &_other_pairs ) {
		for (set<pair<int, int> >::const_iterator it = _one_pairs.begin(); it != _one_pairs.end(); ++it) {
			one_pairs.insert(*it);
		}

		for (set<pair<int, int> >::const_iterator it = _other_pairs.begin(); it != _other_pairs.end(); ++it) {
			other_pairs.insert(*it);
		}
	}
};



bool are_rel_isomorphic_via_perm(const int n, const int (&permutation)[MAX], const int (&relation)[MAX][MAX], const Relation &relation2) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if (relation[i][j] != relation2.relation[ permutation[i] ][ permutation[j] ])
			return false;

	return true;

}

void are_isomorphic_recursive(int level, int (&permutation)[MAX], const int n, const int (&relation)[MAX][MAX], const Relation &relation2, bool &are_iso) {
	if (!are_iso) {
		if (level == n - 1) {

			if (are_rel_isomorphic_via_perm(n, permutation, relation, relation2))
				are_iso = true;

		}
		else {
			for (int i=level; i < n; i++) {
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
				
				are_isomorphic_recursive(level + 1, permutation, n, relation, relation2, are_iso);
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}

			}
		}
	}
}

unsigned long int number_of_tested_configurations=0UL;
const long int dump_state_count=10000000UL;

int compute_set_value(int n, const set<int> *setc) {
	int val = 0;	

	for (set<int>::const_iterator itset = setc->begin(); itset != setc->end(); ++itset)
		val+= (1 << *itset);
	
	return val;
}

bool are_isomorphic(const int n, const int (&relation)[MAX][MAX], const Relation &relation2) {
	bool are_iso = false;
	int permutation[MAX];
	for (int i = 0; i < n; i++)
		permutation[i] = i;

	are_isomorphic_recursive(0, permutation, n, relation, relation2, are_iso);

	return are_iso;
}

bool is_isomorphic(const int n, const int (&relation)[MAX][MAX], const vector<Relation> &relations) {
	/* first relation */
	if (relations.empty())
		return false;


	for (vector<Relation>::const_iterator it = relations.begin(); it != relations.end(); ++it) {
		if (are_isomorphic(n, relation, *it))
			return true;
	}

	return false;
}

bool is_double_negation_permutation(const int n, const int (&permutation)[MAX]) {
	for (int i = 0; i < n; i++)
		if (i != permutation [ permutation [ i ] ])
			return false;

	return true;
}



void admits_reverse_of_relation_recursive(int level, int (&permutation)[MAX], const int n, const int (&relation)[MAX][MAX], bool &found) {
	if (!found) {
		if (level == n - 1) {
			bool is_valid  = true;
			for (int i = 0; i < n && is_valid; i++)
			for (int j = 0; j < n && is_valid; j++)
				if (i != j)
					if (relation[i][j] && !relation[permutation[j]][permutation[i]]) {
						/* wrong permutation */
						is_valid = false;
						break;
					}

			if (is_valid) {
				found = true;


#if 0
				/* test permutation */	
				for (int i = 0; i < n; i++)
					cout << permutation[i] << " ";
				cout << endl;
				cout << "-P-" << endl;
#endif
			}
		}
		else {
			for (int i=level; i < n; i++) {
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
				
				admits_reverse_of_relation_recursive(level + 1, permutation, n, relation, found);
				if (i != level) {
					int tmp = permutation[level];
					permutation[level] = permutation[i];
					permutation[i] = tmp;
				}
					
			}
				
		}
	}
}


bool admits_reverse_of_relation(const int n, const int (&relation)[MAX][MAX]) {
	
	bool found = false;
	int permutation[MAX];
	for (int i = 0; i < n; i++)
		permutation[i] = i;

	admits_reverse_of_relation_recursive(0, permutation, n, relation, found); 

	return found;
	
}

void get_vector_of_orthosupplement_candidates(int level, int (&permutation)[MAX], const int n, const int (&relation)[MAX][MAX], vector<hea::Permutation> &vector_ortho) {

	if (level == n - 1) {
		bool is_valid  = true;
		if (!is_double_negation_permutation(n, permutation))
			is_valid = false;

		for (int i = 0; i < n && is_valid; i++)
		for (int j = 0; j < n && is_valid; j++)
			if (i != j)
				if (relation[i][j] && !relation[permutation[j]][permutation[i]]) {
					/* wrong permutation */
					is_valid = false;
					break;
				}

		if (is_valid) {
			vector_ortho.push_back(hea::Permutation(n, permutation));


#if 0
			/* test permutation */	
			for (int i = 0; i < n; i++)
				cout << permutation[i] << " ";
			cout << endl;
			cout << "-P-" << endl;
#endif
		}
	}
	else {
		for (int i=level; i < n; i++) {
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
			
			get_vector_of_orthosupplement_candidates(level + 1, permutation, n, relation, vector_ortho);
			if (i != level) {
				int tmp = permutation[level];
				permutation[level] = permutation[i];
				permutation[i] = tmp;
			}
				
		}
			
	}
}

/* mapping x-> x', such that x<=y, iff y'<=x' 
 *  returns true, if there is at least one such mapping
 */
bool get_vector_of_orthosupplement_candidates(const int n, const int (&relation)[MAX][MAX], vector<hea::Permutation> &vector_ortho) {
	
	int permutation[MAX];
	for (int i = 0; i < n; i++)
		permutation[i] = i;

	get_vector_of_orthosupplement_candidates(0, permutation, n, relation, vector_ortho); 

	return !vector_ortho.empty();
}


bool is_similar_pairs_via_perm(const int n, const Zoznamy &zoznamy, const set<pair<int, int> > &one_pairs, const set<pair<int, int> > &other_pairs, const int (&permutation)[MAX]) {
	//set<pair<int, int> > one_iso_pairs, other_iso_pairs;

	for (set<pair<int, int> >::const_iterator it_set = one_pairs.begin(); it_set != one_pairs.end(); ++it_set) {
		pair<int, int> iso_pair(permutation[it_set->first - 1] + 1, permutation[it_set->second - 1] + 1);
		if (zoznamy.one_pairs.find(iso_pair) == zoznamy.one_pairs.end())
			/* not found */
			return false;


	//	one_iso_pairs.insert(pair<int, int>(permutation[it_set->first], permutation[it_set->second]));	

	}

	for (set<pair<int, int> >::const_iterator it_set = other_pairs.begin(); it_set != other_pairs.end(); ++it_set) {
		pair<int, int> iso_pair(permutation[it_set->first - 1] + 1, permutation[it_set->second - 1] + 1);
		if (zoznamy.other_pairs.find(iso_pair) == zoznamy.other_pairs.end())
			/* not found */
			return false;


	//	one_iso_pairs.insert(pair<int, int>(permutation[it_set->first], permutation[it_set->second]));	

	}

	return true;
}

/* is  zoznamy isomorphic to one_pairs, other_pairs*/
void is_similar_pairs_recursive(int level, const int n, const Zoznamy &zoznamy, const set<pair<int, int> > &one_pairs, const set<pair<int, int> > &other_pairs, int (&permutation)[MAX], bool &is_similar) {
	if (!is_similar) {
		if (level == n - 1) {
/*
			for (int i = 0; i < n; i++)
				cout << permutation[i] << " ";
			cout << endl;
*/

			if (is_similar_pairs_via_perm(n, zoznamy, one_pairs, other_pairs, permutation)) {
/*
				cout << "Permutation: ";
				for (int i = 0; i < n; i++)
					cout << permutation[i] ;
				cout << endl;
*/
				is_similar = true;
			}

		}
		else {

			for (int i = level; i < n; i++) {
				if (i!=level) {
					int tmp = permutation[i];
					permutation[i] = permutation[level];
					permutation[level] = tmp;
				}

				is_similar_pairs_recursive(level + 1, n, zoznamy, one_pairs, other_pairs, permutation, is_similar);

				if (i!=level) {
					int tmp = permutation[i];
					permutation[i] = permutation[level];
					permutation[level] = tmp;
				}

			}

		}

	}


} 

/* is  zoznamy isomorphic to one_pairs, other_pairs*/
bool is_similar_pairs(const int n, const Zoznamy &zoznamy, const set<pair<int, int> > &one_pairs, const set<pair<int, int> > &other_pairs) {
	int permutation[MAX];
	for (int i = 0; i < n; i++)
		permutation[i] = i;

	bool is_similar = false;

	is_similar_pairs_recursive(0, n, zoznamy, one_pairs, other_pairs, permutation, is_similar);

	return is_similar;
} 

/* search for vec_zoznamy wheater it contains isomorphic (one, other) pairs to one_pairs and other_pairs */
bool is_similar_vec_pairs(const int n, const vector<Zoznamy> &vec_zoznamy, const set<pair<int, int> > &one_pairs, const set<pair<int, int> > &other_pairs) {
	for (vector<Zoznamy>::const_iterator it_vec = vec_zoznamy.begin(); it_vec != vec_zoznamy.end(); ++it_vec) {

		if (is_similar_pairs(n, *it_vec, one_pairs, other_pairs))
			return true;
		
	}

	return false;

} 

/* reduces summable pairs with respect to commutativity property 
 * i.e. it removes (i, j), (j, i) duplicates and replaces it with 
 * just (i, j), i<=j.
 */
void reduce_summable_pairs(const Zoznamy &zoznamy, Zoznamy &reduced) {
	for (set<pair<int, int> >::const_iterator it_set = zoznamy.one_pairs.begin(); it_set != zoznamy.one_pairs.end(); ++it_set)
		if (it_set->first <= it_set->second)
			reduced.one_pairs.insert(*it_set);	

	for (set<pair<int, int> >::const_iterator it_set = zoznamy.other_pairs.begin(); it_set != zoznamy.other_pairs.end(); ++it_set)
		if (it_set->first <= it_set->second)
			reduced.other_pairs.insert(*it_set);	

}

#if 0
struct Candidate {
	pair<int, int> candidate_pair;
	bool	is_one; // true one_pairs, false other_pairs

	Candidate(const pair<int, int> &_candidate_pair, bool _is_one) {
		is_one = _is_one;
		candidate_pair.first = _candidate_pair.first;
		candidate_pair.second = _candidate_pair.second;
	}

	bool operator<(const Candidate &cand2) const {
		return (candidate_pair.first < cand2.candidate_pair.first || 
				(candidate_pair.first == cand2.candidate_pair.first && candidate_pair.second < cand2.candidate_pair.second));
	}
};


// searches candidates for a pair (a,b) a<=b, resp. (b,a) b< a
bool is_summable_candidate(const set<Candidate> &candidates, int a, int b) {
	pair<int, int> my_pair;
	
	if (a <= b) {
		my_pair.first = a;
		my_pair.second = b;
	}
	else {
		my_pair.first = b;
		my_pair.second = a;
	}

	return ( candidates.find(Candidate(my_pair, false)) != candidates.end() );
}


// pair = (x,y) is a canditate pair, i.e. x + y is defined
// then 
//	x must be summable with all elements of y + 0 = 0 + y
// 		and
//	y must be summable with all elements of 0 + x
//
// ! asserts that all pairs (0,x) are already defined !

bool check_summability(const set<Candidate> &candidates, const pair<int, int> &test_pair, const set<int> *(&op)[MAX][MAX]) {

	// x is summable with all elements in y + 0, if x + y is summable
	for (set<int>::const_iterator it_set = op[0][test_pair.second]->begin(); it_set != op[0][test_pair.second]->end(); ++it_set) {
		if (*it_set == 0)
			continue;

		if (!is_summable_candidate(candidates, test_pair.first, *it_set))
			return false;
	}

	if (test_pair.first != test_pair.second) {
		// y is summable with all elements in x + 0, if x + y is summable
		for (set<int>::const_iterator it_set = op[0][test_pair.first]->begin(); it_set != op[0][test_pair.first]->end(); ++it_set) {
			if (*it_set == 0)
				continue;

			if (!is_summable_candidate(candidates, test_pair.second, *it_set))
				return false;
		}
	}

	return true;
}
#endif

#if 0
/* For the case where 0 + x subseteq \{0, x\}, checks wheater candidate for x + x meets desired conditions */
bool testCandidateSetForXPlusXContainingOne(int i, int x, const int n, const set<int> *(&op)[MAX][MAX], set<int> *setCandidates) {
	/* x + x */
	/* x + 0 = \{x\} 
		x + x = A
			A + 0 = A
			1 + 0 \subseteq A + 0

	*/
	if (op[0][x] == &setCandidates[1<<x]) {
		set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
	
		if (resA0 != setCandidates[i]) 
			return false;	
		
		
		int idx01 = op[0][n - 1] - &setCandidates[0];
		if ((idx01 & i) != idx01) 
			return false;	
	}
	else
	/* x + 0 = \{0, x\} 
		x + x = A
			A + 0 = A \union \{0, x\}
			1 + 0 \subseteq A + 0
	*/
	if (op[0][x] == &setCandidates[(1<<x) + 1]) {
		set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		set<int> resA0x = setCandidates[i];
		resA0x.insert(0);
		resA0x.insert(x);
		
		if (resA0 != resA0x) 
			return false;						

		int idx01 = op[0][n - 1] - &setCandidates[0];
		int idxA0 = i | 1;
		idxA0 |= (1 << x); 
		if ((idx01 & idxA0) != idx01) 
			return false;
	}
			
	
	return true;

}

/* For the case where 0 + x subseteq \{0, x\} or 0 + y subseteq \{0, y\} checks wheater candidate for x + x meets desired conditions */
bool testCandidateSetForXPlusYJustOneContainingOne(int i, int x, bool only0x, int y, const int n, const set<int> *(&op)[MAX][MAX], set<int> *setCandidates) {

	int a, b;

	if (only0x) {
		a = x;
		b = y;
	}
	else {
		a = y;
		b = x;
	}

	if (op[0][a] == &setCandidates[1<<a]) {
		set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		if (resA0 != setCandidates[i]) 
			return false;
		
		int idx01 = op[0][n - 1] - &setCandidates[0];
		if ((idx01 & i) != idx01) 
			return false;
	}
	else
	if (op[0][a] == &setCandidates[1+(1<<a)]) {
		set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		set<int> resA_0b = setCandidates[i];
		for (set<int>::const_iterator it_0b=op[0][b]->begin(); it_0b != op[0][b]->end(); ++it_0b)
			resA_0b.insert(*it_0b);
		
		if (resA0 != resA_0b) 
			return false;

		int idx01 = op[0][n - 1] - &setCandidates[0];
		int idxA0 = 0;
		for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
			idxA0 += (1 << (*itA0));
		if ((idx01 & idxA0) != idx01) 
			return false;
	}

	return true;

}

/* For the case where 0 + x subseteq \{0, x\} and 0 + y subseteq \{0, y\} checks wheater candidate for x + x meets desired conditions */
bool testCandidateSetForXPlusYBothContainingOne(int i, int x, int y, const int n, const set<int> *(&op)[MAX][MAX], set<int> *setCandidates) {

	/* x + y */
	if (op[0][x] == &setCandidates[1<<x] && op[0][y] == &setCandidates[1<<y]) {
		/* x + 0 = \{x\} && y + 0 = \{y\} => (x + y) + 0 = (x + y) */
		set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		if (resA0 != setCandidates[i]) 
			return false;
		
		
		int idx01 = op[0][n - 1] - &setCandidates[0];
		if ((idx01 & i) != idx01) 		
			return false;
	}
	else 
	if (op[0][x] == &setCandidates[1<<x] && op[0][y] == &setCandidates[1 + (1<<y)]) {
		/* x + 0 = \{x\} && y + 0 = \{0, y\} => (x + y) + 0 = x + y = (x + y) union \{x\} */
		set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		if (resA0 != setCandidates[i]) 
			return false;

		set<int> resAx = setCandidates[i]; /* (x + y) union \{x\} */
		resAx.insert(x);
		if (resA0 != resAx) 
			return false;

		int idx01 = op[0][n - 1] - &setCandidates[0];
		if ((idx01 & i) != idx01) 
			return false;
	}
	else 
	if (op[0][x] == &setCandidates[1 + (1<<x)] && op[0][y] == &setCandidates[1<<y]) {
		/* x + 0 = \{0, x\} && y + 0 = \{y\} => (x + y) + 0 = x + y =  (x + y) union \{y\} */
		set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		if (resA0 != setCandidates[i]) 
			return false;

		set<int> resAy = setCandidates[i]; /* (x + y) union \{y\} */
		resAy.insert(y);
		if (resA0 != resAy) 
			return false;

		int idx01 = op[0][n - 1] - &setCandidates[0];
		if ((idx01 & i) != idx01) 
			return false;
	}
	else 
	if (op[0][x] == &setCandidates[1 + (1<<x)] && op[0][y] == &setCandidates[1 + (1<<y)]) {
		/* x + 0 = \{0, x\} && y + 0 = \{0, y\} => (x + y) + 0 = (x + y) union \{0, x\} = (x + y) union \{0, y\} */
		set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
		set<int> resA0x = setCandidates[i]; /* (x + y) union \{0, x\} */
		resA0x.insert(0);
		resA0x.insert(x);

		if (resA0 != resA0x) 
			return false;

		set<int> resA0y = setCandidates[i]; /* (x + y) union \{0, y\} */
		resA0y.insert(0);
		resA0y.insert(y);

		if (resA0 != resA0y) 
			return false;

		int idx01 = op[0][n - 1] - &setCandidates[0];
		int idxA0 = i | 1;
		idxA0 |= (1<<x); // union 0, x should be same as union 0, y

		if ((idx01 & idxA0) != idx01) 
			return false;
	}

	return true;

}
#endif


void solve_candidates(set<Candidate>::const_iterator it_cand, const set<Candidate> &candidates, const set<pair<int, int> > &last_cand_for_x, const int n, const set<int> *(&op)[MAX][MAX], const int (&complement)[MAX], set<int> *setCandidates) {

	if (it_cand == candidates.end()) {


		number_of_tested_configurations++;
		if (number_of_tested_configurations == dump_state_count) {
			number_of_tested_configurations = 0UL;

			/* dump state */

			cerr << "State: ";
			for (set<Candidate>::const_iterator itcan = candidates.begin(); itcan != candidates.end(); ++itcan) {

				int val = compute_set_value(n, op[itcan->candidate_pair.first][itcan->candidate_pair.second]);
				cerr << val << "," ; 
			}
			cerr << endl;
		}

/*
				for (int i =0 ; i < n; i++)
					cout << complement[i] << "  ";
				cout << endl;
*/


		if (hea::is_associative(n, op) /*&& has_unique_orthosupplement(n, op)*/) {
			/*int complement[MAX];*/
			/*for (int i = 0; i < n; i++)
				complement[n-1-i] = i;*/

			/*compute_complement(n, op, complement);*/

			if (hea::is_antisymmetric(n, op, complement) /*&& 
				(!transitive || hea::is_transitive(n, op, complement))*/
			) 
			{
				/* display complement */
				for (int i =0 ; i < n; i++) {
					hea::disp_elem(cout, n, i);
					cout << "  ";
				}
				cout << endl;

				for (int i =0 ; i < n; i++) {
					hea::disp_elem(cout, n, complement[i]);
					cout << "  ";
				}
				cout << endl;

				hea::disp_op(cout, n, op);
				cout << "\t-M-" << endl;
			}
		}

		//disp_op(n, op);
	}
	else {
		if (it_cand->candidate_pair.first == 0) {
			if (it_cand->is_one) {
				/* contains 1 == n - 1 */

				for (int i = (1 << (n - 1)); i < 1 << n; i++)

					/* for t = y - 1, y - 2, ... 0, if candidate for [0][y] contains t
					 * then it needs to contain all elements of [0][t] 
					 */

					if (hea::test_subsets_new(n, i, op, it_cand->candidate_pair.second, setCandidates)) {


						/* b \in 0 + y => b + y' are defined */
						bool correct = true;
						for (set<int>::const_iterator it_summable = setCandidates[i].begin(); it_summable != setCandidates[i].end(); ++it_summable) {
							// skip trivial checks 0, y
							if (*it_summable != 0 && *it_summable != it_cand->candidate_pair.second) {
								pair<int, int> my_pair;
								int tmp_compl = complement[ it_cand->candidate_pair.second ];

								if (*it_summable < tmp_compl) {
									my_pair.first = *it_summable;
									my_pair.second = tmp_compl;
								}
								else {
									my_pair.first = tmp_compl;
									my_pair.second = *it_summable;
								}
								
								if (candidates.find(Candidate(my_pair, false)) == candidates.end()) {
									correct = false;	
									break;
								}
										
							}
} 
						if (correct) {

							op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
							op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

							set<Candidate>::const_iterator it_new = it_cand;
							it_new++;
							solve_candidates(it_new, candidates, last_cand_for_x, n, op, complement, setCandidates);		

							// unset filled elements - empty set
							op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
							op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
						}
					}
			}
			else {
				/* does not contain one */
				/* it must be defined - excludes i = 0 */
				/* {0} cannot be used any more  - excludes i = 1 */
				for (int i = 2; i < (1 << (n - 1)); i++)
					/* x\in 0 + x */
					if (setCandidates[i].find(it_cand->candidate_pair.second) != setCandidates[i].end()) {

						// was the candidate already used? 0+x = 0+y => x= y
						bool was_used = false;

						for (int k = it_cand->candidate_pair.second - 1; k >= 1; k--) {
							if (op[0][k] == &setCandidates[i]) {
								was_used = true;	
								break;
							}
						}
						
						if (was_used)
							continue;

						//--

						// Candidate for 0+b: b<=c, then (0+b) is summable with c'
						int b = it_cand->candidate_pair.second;
						bool is_satisfied = true;

						// 1==n-1 cannot be contained and 0 is trivial, skipping
						for (int ii = 1; is_satisfied && ii < n - 1; ii++) {
							// skipping also b as trivial
							if (ii == b)
								continue;

							// b <= ii, i.e. b+ii' is summable
							if (is_summable_candidate(candidates, complement[ii], b)) {
								for (set<int>::const_iterator it_sat = setCandidates[i].begin(); it_sat != setCandidates[i].end(); ++it_sat) {
									// skip trivial checks
									if (*it_sat == 0 || *it_sat == b)
										continue;
									
									if (!is_summable_candidate(candidates, *it_sat, complement[ii])) {
										is_satisfied = false;
										break;
									}
								} 
								
							}
						}
						
						if (!is_satisfied)
							continue;
						// --
						 



						/* for t = y - 1, y - 2, ... 0, if candidate for [0][y] contains t
						 * then it needs to contain all elements of [0][t] 
						 */

						if (hea::test_subsets_new(n, i, op, it_cand->candidate_pair.second, setCandidates)) {

							/* b \in 0 + y => b + y' are defined */
							bool correct = true;
							for (set<int>::const_iterator it_summable = setCandidates[i].begin(); it_summable != setCandidates[i].end(); ++it_summable) {
								// skip trivial checks 0, y
								if (*it_summable != 0 && *it_summable != it_cand->candidate_pair.second) {
									pair<int, int> my_pair;
									int tmp_compl = complement[ it_cand->candidate_pair.second ];

									if (*it_summable < tmp_compl) {
										my_pair.first = *it_summable;
										my_pair.second = tmp_compl;
									}
									else {
										my_pair.first = tmp_compl;
										my_pair.second = *it_summable;
									}
									
									if (candidates.find(Candidate(my_pair, false)) == candidates.end()) {
										correct = false;	
										break;
									}
											
								}
	} 
							if (correct) {

								op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
								op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

								set<Candidate>::const_iterator it_new = it_cand;
								it_new++;
								solve_candidates(it_new, candidates, last_cand_for_x, n, op, complement, setCandidates);		

								// unset filled elements - empty set
								op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
							}
								op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
						}
						
					}
			}


		}
		else {
			if (it_cand->is_one) {
/*
				complement[it_cand->candidate_pair.first] = it_cand->candidate_pair.second;
				complement[it_cand->candidate_pair.second] = it_cand->candidate_pair.first;
*/
				/* contains 1 */
				for (int i = 1 << (n - 1); i < 1 << n; i++) {

					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

					if (!check_summability(candidates, it_cand->candidate_pair, op))
						continue;

#if 0
					bool valid = true;
					if (last_cand_for_x.find(it_cand->candidate_pair) != last_cand_for_x.end()) {

						// x is summable with all elements in y + 0, if x + y is summable
						for (int i = it_cand->candidate_pair.first; i < n - 1 && valid; i++)
							if (hea::is_summable(op, it_cand->candidate_pair.first, i))
								for (set<int>::const_iterator it = op[0][i]->begin(); it != op[0][i]->end(); ++it)
									if (!hea::is_summable(op, it_cand->candidate_pair.first, *it)) {
										valid = false;
										break;
									}
					}

					if (!valid)
						continue;
#endif

					set<Candidate>::const_iterator it_new = it_cand;
					it_new++;
					solve_candidates(it_new, candidates, last_cand_for_x, n, op, complement, setCandidates);

					// unset
					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
				}

			}
			else {
				/* cannot contain 1, or be equal to {0}, or be empty */
				for (int i = 2; i < 1 << (n - 1); i++) {

					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

					if (!check_summability(candidates, it_cand->candidate_pair, op))
						continue;
#if 0
					bool valid = true;
					if (last_cand_for_x.find(it_cand->candidate_pair) != last_cand_for_x.end()) {
						// x is summable with all elements in y + 0, if x + y is summable
						for (int i = it_cand->candidate_pair.first; i < n - 1 && valid; i++)
							if (hea::is_summable(op, it_cand->candidate_pair.first, i))
								for (set<int>::const_iterator it = op[0][i]->begin(); it != op[0][i]->end(); ++it)
									if (!hea::is_summable(op, it_cand->candidate_pair.first, *it)) {
										valid = false;
										break;
									}

					}
					
					if (!valid)
						continue;
#endif

					set<Candidate>::const_iterator it_new = it_cand;
					it_new++;
					solve_candidates(it_new, candidates, last_cand_for_x, n, op, complement, setCandidates);

					// unset
					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
				}
			}

		}
	}
}


#if 0
// C++11 :(
void init_last_cand_for_x_set11(const int n, const set<Candidate> &candidates, set<pair<int, int> > &last_cand_for_x) {
	int last_x = n - 1;
	bool was_inserted = false;

	for (set<Candidate>::const_iterator it_cand = candidates.crbegin(); it_cand != candidates.crend(); ++it_cand) {
		if (!was_inserted) {
			last_cand_for_x.insert(it_cand->candidate_pair);
			was_inserted = true;
		}
		else {
			if (last_x > it->candidate_pair.first) {
				last_x = it->candidate_pair.first;
				last_cand_for_x.insert(it_cand->candidate_pair);
			}
		}
	}
}
#endif


void solve_candidates_from_state(set<Candidate>::const_iterator it_cand, int level, const set<Candidate> &candidates, const set<pair<int, int> > &last_cand_for_x, const int n, const set<int> *(&op)[MAX][MAX], const int (&complement)[MAX], set<int> *setCandidates, int *from_state, bool *used_from_state) {

	if (it_cand == candidates.end()) {
		

		number_of_tested_configurations++;
		if (number_of_tested_configurations == dump_state_count) {
			number_of_tested_configurations = 0UL;

			/* dump state */

			cerr << "State: ";
			for (set<Candidate>::const_iterator itcan = candidates.begin(); itcan != candidates.end(); ++itcan) {

				int val = compute_set_value(n, op[itcan->candidate_pair.first][itcan->candidate_pair.second]);
				cerr << val << "," ; 
			}
			cerr << endl;
		}


/*
				for (int i =0 ; i < n; i++)
					cout << complement[i] << "  ";
				cout << endl;
*/


		if (hea::is_associative(n, op) /*&& has_unique_orthosupplement(n, op)*/) {
			/*int complement[MAX];*/
			/*for (int i = 0; i < n; i++)
				complement[n-1-i] = i;*/

			/*compute_complement(n, op, complement);*/

			if (hea::is_antisymmetric(n, op, complement) /*&& 
				(!transitive || hea::is_transitive(n, op, complement))*/
			) 
			{
				/* display complement */
				for (int i =0 ; i < n; i++) {
					hea::disp_elem(cout, n, i);
					cout << "  ";
				}
				cout << endl;

				for (int i =0 ; i < n; i++) {
					hea::disp_elem(cout, n, complement[i]);
					cout << "  ";
				}
				cout << endl;

				hea::disp_op(cout, n, op);
				cout << "\t-M-" << endl;
			}
		}

		//disp_op(n, op);
	}
	else {
		if (it_cand->candidate_pair.first == 0) {
			if (it_cand->is_one) {
				/* contains 1 == n - 1 */

				int from;
				if (used_from_state[level]) {
					used_from_state[level] = false;
					from = from_state[level];
				}
				else
					from = 1 << (n - 1);

				for (int i = from; i < 1 << n; i++)

					/* for t = y - 1, y - 2, ... 0, if candidate for [0][y] contains t
					 * then it needs to contain all elements of [0][t] 
					 */

					if (hea::test_subsets_new(n, i, op, it_cand->candidate_pair.second, setCandidates)) {

						/* b \in 0 + y => b + y' are defined */
						bool correct = true;
						for (set<int>::const_iterator it_summable = setCandidates[i].begin(); it_summable != setCandidates[i].end(); ++it_summable) {
							// skip trivial checks 0, y
							if (*it_summable != 0 && *it_summable != it_cand->candidate_pair.second) {
								pair<int, int> my_pair;
								int tmp_compl = complement[ it_cand->candidate_pair.second ];

								if (*it_summable < tmp_compl) {
									my_pair.first = *it_summable;
									my_pair.second = tmp_compl;
								}
								else {
									my_pair.first = tmp_compl;
									my_pair.second = *it_summable;
								}
								
								if (candidates.find(Candidate(my_pair, false)) == candidates.end()) {
									correct = false;	
									break;
								}
										
							}
} 
						if (correct) {

							op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
							op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

							set<Candidate>::const_iterator it_new = it_cand;
							it_new++;
							solve_candidates_from_state(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, from_state, used_from_state);		

							// unset filled elements - empty set
							op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
							op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
						}
					}
			}
			else {
				/* does not contain one */
				/* it must be defined - excludes i = 0 */
				/* {0} cannot be used any more  - excludes i = 1 */
				int from;
				if (used_from_state[level]) {
					used_from_state[level] = false;
					from = from_state[level];
				}
				else
					from = 2;
				
				for (int i = from; i < (1 << (n - 1)); i++)
					/* x\in 0 + x */
					if (setCandidates[i].find(it_cand->candidate_pair.second) != setCandidates[i].end()) {
						bool was_used = false;

						for (int k = it_cand->candidate_pair.second - 1; k >= 1; k--) {
							if (op[0][k] == &setCandidates[i]) {
								was_used = true;	
								break;
							}
						}
						
						if (was_used)
							continue;

						// Candidate for 0+b: b<=c, then (0+b) is summable with c'
						//cout << "Level: " << level << ", candidate: " << i << endl;
						int b = it_cand->candidate_pair.second;
						bool is_satisfied = true;

						// 1==n-1 cannot be contained and 0 is trivial, skipping
						for (int ii = 1; is_satisfied && ii < n - 1; ii++) {
							// skipping also b as trivial
							if (ii == b)
								continue;

							// b <= ii, i.e. b+ii' is summable
							if (is_summable_candidate(candidates, complement[ii], b)) {
								//cout << "ii: " << ii << ", ii' = " << complement[ii] << endl;
								for (set<int>::const_iterator it_sat = setCandidates[i].begin(); it_sat != setCandidates[i].end(); ++it_sat) {
									// skip trivial checks
									if (*it_sat == 0 || *it_sat == b)
										continue;
									
									if (!is_summable_candidate(candidates, *it_sat, complement[ii])) {
										
										is_satisfied = false;
										break;
									}
								} 
								
							}
						}
						
						if (!is_satisfied)
							continue;
						// --


						/* for t = y - 1, y - 2, ... 0, if candidate for [0][y] contains t
						 * then it needs to contain all elements of [0][t] 
						 */

						if (hea::test_subsets_new(n, i, op, it_cand->candidate_pair.second, setCandidates)) {


							/* b \in 0 + y => b + y' are defined */
							bool correct = true;
							for (set<int>::const_iterator it_summable = setCandidates[i].begin(); it_summable != setCandidates[i].end(); ++it_summable) {
								// skip trivial checks 0, y
								if (*it_summable != 0 && *it_summable != it_cand->candidate_pair.second) {
									pair<int, int> my_pair;
									int tmp_compl = complement[ it_cand->candidate_pair.second ];

									if (*it_summable < tmp_compl) {
										my_pair.first = *it_summable;
										my_pair.second = tmp_compl;
									}
									else {
										my_pair.first = tmp_compl;
										my_pair.second = *it_summable;
									}
									
									if (candidates.find(Candidate(my_pair, false)) == candidates.end()) {
										correct = false;	
										break;
									}
											
								}
	} 
							if (correct) {

								op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
								op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

								set<Candidate>::const_iterator it_new = it_cand;
								it_new++;
								solve_candidates_from_state(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, from_state, used_from_state);		

								// unset filled elements - empty set
								op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
								op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
							}
						}
						
					}
			}


		}
		else {
			if (it_cand->is_one) {
/*
				complement[it_cand->candidate_pair.first] = it_cand->candidate_pair.second;
				complement[it_cand->candidate_pair.second] = it_cand->candidate_pair.first;
*/
				/* contains 1 */

				int from;
				if (used_from_state[level]) {
					used_from_state[level] = false;
					from = from_state[level];
				}
				else
					from = 1 << (n - 1);

				for (int i = from; i < 1 << n; i++) {

					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

					if (!check_summability(candidates, it_cand->candidate_pair, op))
						continue;
#if 0
					bool valid = true;
					if (last_cand_for_x.find(it_cand->candidate_pair) != last_cand_for_x.end()) {

						// x is summable with all elements in y + 0, if x + y is summable
						for (int i = it_cand->candidate_pair.first; i < n - 1 && valid; i++)
							if (hea::is_summable(op, it_cand->candidate_pair.first, i))
								for (set<int>::const_iterator it = op[0][i]->begin(); it != op[0][i]->end(); ++it)
									if (!hea::is_summable(op, it_cand->candidate_pair.first, *it)) {
										valid = false;
										break;
									}
					}

					if (!valid)
						continue;
#endif

					set<Candidate>::const_iterator it_new = it_cand;
					it_new++;
					solve_candidates_from_state(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, from_state, used_from_state);

					// unset
					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
				}

			}
			else {
				/* cannot contain 1, or be equal to {0}, or be empty */
				int from;
				if (used_from_state[level]) {
					used_from_state[level] = false;
					from = from_state[level];
				}
				else
					from = 2;

				for (int i = from; i < 1 << (n - 1); i++) {

					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

					if (!check_summability(candidates, it_cand->candidate_pair, op))
						continue;

#if 0
					bool valid = true;
					if (last_cand_for_x.find(it_cand->candidate_pair) != last_cand_for_x.end()) {
						// x is summable with all elements in y + 0, if x + y is summable
						for (int i = it_cand->candidate_pair.first; i < n - 1 && valid; i++)
							if (hea::is_summable(op, it_cand->candidate_pair.first, i))
								for (set<int>::const_iterator it = op[0][i]->begin(); it != op[0][i]->end(); ++it)
									if (!hea::is_summable(op, it_cand->candidate_pair.first, *it)) {
										valid = false;
										break;
									}

					}
					
					if (!valid)
						continue;

#endif

					set<Candidate>::const_iterator it_new = it_cand;
					it_new++;
					solve_candidates_from_state(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, from_state, used_from_state);

					// unset
					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
				}
			}

		}
	}
}


void solve_candidates_from_state_to_state(set<Candidate>::const_iterator it_cand, int level, const set<Candidate> &candidates, const set<pair<int, int> > &last_cand_for_x, const int n, const set<int> *(&op)[MAX][MAX], const int (&complement)[MAX], set<int> *setCandidates, int *from_state, bool *used_from_state, int *to_state, bool *used_to_state) {

	if (it_cand == candidates.end()) {
		

		number_of_tested_configurations++;
		if (number_of_tested_configurations == dump_state_count) {
			number_of_tested_configurations = 0UL;

			/* dump state */

			cerr << "State: ";
			for (set<Candidate>::const_iterator itcan = candidates.begin(); itcan != candidates.end(); ++itcan) {

				int val = compute_set_value(n, op[itcan->candidate_pair.first][itcan->candidate_pair.second]);
				cerr << val << "," ; 
			}
			cerr << endl;
		}


/*
				for (int i =0 ; i < n; i++)
					cout << complement[i] << "  ";
				cout << endl;
*/


		if (hea::is_associative(n, op) /*&& has_unique_orthosupplement(n, op)*/) {
			/*int complement[MAX];*/
			/*for (int i = 0; i < n; i++)
				complement[n-1-i] = i;*/

			/*compute_complement(n, op, complement);*/

			if (hea::is_antisymmetric(n, op, complement) /*&& 
				(!transitive || hea::is_transitive(n, op, complement))*/
			) 
			{
				/* display complement */
				for (int i =0 ; i < n; i++) {
					hea::disp_elem(cout, n, i);
					cout << "  ";
				}
				cout << endl;

				for (int i =0 ; i < n; i++) {
					hea::disp_elem(cout, n, complement[i]);
					cout << "  ";
				}
				cout << endl;

				hea::disp_op(cout, n, op);
				cout << "\t-M-" << endl;
			}
		}

		//disp_op(n, op);
	}
	else {
		if (it_cand->candidate_pair.first == 0) {
			if (it_cand->is_one) {
				/* contains 1 == n - 1 */

				int from;
				if (used_from_state[level]) {
					used_from_state[level] = false;
					from = from_state[level];
				}
				else
					from = 1 << (n - 1);

				int to;
				if (used_to_state[level]) {
					used_to_state[level] = false;
					to = to_state[level];
				}
				else
					to = (1 << n) - 1;

				for (int i = from; i <= to; i++)

					/* for t = y - 1, y - 2, ... 0, if candidate for [0][y] contains t
					 * then it needs to contain all elements of [0][t] 
					 */

					if (hea::test_subsets_new(n, i, op, it_cand->candidate_pair.second, setCandidates)) {

						/* b \in 0 + y => b + y' are defined */
						bool correct = true;
						for (set<int>::const_iterator it_summable = setCandidates[i].begin(); it_summable != setCandidates[i].end(); ++it_summable) {
							// skip trivial checks 0, y
							if (*it_summable != 0 && *it_summable != it_cand->candidate_pair.second) {
								pair<int, int> my_pair;
								int tmp_compl = complement[ it_cand->candidate_pair.second ];

								if (*it_summable < tmp_compl) {
									my_pair.first = *it_summable;
									my_pair.second = tmp_compl;
								}
								else {
									my_pair.first = tmp_compl;
									my_pair.second = *it_summable;
								}
								
								if (candidates.find(Candidate(my_pair, false)) == candidates.end()) {
									correct = false;	
									break;
								}
										
							}
} 
						if (correct) {

							op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
							op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

							set<Candidate>::const_iterator it_new = it_cand;
							it_new++;
							solve_candidates_from_state_to_state(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, from_state, used_from_state, to_state, used_to_state);		

							// unset filled elements - empty set
							op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
							op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
						}
					}
			}
			else {
				/* does not contain one */
				/* it must be defined - excludes i = 0 */
				/* {0} cannot be used any more  - excludes i = 1 */
				int from;
				if (used_from_state[level]) {
					used_from_state[level] = false;
					from = from_state[level];
				}
				else
					from = 2;

				int to;
				if (used_to_state[level]) {
					used_to_state[level] = false;
					to = to_state[level];
				}
				else
					to = (1 << (n - 1)) - 1;
				
				for (int i = from; i <= to; i++)
					/* x\in 0 + x */
					if (setCandidates[i].find(it_cand->candidate_pair.second) != setCandidates[i].end()) {
						bool was_used = false;

						for (int k = it_cand->candidate_pair.second - 1; k >= 1; k--) {
							if (op[0][k] == &setCandidates[i]) {
								was_used = true;	
								break;
							}
						}
						
						if (was_used)
							continue;

						// Candidate for 0+b: b<=c, then (0+b) is summable with c'
						//cout << "Level: " << level << ", candidate: " << i << endl;
						int b = it_cand->candidate_pair.second;
						bool is_satisfied = true;

						// 1==n-1 cannot be contained and 0 is trivial, skipping
						for (int ii = 1; is_satisfied && ii < n - 1; ii++) {
							// skipping also b as trivial
							if (ii == b)
								continue;

							// b <= ii, i.e. b+ii' is summable
							if (is_summable_candidate(candidates, complement[ii], b)) {
								//cout << "ii: " << ii << ", ii' = " << complement[ii] << endl;
								for (set<int>::const_iterator it_sat = setCandidates[i].begin(); it_sat != setCandidates[i].end(); ++it_sat) {
									// skip trivial checks
									if (*it_sat == 0 || *it_sat == b)
										continue;
									
									if (!is_summable_candidate(candidates, *it_sat, complement[ii])) {
										
										is_satisfied = false;
										break;
									}
								} 
								
							}
						}
						
						if (!is_satisfied)
							continue;
						// --


						/* for t = y - 1, y - 2, ... 0, if candidate for [0][y] contains t
						 * then it needs to contain all elements of [0][t] 
						 */

						if (hea::test_subsets_new(n, i, op, it_cand->candidate_pair.second, setCandidates)) {


							/* b \in 0 + y => b + y' are defined */
							bool correct = true;
							for (set<int>::const_iterator it_summable = setCandidates[i].begin(); it_summable != setCandidates[i].end(); ++it_summable) {
								// skip trivial checks 0, y
								if (*it_summable != 0 && *it_summable != it_cand->candidate_pair.second) {
									pair<int, int> my_pair;
									int tmp_compl = complement[ it_cand->candidate_pair.second ];

									if (*it_summable < tmp_compl) {
										my_pair.first = *it_summable;
										my_pair.second = tmp_compl;
									}
									else {
										my_pair.first = tmp_compl;
										my_pair.second = *it_summable;
									}
									
									if (candidates.find(Candidate(my_pair, false)) == candidates.end()) {
										correct = false;	
										break;
									}
											
								}
	} 
							if (correct) {

								op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
								op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

								set<Candidate>::const_iterator it_new = it_cand;
								it_new++;
								solve_candidates_from_state_to_state(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, from_state, used_from_state, to_state, used_to_state);		

								// unset filled elements - empty set
								op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
								op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
							}
						}
						
					}
			}


		}
		else {
			if (it_cand->is_one) {
/*
				complement[it_cand->candidate_pair.first] = it_cand->candidate_pair.second;
				complement[it_cand->candidate_pair.second] = it_cand->candidate_pair.first;
*/
				/* contains 1 */

				int from;
				if (used_from_state[level]) {
					used_from_state[level] = false;
					from = from_state[level];
				}
				else
					from = 1 << (n - 1);

				int to;
				if (used_to_state[level]) {
					used_to_state[level] = false;
					to = to_state[level];
				}
				else
					to = (1 << n) - 1;

				for (int i = from; i <= to; i++) {

					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

					if (!check_summability(candidates, it_cand->candidate_pair, op))
						continue;
#if 0
					bool valid = true;
					if (last_cand_for_x.find(it_cand->candidate_pair) != last_cand_for_x.end()) {

						// x is summable with all elements in y + 0, if x + y is summable
						for (int i = it_cand->candidate_pair.first; i < n - 1 && valid; i++)
							if (hea::is_summable(op, it_cand->candidate_pair.first, i))
								for (set<int>::const_iterator it = op[0][i]->begin(); it != op[0][i]->end(); ++it)
									if (!hea::is_summable(op, it_cand->candidate_pair.first, *it)) {
										valid = false;
										break;
									}
					}

					if (!valid)
						continue;
#endif

					set<Candidate>::const_iterator it_new = it_cand;
					it_new++;
					solve_candidates_from_state_to_state(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, from_state, used_from_state, to_state, used_to_state);

					// unset
					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
				}

			}
			else {
				/* cannot contain 1, or be equal to {0}, or be empty */
				int from;
				if (used_from_state[level]) {
					used_from_state[level] = false;
					from = from_state[level];
				}
				else
					from = 2;

				int to;
				if (used_to_state[level]) {
					used_to_state[level] = false;
					to = to_state[level];
				}
				else
					to = (1 << (n - 1)) - 1;

				for (int i = from; i <= to; i++) {

					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

					if (!check_summability(candidates, it_cand->candidate_pair, op))
						continue;

#if 0
					bool valid = true;
					if (last_cand_for_x.find(it_cand->candidate_pair) != last_cand_for_x.end()) {
						// x is summable with all elements in y + 0, if x + y is summable
						for (int i = it_cand->candidate_pair.first; i < n - 1 && valid; i++)
							if (hea::is_summable(op, it_cand->candidate_pair.first, i))
								for (set<int>::const_iterator it = op[0][i]->begin(); it != op[0][i]->end(); ++it)
									if (!hea::is_summable(op, it_cand->candidate_pair.first, *it)) {
										valid = false;
										break;
									}

					}
					
					if (!valid)
						continue;

#endif

					set<Candidate>::const_iterator it_new = it_cand;
					it_new++;
					solve_candidates_from_state_to_state(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, from_state, used_from_state, to_state, used_to_state);

					// unset
					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
				}
			}

		}
	}
}

#if 0
unsigned int set_to_integer(const set<int> &set) {
	unsigned int result=0;

	for (set<int>::const_iterator it_set = set.begin(); it_set != set.end(); ++it_set)
		result |= (1 << *it_set);

	return result;
}
#endif

/* checks 0 + A = A, provided that results for 0+x are already defined */

bool is_0_plus_set_equal_to_set(const set<int> *(&op)[MAX][MAX], int set_candidate_no, set<int> *setCandidates) {

	set<int> result = hea::compute_sum(op, 0, setCandidates[set_candidate_no]);

	return result == setCandidates[set_candidate_no];
}

void clear_used_from(fromToState &current_setup, int level) {
	int j = level + 1;
	while (j < MAX * MAX)  {
		current_setup.used_from_state[j] = false;
		j++;
	}

}

void solve_candidates_state_setup(set<Candidate>::const_iterator it_cand, int level, const set<Candidate> &candidates, const set<pair<int, int> > &last_cand_for_x, const int n, const set<int> *(&op)[MAX][MAX], const int (&complement)[MAX], set<int> *setCandidates, fromToState &current_setup) {

	if (it_cand == candidates.end()) {
		

		number_of_tested_configurations++;
		if (number_of_tested_configurations == dump_state_count) {
			number_of_tested_configurations = 0UL;

			/* dump state */

			cerr << "State: ";
			for (set<Candidate>::const_iterator itcan = candidates.begin(); itcan != candidates.end(); ++itcan) {

				int val = compute_set_value(n, op[itcan->candidate_pair.first][itcan->candidate_pair.second]);
				cerr << val << "," ; 
			}
			cerr << endl;
		}


/*
				for (int i =0 ; i < n; i++)
					cout << complement[i] << "  ";
				cout << endl;
*/


		if (hea::is_associative(n, op) /*&& has_unique_orthosupplement(n, op)*/) {
			/*int complement[MAX];*/
			/*for (int i = 0; i < n; i++)
				complement[n-1-i] = i;*/

			/*compute_complement(n, op, complement);*/

			if (hea::is_antisymmetric(n, op, complement) /*&& 
				(!transitive || hea::is_transitive(n, op, complement))*/
			) 
			{
				/* display complement */
				for (int i =0 ; i < n; i++) {
					hea::disp_elem(cout, n, i);
					cout << "  ";
				}
				cout << endl;

				for (int i =0 ; i < n; i++) {
					hea::disp_elem(cout, n, complement[i]);
					cout << "  ";
				}
				cout << endl;

				hea::disp_op(cout, n, op);
				cout << "\t-M-" << endl;
			}
		}

		//disp_op(n, op);
	}
	else {
		if (it_cand->candidate_pair.first == 0) {
			if (it_cand->is_one) {
				/* contains 1 == n - 1 */

				int from;
				bool was_cleared = true; 
				if (current_setup.bFromState && current_setup.used_from_state[level]) {
					current_setup.used_from_state[level] = false;
					from = current_setup.from_state[level];
					was_cleared = false;
				}
				else
					from = 1 << (n - 1);

				int to;
				if (current_setup.bToState && current_setup.used_to_state[level]) {
					current_setup.used_to_state[level] = false;
					to = current_setup.to_state[level];
				}
				else
					to = (1 << n) - 1;

				for (int i = from; i <= to; i++)

					/* for t = y - 1, y - 2, ... 0, if candidate for [0][y] contains t
					 * then it needs to contain all elements of [0][t] 
					 */

					if (hea::test_subsets_new(n, i, op, it_cand->candidate_pair.second, setCandidates)) {

						/* b \in 0 + y => b + y' are defined */
						bool correct = true;
						for (set<int>::const_iterator it_summable = setCandidates[i].begin(); it_summable != setCandidates[i].end(); ++it_summable) {
							// skip trivial checks 0, y
							if (*it_summable != 0 && *it_summable != it_cand->candidate_pair.second) {
								pair<int, int> my_pair;
								int tmp_compl = complement[ it_cand->candidate_pair.second ];

								if (*it_summable < tmp_compl) {
									my_pair.first = *it_summable;
									my_pair.second = tmp_compl;
								}
								else {
									my_pair.first = tmp_compl;
									my_pair.second = *it_summable;
								}
								
								if (candidates.find(Candidate(my_pair, false)) == candidates.end()) {
									correct = false;	
									break;
								}
										
							}
} 
						if (correct) {

							op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
							op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

							set<Candidate>::const_iterator it_new = it_cand;
							it_new++;
							solve_candidates_state_setup(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, current_setup);		

							// unset filled elements - empty set
							op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
							op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
						}
						else {
							/* we are moving to another i (from + 1), thus we need to set all successive bFromState to false */
							if (!was_cleared) {
								clear_used_from(current_setup, level);

								was_cleared = true;
							}
						}
					}
			}
			else {
				/* does not contain one */
				/* it must be defined - excludes i = 0 */
				/* {0} cannot be used any more  - excludes i = 1 */

				int from;
				bool was_cleared = true; 
				if (current_setup.bFromState && current_setup.used_from_state[level]) {
					current_setup.used_from_state[level] = false;
					from = current_setup.from_state[level];

					was_cleared = false; 
				}
				else
					from = 2;

				int to;
				if (current_setup.bToState && current_setup.used_to_state[level]) {
					current_setup.used_to_state[level] = false;
					to = current_setup.to_state[level];
				}
				else
					to = (1 << (n - 1)) - 1;
				
				for (int i = from; i <= to; i++)
					/* x\in 0 + x */
					if (setCandidates[i].find(it_cand->candidate_pair.second) != setCandidates[i].end()) {
						bool was_used = false;

						for (int k = it_cand->candidate_pair.second - 1; k >= 1; k--) {
							if (op[0][k] == &setCandidates[i]) {
								was_used = true;	
								break;
							}
						}
						
						if (was_used) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						// Candidate for 0+b: b<=c, then (0+b) is summable with c'
						//cout << "Level: " << level << ", candidate: " << i << endl;
						int b = it_cand->candidate_pair.second;
						bool is_satisfied = true;

						// 1==n-1 cannot be contained and 0 is trivial, skipping
						for (int ii = 1; is_satisfied && ii < n - 1; ii++) {
							// skipping also b as trivial
							if (ii == b)
								continue;

							// b <= ii, i.e. b+ii' is summable
							if (is_summable_candidate(candidates, complement[ii], b)) {
								//cout << "ii: " << ii << ", ii' = " << complement[ii] << endl;
								for (set<int>::const_iterator it_sat = setCandidates[i].begin(); it_sat != setCandidates[i].end(); ++it_sat) {
									// skip trivial checks
									if (*it_sat == 0 || *it_sat == b)
										continue;
									
									if (!is_summable_candidate(candidates, *it_sat, complement[ii])) {
										
										is_satisfied = false;
										break;
									}
								} 
								
							}
						}
						
						if (!is_satisfied) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
						// --


						/* for t = y - 1, y - 2, ... 0, if candidate for [0][y] contains t
						 * then it needs to contain all elements of [0][t] 
						 */

						if (hea::test_subsets_new(n, i, op, it_cand->candidate_pair.second, setCandidates)) {


							/* b \in 0 + y => b + y' are defined */
							bool correct = true;
							for (set<int>::const_iterator it_summable = setCandidates[i].begin(); it_summable != setCandidates[i].end(); ++it_summable) {
								// skip trivial checks 0, y
								if (*it_summable != 0 && *it_summable != it_cand->candidate_pair.second) {
									pair<int, int> my_pair;
									int tmp_compl = complement[ it_cand->candidate_pair.second ];

									if (*it_summable < tmp_compl) {
										my_pair.first = *it_summable;
										my_pair.second = tmp_compl;
									}
									else {
										my_pair.first = tmp_compl;
										my_pair.second = *it_summable;
									}
									
									if (candidates.find(Candidate(my_pair, false)) == candidates.end()) {
										correct = false;	
										break;
									}
											
								}
	} 
							if (correct) {

								op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
								op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

								set<Candidate>::const_iterator it_new = it_cand;
								it_new++;
								solve_candidates_state_setup(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, current_setup);		

								// unset filled elements - empty set
								op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
								op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
							}
							else {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
							}
						}
						
					}
			}


		}
		else {
			if (it_cand->is_one) {
/*
				complement[it_cand->candidate_pair.first] = it_cand->candidate_pair.second;
				complement[it_cand->candidate_pair.second] = it_cand->candidate_pair.first;
*/
				/* contains 1 */

				int from;
				bool was_cleared = true; 

				if (current_setup.bFromState && current_setup.used_from_state[level]) {
					current_setup.used_from_state[level] = false;
					from = current_setup.from_state[level];
					was_cleared = false;
				}
				else
					from = 1 << (n - 1);

				int to;
				if (current_setup.bToState && current_setup.used_to_state[level]) {
					current_setup.used_to_state[level] = false;
					to = current_setup.to_state[level];
				}
				else
					to = (1 << n) - 1;





#if (modelno == 25)
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 4) {
				// 0 + a = {a}	
					if (op[0][1] == &setCandidates[2]) {
					
							int from_new = op[0][n - 1] - &setCandidates[0];
						
							if (from_new > from) {
								from = from_new;
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
							}
					}
					else {
						// 0 + a = {0, a}
						if (op[0][1] == &setCandidates[3]) {
							int from_new = op[0][n - 1] - &setCandidates[0];
							from_new &= ~(op[0][4] - &setCandidates[0]);

							if (from_new > from) {
								from = from_new;
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
							}
						}
					}
				}

				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 3) {
					int from_new = op[0][n - 1] - &setCandidates[0];

					for (set<int>::const_iterator it_s = op[0][2]->begin(); it_s != op[0][2]->end(); ++it_s) {
						if (*it_s != 2) {
							from_new &= ~(op[*it_s][3] - &setCandidates[0]);
						}
					}
					
					if (from_new > from) {
						from = from_new;
						if (!was_cleared) {
							clear_used_from(current_setup, level);
							was_cleared = true;
						}
					}
				}
#endif

#if (modelno == 24)
	if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {

		if (op[0][1] == &setCandidates[2]) {
					/* only for a + a and model number 24 and a + 0 = {a} ! */
						int from_new = op[0][n - 1] - &setCandidates[0];

							//std::cerr << "From new: " << from_new << endl;

						if (from_new > from) {
							from = from_new;
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
						}
		}
		else
		/* a + 0 = \{0, a\}, (a + a) + 0 = (a + a) union \{0, a\} */
		if (op[0][1] == &setCandidates[3]) {
					int from_new = op[0][n - 1] - &setCandidates[0];
					from_new &= ~3;

						//std::cerr << "From new: " << from_new << endl;

					if (from_new > from) {
						from = from_new;
						if (!was_cleared) {
							clear_used_from(current_setup, level);
							was_cleared = true;
						}

					}
				
		}
	}
	else
	if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
	}
#endif


				for (int i = from; i <= to; i++) {

#if (modelno == 1)
				
				/* a + a */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0a = setCandidates[i];
						resA0a.insert(0);
						resA0a.insert(1);
						
						if (resA0 != resA0a) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
						
				}
				else
				/* b + b */
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
					if (op[0][2] == &setCandidates[4]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0b = setCandidates[i];
						resA0b.insert(0);
						resA0b.insert(2);
						
						if (resA0 != resA0b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
						
				}
				else
				/* c + c */
				if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 3) {
					if (op[0][3] == &setCandidates[8]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][3] == &setCandidates[9]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0c = setCandidates[i];
						resA0c.insert(0);
						resA0c.insert(3);
						
						if (resA0 != resA0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
				}
				else
				/* d + d */
				if (it_cand->candidate_pair.first == 4 && it_cand->candidate_pair.second == 4) {
					if (op[0][4] == &setCandidates[16]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][4] == &setCandidates[17]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0d = setCandidates[i];
						resA0d.insert(0);
						resA0d.insert(4);
						
						if (resA0 != resA0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
				}
#endif

#if (modelno == 2)
				/* a + a */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0a = setCandidates[i];
						resA0a.insert(0);
						resA0a.insert(1);
						
						if (resA0 != resA0a) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
						
				}
				else
				/* b + b */
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
					if (op[0][2] == &setCandidates[4]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0b = setCandidates[i];
						resA0b.insert(0);
						resA0b.insert(2);
						
						if (resA0 != resA0b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
						
				}
				else
				/* c + d */
				if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
					if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[16]) {
						/* c + 0 = \{c\} && d + 0 = \{d\} => (c + d) + 0 = (c + d) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					
					}
					else 
					if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[17]) {
						/* c + 0 = \{c\} && d + 0 = \{0, d\} => (c + d) + 0 = c + d = (c + d) union \{c\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAc = setCandidates[i]; /* (c + d) union \{c\} */
						resAc.insert(3);
						if (resA0 != resAc) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[16]) {
						/* c + 0 = \{0, c\} && d + 0 = \{d\} => (c + d) + 0 = c + d =  (c + d) union \{d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAd = setCandidates[i]; /* (c + d) union \{d\} */
						resAd.insert(4);
						if (resA0 != resAd) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[17]) {
						/* c + 0 = \{0, c\} && d + 0 = \{0, d\} => (c + d) + 0 = (c + d) union \{0, c\} = (c + d) union \{0, d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0c = setCandidates[i]; /* (c + d) union \{0, c\} */
						resA0c.insert(0);
						resA0c.insert(3);
	
						if (resA0 != resA0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resA0d = setCandidates[i]; /* (c + d) union \{0, d\} */
						resA0d.insert(0);
						resA0d.insert(4);
	
						if (resA0 != resA0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}

				}
#endif



#if (modelno == 3)
				/* a + b */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
					if (op[0][1] == &setCandidates[2] && op[0][2] == &setCandidates[4]) {
						/* a + 0 = \{a\} && b + 0 = \{b\} => (a + b) + 0 = (a + b) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					
					}
					else 
					if (op[0][1] == &setCandidates[2] && op[0][2] == &setCandidates[5]) {
						/* a + 0 = \{a\} && b + 0 = \{0, b\} => (a + b) + 0 = a + b = (a + b) union \{a\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAa = setCandidates[i]; /* (a + b) union \{a\} */
						resAa.insert(1);
						if (resA0 != resAa) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][1] == &setCandidates[3] && op[0][2] == &setCandidates[4]) {
						/* a + 0 = \{0, a\} && b + 0 = \{b\} => (a + b) + 0 = a + b =  (a + b) union \{b\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAb = setCandidates[i]; /* (a + b) union \{b\} */
						resAb.insert(2);
						if (resA0 != resAb) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][1] == &setCandidates[3] && op[0][2] == &setCandidates[5]) {
						/* a + 0 = \{0, a\} && b + 0 = \{0, b\} => (a + b) + 0 = (a + b) union \{0, a\} = (a + b) union \{0, b\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0a = setCandidates[i]; /* (a + b) union \{0, a\} */
						resA0a.insert(0);
						resA0a.insert(1);
	
						if (resA0 != resA0a) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resA0b = setCandidates[i]; /* (a + b) union \{0, b\} */
						resA0b.insert(0);
						resA0b.insert(2);
	
						if (resA0 != resA0b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}

				}
				else
				/* c + d */
				if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
					if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[16]) {
						/* c + 0 = \{c\} && d + 0 = \{d\} => (c + d) + 0 = (c + d) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					
					}
					else 
					if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[17]) {
						/* c + 0 = \{c\} && d + 0 = \{0, d\} => (c + d) + 0 = c + d = (c + d) union \{c\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAc = setCandidates[i]; /* (c + d) union \{c\} */
						resAc.insert(3);
						if (resA0 != resAc) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[16]) {
						/* c + 0 = \{0, c\} && d + 0 = \{d\} => (c + d) + 0 = c + d =  (c + d) union \{d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAd = setCandidates[i]; /* (c + d) union \{d\} */
						resAd.insert(4);
						if (resA0 != resAd) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[17]) {
						/* c + 0 = \{0, c\} && d + 0 = \{0, d\} => (c + d) + 0 = (c + d) union \{0, c\} = (c + d) union \{0, d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0c = setCandidates[i]; /* (c + d) union \{0, c\} */
						resA0c.insert(0);
						resA0c.insert(3);
	
						if (resA0 != resA0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resA0d = setCandidates[i]; /* (c + d) union \{0, d\} */
						resA0d.insert(0);
						resA0d.insert(4);
	
						if (resA0 != resA0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}

				}
#endif

#if (modelno == 4) 
				/* a + a */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0a = setCandidates[i];
						resA0a.insert(0);
						resA0a.insert(1);
						
						if (resA0 != resA0a) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
						
				}
				else
				/* b + b */
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
					if (op[0][2] == &setCandidates[4]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0b = setCandidates[i];
						resA0b.insert(0);
						resA0b.insert(2);
						
						if (resA0 != resA0b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
						
				}
				else
				/* c + d */
				if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
					if (op[0][3] == &setCandidates[8]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][3] == &setCandidates[9]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0d = setCandidates[i];
						for (set<int>::const_iterator it_0d=op[0][4]->begin(); it_0d != op[0][4]->end(); ++it_0d)
							resA_0d.insert(*it_0d);
						
						if (resA0 != resA_0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}



#endif

#if (modelno == 5) 
				/* a + b */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
					if (op[0][1] == &setCandidates[2] && op[0][2] == &setCandidates[4]) {
						/* a + 0 = \{a\} && b + 0 = \{b\} => (a + b) + 0 = (a + b) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					
					}
					else 
					if (op[0][1] == &setCandidates[2] && op[0][2] == &setCandidates[5]) {
						/* a + 0 = \{a\} && b + 0 = \{0, b\} => (a + b) + 0 = a + b = (a + b) union \{a\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAa = setCandidates[i]; /* (a + b) union \{a\} */
						resAa.insert(1);
						if (resA0 != resAa) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][1] == &setCandidates[3] && op[0][2] == &setCandidates[4]) {
						/* a + 0 = \{0, a\} && b + 0 = \{b\} => (a + b) + 0 = a + b =  (a + b) union \{b\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAb = setCandidates[i]; /* (a + b) union \{b\} */
						resAb.insert(2);
						if (resA0 != resAb) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][1] == &setCandidates[5] && op[0][2] == &setCandidates[5]) {
						/* a + 0 = \{0, a\} && b + 0 = \{0, b\} => (a + b) + 0 = (a + b) union \{0, a\} = (a + b) union \{0, b\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0a = setCandidates[i]; /* (a + b) union \{0, a\} */
						resA0a.insert(0);
						resA0a.insert(1);
	
						if (resA0 != resA0a) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resA0b = setCandidates[i]; /* (a + b) union \{0, b\} */
						resA0b.insert(0);
						resA0b.insert(2);
	
						if (resA0 != resA0b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}

				}
				else
				/* c + d */
				if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
					if (op[0][3] == &setCandidates[8]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][3] == &setCandidates[9]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0d = setCandidates[i];
						for (set<int>::const_iterator it_0d=op[0][4]->begin(); it_0d != op[0][4]->end(); ++it_0d)
							resA_0d.insert(*it_0d);
						
						if (resA0 != resA_0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}



#endif

#if (modelno == 6)
				/* a + a */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
					/* a + 0 = \{a\} 
						a + a = A
							A + 0 = A
							1 + 0 \subseteq A + 0

					*/
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					/* a + 0 = \{0, a\} 
						a + a = A
							A + 0 = A \union \{0, a\}
							1 + 0 \subseteq A + 0
					*/
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0a = setCandidates[i];
						resA0a.insert(0);
						resA0a.insert(1);
						
						if (resA0 != resA0a) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
						
				}
				else
				/* b + c */
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 3) {
					if (op[0][2] == &setCandidates[4]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0c = setCandidates[i];
						for (set<int>::const_iterator it_0c=op[0][3]->begin(); it_0c != op[0][3]->end(); ++it_0c)
							resA_0c.insert(*it_0c);
						
						if (resA0 != resA_0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}
				else
				/* d + d */
				if (it_cand->candidate_pair.first == 4 && it_cand->candidate_pair.second == 4) {
					if (op[0][4] == &setCandidates[16]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][4] == &setCandidates[17]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0d = setCandidates[i];
						resA0d.insert(0);
						resA0d.insert(4);
						
						if (resA0 != resA0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
						
				}

#endif


#if (modelno == 7)
				/* a + a */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
					/* a + 0 = \{a\} 
						a + a = A
							A + 0 = A
							1 + 0 \subseteq A + 0

					*/
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					/* a + 0 = \{0, a\} 
						a + a = A
							A + 0 = A \union \{0, a\}
							1 + 0 \subseteq A + 0
					*/
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0a = setCandidates[i];
						resA0a.insert(0);
						resA0a.insert(1);
						
						if (resA0 != resA0a) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
						
				}
				else
				/* b + d */
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 4) {
					if (op[0][2] == &setCandidates[4]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0d = setCandidates[i];
						for (set<int>::const_iterator it_0d=op[0][4]->begin(); it_0d != op[0][4]->end(); ++it_0d)
							resA_0d.insert(*it_0d);
						
						if (resA0 != resA_0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}

#endif


#if (modelno == 8)
				/* a + a */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
					/* a + 0 = \{a\} 
						a + a = A
							A + 0 = A
							1 + 0 \subseteq A + 0

					*/
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					/* a + 0 = \{0, a\} 
						a + a = A
							A + 0 = A \union \{0, a\}
							1 + 0 \subseteq A + 0
					*/
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0a = setCandidates[i];
						resA0a.insert(0);
						resA0a.insert(1);
						
						if (resA0 != resA0a) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
						
				}
				/* b + b */
				else
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
					/* b + 0 = \{b\} 

						b + b = A;
							A + 0 = A
							1 + 0 \subseteq A + 0  = A
					*/
					
					if (op[0][2] == &setCandidates[4]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					/* b + 0 = \{0, b\} 

						b + b = A;
							A + 0 = A \union \{0, b\}
							1 + 0 \subseteq A + 0  
					
					*/
					if (op[0][2] == &setCandidates[5]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0b = setCandidates[i];
						resA0b.insert(0);
						resA0b.insert(2);
						
						if (resA0 != resA0b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}

				}
				/* c + d */
				if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
					if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[16]) {
						/* c + 0 = \{c\} && d + 0 = \{d\} => (c + d) + 0 = (c + d) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					
					}
					else 
					if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[17]) {
						/* c + 0 = \{c\} && d + 0 = \{0, d\} => (c + d) + 0 = c + d = (c + d) union \{c\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAc = setCandidates[i]; /* (c + d) union \{c\} */
						resAc.insert(3);
						if (resA0 != resAc) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[16]) {
						/* c + 0 = \{0, c\} && d + 0 = \{d\} => (c + d) + 0 = c + d =  (c + d) union \{d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAd = setCandidates[i]; /* (c + d) union \{d\} */
						resAd.insert(4);
						if (resA0 != resAd) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[17]) {
						/* c + 0 = \{0, c\} && d + 0 = \{0, d\} => (c + d) + 0 = (c + d) union \{0, c\} = (c + d) union \{0, d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0c = setCandidates[i]; /* (c + d) union \{0, c\} */
						resA0c.insert(0);
						resA0c.insert(3);
	
						if (resA0 != resA0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resA0d = setCandidates[i]; /* (c + d) union \{0, d\} */
						resA0d.insert(0);
						resA0d.insert(4);
	
						if (resA0 != resA0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}

				}

#endif

#if (modelno == 10)
				/* a + d */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 4) {
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0d = setCandidates[i];
						for (set<int>::const_iterator it_0d=op[0][4]->begin(); it_0d != op[0][4]->end(); ++it_0d)
							resA_0d.insert(*it_0d);
						
						if (resA0 != resA_0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}
				else
				/* b + c */
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 3) {
					if (op[0][2] == &setCandidates[4]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0c = setCandidates[i];
						for (set<int>::const_iterator it_0c=op[0][3]->begin(); it_0c != op[0][3]->end(); ++it_0c)
							resA_0c.insert(*it_0c);
						
						if (resA0 != resA_0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}

#endif

#if (modelno == 9 || modelno == 11)
				/* a + c */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 3) {
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0c = setCandidates[i];
						for (set<int>::const_iterator it_0c=op[0][3]->begin(); it_0c != op[0][3]->end(); ++it_0c)
							resA_0c.insert(*it_0c);
						
						if (resA0 != resA_0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}
				else
				/* b + d */
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 4) {
					if (op[0][2] == &setCandidates[4]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0d = setCandidates[i];
						for (set<int>::const_iterator it_0d=op[0][4]->begin(); it_0d != op[0][4]->end(); ++it_0d)
							resA_0d.insert(*it_0d);
						
						if (resA0 != resA_0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}

#endif

#if (modelno == 12)
				/* a + c */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 3) {
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0c = setCandidates[i];
						for (set<int>::const_iterator it_0c=op[0][3]->begin(); it_0c != op[0][3]->end(); ++it_0c)
							resA_0c.insert(*it_0c);
						
						if (resA0 != resA_0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}
				else
				/* b + d */
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 4) {
					if (op[0][2] == &setCandidates[4] && op[0][4] == &setCandidates[16]) {
						/* b + 0 = \{b\} && d + 0 = \{d\} => (b + d) + 0 = (b + d) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[4] && op[0][4] == &setCandidates[17]) {
						/* b + 0 = \{b\} && d + 0 = \{0, d\} => (b + d) + 0 = (b + d)  = (b + d) union \{b\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						set<int> resAb = setCandidates[i];
						resAb.insert(2);

						if (resA0 != resAb) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

							
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5] && op[0][4] == &setCandidates[16]) {
						/* b + 0 = \{0, b\} && d + 0 = \{d\} => (b + d) + 0 = (b + d)  = (b + d) union \{d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						set<int> resAd = setCandidates[i];
						resAd.insert(4);

						if (resA0 != resAd) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

							
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5] && op[0][4] == &setCandidates[17]) {
						/* b + 0 = \{0, b\} && d + 0 = \{0, d\} => (b + d) + 0 = (b + d) union \{0, b\}  = (b + d) union \{0, d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0b = setCandidates[i];
						resA0b.insert(0);
						resA0b.insert(2);

						if (resA0 != resA0b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						set<int> resA0d = setCandidates[i];
						resA0d.insert(0);
						resA0d.insert(4);

						if (resA0 != resA0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
							
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}

#endif

#if (modelno == 13)
				/* a + c */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 3) {
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0c = setCandidates[i];
						for (set<int>::const_iterator it_0c=op[0][3]->begin(); it_0c != op[0][3]->end(); ++it_0c)
							resA_0c.insert(*it_0c);
						
						if (resA0 != resA_0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}
				else
				/* b + d */
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 4) {
					if (op[0][2] == &setCandidates[4]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0d = setCandidates[i];
						for (set<int>::const_iterator it_0d=op[0][4]->begin(); it_0d != op[0][4]->end(); ++it_0d)
							resA_0d.insert(*it_0d);
						
						if (resA0 != resA_0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}
#endif

#if (modelno == 14)
				/* a + c */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 3) {
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0c = setCandidates[i];
						for (set<int>::const_iterator it_0c=op[0][3]->begin(); it_0c != op[0][3]->end(); ++it_0c)
							resA_0c.insert(*it_0c);
						
						if (resA0 != resA_0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}
				else
				/* b + d */
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 4) {
					if (op[0][2] == &setCandidates[4] && op[0][4] == &setCandidates[16]) {
						/* b + 0 = \{b\} && d + 0 = \{d\} => (b + d) + 0 = (b + d) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[4] && op[0][4] == &setCandidates[17]) {
						/* b + 0 = \{b\} && d + 0 = \{0, d\} => (b + d) + 0 = (b + d)  = (b + d) union \{b\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						set<int> resAb = setCandidates[i];
						resAb.insert(2);

						if (resA0 != resAb) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

							
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5] && op[0][4] == &setCandidates[16]) {
						/* b + 0 = \{0, b\} && d + 0 = \{d\} => (b + d) + 0 = (b + d)  = (b + d) union \{d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						set<int> resAd = setCandidates[i];
						resAd.insert(4);

						if (resA0 != resAd) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

							
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5] && op[0][4] == &setCandidates[17]) {
						/* b + 0 = \{0, b\} && d + 0 = \{0, d\} => (b + d) + 0 = (b + d) union \{0, b\}  = (b + d) union \{0, d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0b = setCandidates[i];
						resA0b.insert(0);
						resA0b.insert(2);

						if (resA0 != resA0b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						set<int> resA0d = setCandidates[i];
						resA0d.insert(0);
						resA0d.insert(4);

						if (resA0 != resA0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
							
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}


#endif

#if (modelno == 15)
				/* c + c */
				if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 3) {

					if (op[0][3] == &setCandidates[8]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
					}
					else
					if (op[0][3] == &setCandidates[9]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0c = setCandidates[i]; /* (c + c) union \{0, c\} */
						resA0c.insert(0);
						resA0c.insert(3);


						if (resA0 != resA0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
			
				}
				else
				/* d + d */
				if (it_cand->candidate_pair.first == 4 && it_cand->candidate_pair.second == 4) {

					if (op[0][4] == &setCandidates[16]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
					}
					else
					if (op[0][4] == &setCandidates[17]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0d = setCandidates[i]; /* (d + d) union \{0, d\} */
						resA0d.insert(0);
						resA0d.insert(4);


						if (resA0 != resA0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
			
				}
				else
				/* a + b */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
					if (op[0][1] == &setCandidates[2]) {
						/* a + 0 = \{a\} => (a + b) + 0 = (a + b) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					
					}
					else 
					if (op[0][1] == &setCandidates[3]) {
						/* a + 0 = \{0, a\} => (a + b) + 0 = (a + b) union (0 + b) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> res_a0_b = setCandidates[i]; /* (a + b) union (0 + b) */
						for (set<int>::const_iterator it_0b = op[0][2]->begin(); it_0b != op[0][2]->end(); ++it_0b)
							res_a0_b.insert(*it_0b);

						if (resA0 != res_a0_b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}

				}

#endif

#if (modelno == 16)
				/* a + b */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
					if (op[0][1] == &setCandidates[2]) {
						/* a + 0 = \{a\} => (a + b) + 0 = (a + b) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					
					}
					else 
					if (op[0][1] == &setCandidates[3]) {
						/* a + 0 = \{0, a\} => (a + b) + 0 = (a + b) union (0 + b) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> res_a0_b = setCandidates[i]; /* (a + b) union (0 + b) */
						for (set<int>::const_iterator it_0b = op[0][2]->begin(); it_0b != op[0][2]->end(); ++it_0b)
							res_a0_b.insert(*it_0b);

						if (resA0 != res_a0_b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}

				}
				else
				/* c + d */
				if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
					if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[16]) {
						/* c + 0 = \{c\} && d + 0 = \{d\} => (c + d) + 0 = (c + d) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					
					}
					else 
					if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[17]) {
						/* c + 0 = \{c\} && d + 0 = \{0, d\} => (c + d) + 0 = c + d = (c + d) union \{c\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAc = setCandidates[i]; /* (c + d) union \{c\} */
						resAc.insert(3);
						if (resA0 != resAc) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[16]) {
						/* c + 0 = \{0, c\} && d + 0 = \{d\} => (c + d) + 0 = c + d =  (c + d) union \{d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAd = setCandidates[i]; /* (c + d) union \{d\} */
						resAd.insert(4);
						if (resA0 != resAd) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[17]) {
						/* c + 0 = \{0, c\} && d + 0 = \{0, d\} => (c + d) + 0 = (c + d) union \{0, c\} = (c + d) union \{0, d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0c = setCandidates[i]; /* (c + d) union \{0, c\} */
						resA0c.insert(0);
						resA0c.insert(3);
	
						if (resA0 != resA0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resA0d = setCandidates[i]; /* (c + d) union \{0, d\} */
						resA0d.insert(0);
						resA0d.insert(4);
	
						if (resA0 != resA0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}

				}
				
#endif

#if (modelno == 17)
				/* a + b */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
					/* a + 0 = \{a\} */

					/* a + b = A 

						A + 0 = A
						1 + 0 \subseteq A + 0
					*/

					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						if (op[0][2] == &setCandidates[5]) {
							set<int> resAa = setCandidates[i];
							resAa.insert(1);

							if (resA0 != resAa) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> res_a0_b = setCandidates[i]; /* (a + b) union (0 + b) */
						for (set<int>::const_iterator it_0b = op[0][2]->begin(); it_0b != op[0][2]->end(); ++it_0b)
							res_a0_b.insert(*it_0b);

						if (resA0 != res_a0_b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
		
						if (op[0][2] == &setCandidates[4]) {
							/* a  + b */	
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][2] == &setCandidates[5]) {
							/* (a + b) union \{0, a\} */

							set<int> res_a_b0 = setCandidates[i];
							res_a_b0.insert(0);
							res_a_b0.insert(1);
			
							if (resA0 != res_a_b0) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
					}
				}
				else
				/* c + d */
				if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
					if (op[0][3] == &setCandidates[8]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
//						if (op[0][2] == &setCandidates[5]) {
//							set<int> resAa = setCandidates[i];
//							resAa.insert(1);
//
//							if (resA0 != resAa) {
//								if (!was_cleared) {
//									clear_used_from(current_setup, level);
//									was_cleared = true;
//								}
//								continue;
//							}
//						
//						}

						if (op[0][4] == &setCandidates[17]) { /* d + 0 = \{0, d\} */
							/* (c + d) union \{c\} */
							set<int> resAc = setCandidates[i];
							resAc.insert(3);
							if (resA0 != resAc) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
							
						}
						else
						if (op[0][4] == &setCandidates[24]) { /* d + 0 = \{c, d\} */
							/* (c + d) union (c + c) */
							set<int> resA_cc = setCandidates[i];
							for (set<int>::const_iterator it_cc = op[3][3]->begin(); it_cc != op[3][3]->end(); ++it_cc)
								resA_cc.insert(*it_cc);

							if (resA0 != resA_cc) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
							
						}
						else
						if (op[0][4] == &setCandidates[24]) { /* d + 0 = \{0, c, d\} */
							/* (c + d) union (c + c) union \{c\} */

							set<int> resA_cd0 = setCandidates[i];
							resA_cd0.insert(3);
							for (set<int>::const_iterator it_cc = op[3][3]->begin(); it_cc != op[3][3]->end(); ++it_cc)
								resA_cd0.insert(*it_cc);

							if (resA0 != resA_cd0) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}

						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][3] == &setCandidates[9]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						/* (c + d) union (d + 0) */
						set<int> resA_0d = setCandidates[i];
						for (set<int>::const_iterator it_0d = op[0][4]->begin(); it_0d != op[0][4]->end(); ++it_0d)
							resA_0d.insert(*it_0d);
						if (resA0 != resA_0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						if (op[0][4]==&setCandidates[16]) {
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][4]==&setCandidates[17]) {
							set<int> resA0c = setCandidates[i];
							resA0c.insert(0);
							resA0c.insert(3);

							if (resA0 != resA0c) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][4]==&setCandidates[24]) {
							set<int> resA_cc = setCandidates[i];
							for (set<int>::const_iterator it_cc = op[3][3]->begin(); it_cc != op[3][3]->end(); ++it_cc)
								resA_cc.insert(*it_cc);

							if (resA0 != resA_cc) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
			
						}
						else
						if (op[0][4]==&setCandidates[25]) {
							set<int> resA0c_cc = setCandidates[i];
							resA0c_cc.insert(0);
							resA0c_cc.insert(3);
							for (set<int>::const_iterator it_cc = op[3][3]->begin(); it_cc != op[3][3]->end(); ++it_cc)
								resA0c_cc.insert(*it_cc);

							if (resA0 != resA0c_cc) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				}
				

#endif

#if (modelno == 18)
				/* a + a */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
					/* a + 0 = \{a\} 
						a + a = A
							A + 0 = A
							1 + 0 \subseteq A + 0

					*/
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					/* a + 0 = \{0, a\} 
						a + a = A
							A + 0 = A \union \{0, a\}
							1 + 0 \subseteq A + 0
					*/
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0a = setCandidates[i];
						resA0a.insert(0);
						resA0a.insert(1);
						
						if (resA0 != resA0a) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
						
				}

				/* b + b */
				else
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
					/* b + 0 = \{b\} 

						b + b = A;
							A + 0 = A
							1 + 0 \subseteq A + 0  = A
					*/
					
					if (op[0][2] == &setCandidates[4]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					/* b + 0 = \{0, b\} 

						b + b = A;
							A + 0 = A \union \{0, b\}
							1 + 0 \subseteq A + 0  
					
					*/
					if (op[0][2] == &setCandidates[5]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0b = setCandidates[i];
						resA0b.insert(0);
						resA0b.insert(2);
						
						if (resA0 != resA0b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}

				}

				/* c + d */
				else
				if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
					/* c + 0 = \{c\} && d + 0 = \{d\} 
						c + d = A
							A + 0 = A
							1 + 0 \subseteq A + 0 = A

					 */
					if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[16]) {
						
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					/* c + 0 = \{0, c\} && d + 0 = \{d\} 
						A \union {d} = A = A + 0
						1 + 0 \subseteq A + 0 = A

					*/
					if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[16]) {
						
						if ( !(i & 16) ) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
	
						}
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					/* c + 0 = \{c\} && d + 0 = \{0, d\} 

						A + 0 = A = A \union \{c\}
						1 + 0 \subseteq A + 0 = A
					*/
					if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[17]) {
						
						if ( !(i & 8) ) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
	
						}
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					/* c + 0 = \{0, c\} && d + 0 = \{0, d\} 
						a) A + 0 = A \union {0, c} = A \union \{0, d\}
						b) 1 + 0 \subseteq A + 0
					*/
					if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[17]) {
						
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0c = setCandidates[i];
						resA0c.insert(0);
						resA0c.insert(3);
						

						if (resA0 != resA0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resA0d = setCandidates[i];
						resA0d.insert(0);
						resA0d.insert(4);

						if (resA0 != resA0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				
				}

#endif

#if (modelno == 19)
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
						if (op[0][1] == &setCandidates[2] && op[0][2] == &setCandidates[4]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][1] == &setCandidates[2] && op[0][2] == &setCandidates[5]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
							set<int> resAa = setCandidates[i];
							resAa.insert(1);

							if (resA0 != resAa) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][1] == &setCandidates[3] && op[0][2] == &setCandidates[4]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							set<int> resAb = setCandidates[i];
							resAb.insert(2);

							if (resA0 != resAb) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][1] == &setCandidates[3] && op[0][2] == &setCandidates[5]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							set<int> resA0a = setCandidates[i];
							resA0a.insert(0);
							resA0a.insert(1);

							if (resA0 != resA0a) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							set<int> resA0b = setCandidates[i];
							resA0b.insert(0);
							resA0b.insert(2);

							if (resA0 != resA0b) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							int idxA0 = 0;
							for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
								idxA0 += (1 << (*itA0));
							if ((idx01 & idxA0) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
					}
					else
					if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
						if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[16]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[17]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
							set<int> resAc = setCandidates[i];
							resAc.insert(3);

							if (resA0 != resAc) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[16]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
							set<int> resAd = setCandidates[i];
							resAd.insert(4);

							if (resA0 != resAd) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[17]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							set<int> resA0c = setCandidates[i];
							resA0c.insert(0);
							resA0c.insert(3);

							if (resA0 != resA0c) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							set<int> resA0d = setCandidates[i];
							resA0d.insert(0);
							resA0d.insert(4);

							if (resA0 != resA0d) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							int idxA0 = 0;
							for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
								idxA0 += (1 << (*itA0));
							if ((idx01 & idxA0) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
					}
#endif

#if (modelno == 22)
				/* a + a */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
					/* a + 0 = \{a\} 
						a + a = A
							A + 0 = A
							1 + 0 \subseteq A + 0

					*/
					if (op[0][1] == &setCandidates[2]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					/* a + 0 = \{0, a\} 
						a + a = A
							A + 0 = A \union \{0, a\}
							1 + 0 \subseteq A + 0
					*/
					if (op[0][1] == &setCandidates[3]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0a = setCandidates[i];
						resA0a.insert(0);
						resA0a.insert(1);
						
						if (resA0 != resA0a) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
						
				}

				/* b + b */
				else
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
					/* b + 0 = \{b\} 

						b + b = A;
							A + 0 = A
							1 + 0 \subseteq A + 0  = A
					*/
					
					if (op[0][2] == &setCandidates[4]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					/* b + 0 = \{0, b\} 

						b + b = A;
							A + 0 = A \union \{0, b\}
							1 + 0 \subseteq A + 0  
					
					*/
					if (op[0][2] == &setCandidates[5]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0b = setCandidates[i];
						resA0b.insert(0);
						resA0b.insert(2);
						
						if (resA0 != resA0b) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}

				}
				

				/* c + d */
				else
				if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
					/* c + 0 = \{c\} && d + 0 = \{d\} 
						c + d = A
							A + 0 = A
							1 + 0 \subseteq A + 0 = A

					 */
					if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[16]) {
						
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					/* c + 0 = \{0, c\} && d + 0 = \{d\} 
						A \union {d} = A = A + 0
						1 + 0 \subseteq A + 0 = A

					*/
					if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[16]) {
						
						if ( !(i & 16) ) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
	
						}
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					/* c + 0 = \{c\} && d + 0 = \{0, d\} 

						A + 0 = A = A \union \{c\}
						1 + 0 \subseteq A + 0 = A
					*/
					if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[17]) {
						
						if ( !(i & 8) ) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
	
						}
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					/* c + 0 = \{0, c\} && d + 0 = \{0, d\} 
						a) A + 0 = A \union {0, c} = A \union \{0, d\}
						b) 1 + 0 \subseteq A + 0
					*/
					if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[17]) {
						
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0c = setCandidates[i];
						resA0c.insert(0);
						resA0c.insert(3);
						

						if (resA0 != resA0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resA0d = setCandidates[i];
						resA0d.insert(0);
						resA0d.insert(4);

						if (resA0 != resA0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					
						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
				
				}

#endif

#if (modelno == 23)
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 3) {
						if (op[0][1] == &setCandidates[2] && op[0][3] == &setCandidates[8]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][1] == &setCandidates[2] && op[0][3] == &setCandidates[9]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
							set<int> resAa = setCandidates[i];
							resAa.insert(1);

							if (resA0 != resAa) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][1] == &setCandidates[3] && op[0][3] == &setCandidates[8]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							set<int> resAc = setCandidates[i];
							resAc.insert(3);

							if (resA0 != resAc) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][1] == &setCandidates[3] && op[0][3] == &setCandidates[9]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							set<int> resA0a = setCandidates[i];
							resA0a.insert(0);
							resA0a.insert(1);

							if (resA0 != resA0a) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							set<int> resA0c = setCandidates[i];
							resA0c.insert(0);
							resA0c.insert(3);

							if (resA0 != resA0c) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							int idxA0 = 0;
							for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
								idxA0 += (1 << (*itA0));
							if ((idx01 & idxA0) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
					}
					else
					if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 4) {
						if (op[0][2] == &setCandidates[4] && op[0][4] == &setCandidates[16]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][2] == &setCandidates[4] && op[0][4] == &setCandidates[17]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
							set<int> resAb = setCandidates[i];
							resAb.insert(2);

							if (resA0 != resAb) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][2] == &setCandidates[5] && op[0][4] == &setCandidates[16]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
							set<int> resAd = setCandidates[i];
							resAd.insert(4);

							if (resA0 != resAd) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						if (op[0][2] == &setCandidates[5] && op[0][4] == &setCandidates[17]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							set<int> resA0b = setCandidates[i];
							resA0b.insert(0);
							resA0b.insert(2);

							if (resA0 != resA0b) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							set<int> resA0d = setCandidates[i];
							resA0d.insert(0);
							resA0d.insert(4);

							if (resA0 != resA0d) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							int idxA0 = 0;
							for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
								idxA0 += (1 << (*itA0));
							if ((idx01 & idxA0) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
					}
#endif


#if (modelno == 24) 
					/* a + 0 = {0, a} */
					/* a + a = A, potom 
							1) {0, a} union A = A + 0 
							2) 1 + 0 \subseteq A + 0 = {0, a} union A
					*/
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
						if (op[0][1] == &setCandidates[2]) {
							/* setCandidate[i] + 0 = setCandidate[i] */

							/* compute (setCandidates[i]) + 0 */
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);

							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

						}

						if (op[0][1] == &setCandidates[3]) {
							/* compute setCandidates[i] union {0, a} */
							set<int> resAa0 = setCandidates[i];
							resAa0.insert(0); // insert element 0
							resAa0.insert(1); // insert element a

							/* compute (setCandidates[i]) + 0 */
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);

							if (resAa0 != resA0) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
					}
					else
					if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
						/* c + 0 = \{c\} && d + 0 = \{d\} 
							c + d = A
								A + 0 = A
								1 + 0 \subseteq A + 0 = A

						 */
						if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[16]) {
							
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
							
							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						/* c + 0 = \{0, c\} && d + 0 = \{d\} 
							A \union {d} = A = A + 0
							1 + 0 \subseteq A + 0 = A

						*/
						if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[16]) {
							
							if ( !(i & 16) ) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
		
							}
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
							
							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else
						/* c + 0 = \{c\} && d + 0 = \{0, d\} 

							A + 0 = A = A \union \{c\}
							1 + 0 \subseteq A + 0 = A
						*/
						if (op[0][3] == &setCandidates[8] && op[0][4] == &setCandidates[17]) {
							
							if ( !(i & 8) ) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
		
							}
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
							
							int idx01 = op[0][n - 1] - &setCandidates[0];
							if ((idx01 & i) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
						else 
						/* c + 0 = \{0, c\} && d + 0 = \{0, d\} 
							a) A + 0 = A \union {0, c} = A \union \{0, d\}
							b) 1 + 0 \subseteq A + 0
						*/
						if (op[0][3] == &setCandidates[9] && op[0][4] == &setCandidates[17]) {
							
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							set<int> resA0c = setCandidates[i];
							resA0c.insert(0);
							resA0c.insert(3);
							

							if (resA0 != resA0c) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							set<int> resA0d = setCandidates[i];
							resA0d.insert(0);
							resA0d.insert(4);

							if (resA0 != resA0d) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						
							int idx01 = op[0][n - 1] - &setCandidates[0];
							int idxA0 = 0;
							for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
								idxA0 += (1 << (*itA0));
							if ((idx01 & idxA0) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
				
				}

#endif

#if (modelno == 25)
					/* a + d */
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 4) {
						if (op[0][1] == &setCandidates[2]) {
							/* setCandidate[i] + 0 = setCandidate[i] */

							/* compute (setCandidates[i]) + 0 */
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);

							if (resA0 != setCandidates[i]) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

						}
						else
						if (op[0][1] == &setCandidates[3]) {
							set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
							set<int> resA_0d = setCandidates[i];
							for (set<int>::const_iterator it_0d=op[0][4]->begin(); it_0d != op[0][4]->end(); ++it_0d)
								resA_0d.insert(*it_0d);
							
							if (resA0 != resA_0d) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}

							int idx01 = op[0][n - 1] - &setCandidates[0];
							int idxA0 = 0;
							for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
								idxA0 += (1 << (*itA0));
							if ((idx01 & idxA0) != idx01) {
								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}
								continue;
							}
						}
					}


#endif

#if (modelno == 26)
				/* a + d */
				if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 4) {
					if (op[0][1] == &setCandidates[2] && op[0][4] == &setCandidates[16]) {
						/* a + 0 = \{a\} && d + 0 = \{d\} => (a + d) + 0 = (a + d) */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					
					}
					else 
					if (op[0][1] == &setCandidates[2] && op[0][4] == &setCandidates[17]) {
						/* a + 0 = \{a\} && d + 0 = \{0, d\} => (a + d) + 0 = a + d = (a + d) union \{a\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAa = setCandidates[i]; /* (a + d) union \{a\} */
						resAa.insert(1);
						if (resA0 != resAa) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][1] == &setCandidates[3] && op[0][4] == &setCandidates[16]) {
						/* a + 0 = \{0, a\} && d + 0 = \{d\} => (a + d) + 0 = a + d =  (a + d) union \{d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resAd = setCandidates[i]; /* (a + d) union \{d\} */
						resAd.insert(4);
						if (resA0 != resAd) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else 
					if (op[0][1] == &setCandidates[3] && op[0][4] == &setCandidates[17]) {
						/* a + 0 = \{0, a\} && d + 0 = \{0, d\} => (a + d) + 0 = (a + d) union \{0, a\} = (a + d) union \{0, d\} */
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA0a = setCandidates[i]; /* (a + d) union \{0, a\} */
						resA0a.insert(0);
						resA0a.insert(1);
	
						if (resA0 != resA0a) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						set<int> resA0d = setCandidates[i]; /* (a + d) union \{0, d\} */
						resA0d.insert(0);
						resA0d.insert(4);
	
						if (resA0 != resA0d) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}

				}
				else
				/* b + c */
				if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 3) {
					if (op[0][2] == &setCandidates[4]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						if (resA0 != setCandidates[i]) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
						
						int idx01 = op[0][n - 1] - &setCandidates[0];
						if ((idx01 & i) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					else
					if (op[0][2] == &setCandidates[5]) {
						set<int> resA0 = hea::compute_sum(op, 0, setCandidates[i]);
						set<int> resA_0c = setCandidates[i];
						for (set<int>::const_iterator it_0c=op[0][3]->begin(); it_0c != op[0][3]->end(); ++it_0c)
							resA_0c.insert(*it_0c);
						
						if (resA0 != resA_0c) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

						int idx01 = op[0][n - 1] - &setCandidates[0];
						int idxA0 = 0;
						for (set<int>::const_iterator itA0 = resA0.begin(); itA0 != resA0.end(); ++itA0) 
							idxA0 += (1 << (*itA0));
						if ((idx01 & idxA0) != idx01) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
	
				}
					
#endif

					/* x = y', then 0 + 1 \subseteq x + y, if 0+x = {x}, or 0+y = {y} ! */


					if (	op[0][it_cand->candidate_pair.first]->size() == 1 || 
						((it_cand->candidate_pair.first != it_cand->candidate_pair.second) && (op[0][it_cand->candidate_pair.second]->size() == 1)) /* || 
						is_0_plus_set_equal_to_set(op, i, setCandidates)*/

					) {

						bool is_ok = true;
						for (set<int>::const_iterator it_01set = op[0][n - 1]->begin(); it_01set != op[0][n - 1]->end(); ++it_01set)
							if ( !(i & (1 << (*it_01set)))) {
								is_ok = false;
								break;
							}
						/* -- */

						if (!is_ok) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}

					}
					


					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

					if (!check_summability(candidates, it_cand->candidate_pair, op)) {

						if (!was_cleared) {
							clear_used_from(current_setup, level);
							was_cleared = true;
						}
					
						continue;
					}
#if 0
					bool valid = true;
					if (last_cand_for_x.find(it_cand->candidate_pair) != last_cand_for_x.end()) {

						// x is summable with all elements in y + 0, if x + y is summable
						for (int i = it_cand->candidate_pair.first; i < n - 1 && valid; i++)
							if (hea::is_summable(op, it_cand->candidate_pair.first, i))
								for (set<int>::const_iterator it = op[0][i]->begin(); it != op[0][i]->end(); ++it)
									if (!hea::is_summable(op, it_cand->candidate_pair.first, *it)) {
										valid = false;
										break;
									}
					}

					if (!valid)
						continue;
#endif

					set<Candidate>::const_iterator it_new = it_cand;
					it_new++;
					solve_candidates_state_setup(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, current_setup);

					// unset
					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
				}

			}
			else {
				/* cannot contain 1, or be equal to \{0\}, or be empty */

				int from;
				bool was_cleared = true;
				if (current_setup.bFromState && current_setup.used_from_state[level]) {
					current_setup.used_from_state[level] = false;
					from = current_setup.from_state[level];
					was_cleared = false;
				}
				else
					from = 2;

				int to;
				if (current_setup.bToState && current_setup.used_to_state[level]) {
					current_setup.used_to_state[level] = false;
					to = current_setup.to_state[level];
				}
				else
					to = (1 << (n - 1)) - 1;

				for (int i = from; i <= to; i++) {

					// skip x + y = {x}, {0, x}, {y}, {0, y}
					if (i == (1<<it_cand->candidate_pair.first) || i == 1 + (1<<it_cand->candidate_pair.first) || i == (1<<it_cand->candidate_pair.second) || i == 1 + (1<<it_cand->candidate_pair.second)) {
						if (!was_cleared) {
							clear_used_from(current_setup, level);
							was_cleared = true;
						}

						continue;
					}

#if (modelno == 4 || modelno == 5)
			/* c + c */
			if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 3) {
				if (i < 6) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}

					continue;
				}
			}
#endif

#if (modelno == 6)
			/* b + d */
			if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 4) {
				/* not subset 0a, 0cd, 0b (triv), 0d (triv)*/
				/* ab, 0ab, ac, 0ac, bc, 0bc, abc, 0abc, ad, 0ad, bd, 0bd, abd, 0abd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
				/*  6,   7, 10,  11, 12,  13,  14,   15, 18,  19, 20,  21,  22,   23,  26,   27,  28,   29,   30,    31 */
				if (i < 6 || i == 8 || i == 9 /* || i == 16 || i == 17 */ || i == 24 || i == 25 ) {
						if (!was_cleared) {
							clear_used_from(current_setup, level);
							was_cleared = true;
						}

						continue;
				}
				
			}
			
#endif

#if (modelno == 7)
			if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
				/* not subset of 0a, 0b=triv */
				if (i < 4) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
			else
			if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 3) {
				/* not subset of 0a, 0bc, 0b (triv) */
				/* ab, 0ab, ac, 0ac, abc, 0abc,  d, 0d, ad, 0ad, bd, 0bd, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
				/*  6,   7, 10,  11,  14,   15, 16, 17, 18,  19, 20,  21,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31 */

				if (i < 6 || i == 8 || i == 9 || i==12 || i == 13) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
			
#endif


#if (modelno == 8)
			if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
				/* not subset of 0a, 0bd, 0cd, 0bc */
				/* ab, 0ab, ac, 0ac, abc, 0abc, ad, 0ad, abd, 0abd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
				/*  6,   7, 10,  11,  14,   15, 18,  19,  22,   23,  26,   27,  28,   29,   30,    31 */

				if (i < 6 || i == 8 || i == 9 || i == 12 || i == 13 || i == 16 || i == 17 || i == 20 || i == 21 || i == 24 || i == 25) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
			else
			if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 3) {
				/* not subset of 0a, 0bd, 0bc */
				/* ab, 0ab, ac, 0ac, abc, 0abc, ad, 0ad, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
				/*  6,   7, 10,  11,  14,   15, 18,  19,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31 */

				if (i < 6 || i == 8 || i == 9 || i == 12 || i == 13 || i == 16 || i == 17 || i == 20 || i == 21) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}

		
#endif

#if (modelno == 9)
			if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
				/* ab, 0ab, ac, 0ac, abc, 0abc, bd, 0bd, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
				/*  6,   7, 10,  11,  14,   15, 20,  21,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31 */

				if (i < 6 || i == 8 || i == 9 || i == 12 || i == 13 || (i > 15 && i < 20)) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
#endif

#if (modelno == 10)
			if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
				/* not subset of 0bc, 0b, 0a */
				/* ab, 0ab, ac, 0ac, abc, 0abc,  d, 0d, ad, 0ad, bd, 0bd, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
				/*  6,   7, 10,  11,  14,   15, 16, 17, 18,  19, 20,  21,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31 */
				if (i < 6 || i == 8 || i == 9 || i == 12 || i == 13) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
			else
			if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
				/* not subset of 0ad, 0b, 0a */
				/* ab, 0ab, c, 0c, ac, 0ac, bc, 0bc, abc, 0abc, bd, 0bd, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
				/*  6,   7, 8,  9, 10,  11, 12,  13,  14,   15, 20,  21,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31 */
				if (i < 6 || (i>15 && i < 20)) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
#endif

#if (modelno == 11)
			if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
				/* ab, 0ab, ac, 0ac, abc, 0abc,  d, 0d, ad, 0ad, bd, 0bd, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
				/*  6,   7, 10,  11,  14,   15, 16, 17, 18,  19, 20,  21,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31 */
				if ((i < 6) || (i == 8) || (i == 9) || (i == 12) || (i == 13)) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
			else
			if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
				if (i == 2 || i == 3) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
#endif

#if (modelno == 12)
			if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
				/* ab, 0ab, ac, 0ac, abc, 0abc, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
				/*  6,   7, 10,  11,  14,   15,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31 */
				if ((i < 6) || (i == 8) || (i == 9) || (i == 12) || (i == 13) || (i > 15 && i < 22)) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
			else
			if (it_cand->candidate_pair.first == 4 && it_cand->candidate_pair.second == 4) {
				/* ab, 0ab, ac, 0ac, abc, 0abc, bd, 0bd, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
				/*  6,   7, 10,  11,  14,   15, 20,  21,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31 */
				if ((i < 6) || (i == 8) || (i == 9) || (i == 12) || (i == 13) || (i > 15 && i < 20)) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}

#endif

#if (modelno == 13)
			/* a + a */
			if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
				/* not subset of 0a (triv), 0b */
				if (i == 4 || i == 5) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
			else 
			/* b + b */
			if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
				/* not subset of 0a, 0b (triv) */
				if (i == 2 || i == 3) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
			
#endif

#if (modelno == 14)
			/* a + a */
			if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
			/* ab, 0ab, c, 0c, ac, 0ac, bc, 0bc, abc, 0abc, ad, 0ad, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
			/*  6,   7, 8,  9, 10,  11, 12,  13,  14,   15, 18,  19,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31 */
				if ((i < 5) || (i == 16) || (i == 17) || (i == 20) || (i == 21)) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
			else 
			/* a + b */
			if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
				/* ab, 0ab, c, 0c, ac, 0ac, bc, 0bc, abc, 0abc, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
				/*  6,   7, 8,  9, 10,  11, 12,  13,  14,   15,  22,   23, 24,  25,  26    27,  28,   29,   30,    31 */
				if ((i < 5) || (i > 15 && i < 22)) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}
			else 
			/* d + d */
			if (it_cand->candidate_pair.first == 4 && it_cand->candidate_pair.second == 4) {
				/* bd, 0bd, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
				/* 20,  21,  22,   23, 24,  25,  26,   27,  28,   39,   30,    31 */
				if (i < 20) {
					if (!was_cleared) {
						clear_used_from(current_setup, level);
						was_cleared = true;
					}
					continue;
				}
			}

#endif

#if (modelno == 15 || modelno == 16)
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 3) {
						if (op[0][1] == &setCandidates[2]) {

							if ( (i != 6) && (i != 7) && (i != 14) && (i != 15) && (i != 22) && (i != 23) &&
								(i != 26) && (i != 27) && (i != 30) && (i != 31) ) {

								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}

								continue;
							}
						}
						else
						if (op[0][1] == &setCandidates[3]) {
							if (op[0][3] == &setCandidates[8]) {
								/* c in */
								if ( (i != 14) && (i != 15) &&
									(i != 26) && (i != 27) && (i != 30) && (i != 31) ) {

									if (!was_cleared) {
										clear_used_from(current_setup, level);
										was_cleared = true;
									}

									continue;
								}
							}
							else
							if (op[0][3] == &setCandidates[9]) {
								if ( (i != 6) && (i != 7) && (i != 14) && (i != 15) && (i != 22) && (i != 23) &&
								(i != 26) && (i != 27) && (i != 30) && (i != 31) ) {

									if (!was_cleared) {
										clear_used_from(current_setup, level);
										was_cleared = true;
									}

									continue;
								}
							}
						}
					}
					else
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 4) {
						if (op[0][1] == &setCandidates[2]) {

							if ( (i != 6) && (i != 7) && (i != 14) && (i != 15) && (i != 22) && (i != 23) &&
								(i != 26) && (i != 27) && (i != 30) && (i != 31) ) {

								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}

								continue;
							}
						}
						else
						if (op[0][1] == &setCandidates[3]) {
							if (op[0][4] == &setCandidates[16]) {
								/* d in */
								if ( (i != 22) && (i != 23) &&
									(i != 26) && (i != 27) && (i != 30) && (i != 31) ) {

									if (!was_cleared) {
										clear_used_from(current_setup, level);
										was_cleared = true;
									}

									continue;
								}
							}
							else
							if (op[0][4] == &setCandidates[17]) {
								
								if ( (i != 6) && (i != 7) && (i != 14) && (i != 15) && (i != 22) && (i != 23) &&
									(i != 26) && (i != 27) && (i != 30) && (i != 31) ) {

									if (!was_cleared) {
										clear_used_from(current_setup, level);
										was_cleared = true;
									}

									continue;
								}
							}
						}
					}

#endif

#if (modelno == 17)
					if ((it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 3)) {

						/* a + 0 = \{a\} */
						if (op[0][1] == &setCandidates[2]) {
							/* ab, 0ab, abc, 0abc, ad, 0ad, abd, 0abd, acd, 0acd, abcd, 0abcd 
							    6,   7,  14,   15, 18,  19,  22,   23,  26,   27,   30,    31
							*/

							if ( (i != 6) && (i != 7) && (i != 14) && (i != 15) &&
							     (i != 18) && (i != 19) && (i != 22) && (i != 23) &&
							     (i != 26) && (i != 27) && (i != 30) && (i != 31) ) {

								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}

								continue;
							}
						}
						else
						/* a + 0 = \{0, a\} */
						if (op[0][1] == &setCandidates[3]) {
						
							/* abc, 0abc, acd, 0acd, abcd, 0abcd 
							    14,   15,  26,   27,   30,    31
							*/

							if ( (i != 14) && (i != 15) &&
							     (i != 26) && (i != 27) && (i != 30) && (i != 31) ) {

								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}

								continue;
							}
						}

					}
					else
					if ((it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 4)) {

						/* ab, 0ab, abc, 0abc, abd, 0abd, abcd, 0abcd 
						    6,   7,  14,   15,  22,   23,   30,    31
						*/

						if ( (i !=  6) && (i !=  7) && (i != 14) && (i != 15) &&
						     (i != 22) && (i != 23) && (i != 30) && (i != 31) ) {

								if (!was_cleared) {
									clear_used_from(current_setup, level);
									was_cleared = true;
								}

								continue;
						}
					}
					else
					/* c + c */
					if ((it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 3)) {
						if ( (i == 2) || (i == 3) || (i == 8) || (i == 9) || (i == 10) || (i == 11) ) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}
					

#endif


#if (modelno == 18 || modelno == 19) 
					if ((it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 4) ||
						(it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 4)) {

						/* abd, 0abd, acd, 0acd, bcd, 0bcd, abcd, 0abcd 
						    22,   23,  26,   27,  28,   29,   30,    31
						*/
						if ((i < 26) && (i != 22) && (i != 23)) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
							
						}
					}
					else
					if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 3) {
						/* abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd 
						    22,   23, 24,  25,  26,   27,  28,   29,   30,    31
						*/

						if (i < 22) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
							
						}
					}
#endif



#if (modelno == 22)
					if ((it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 4) ||
						(it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 3)) {	
						/* ab, 0ab, abc, 0abc, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd 
						    6,   7,  14,   15,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31
						*/

						if ((i < 22) && (i != 6) && (i != 7) && (i != 14) && (i != 15)) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
							
						}

						

					}
#endif

#if (modelno == 23)
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {
						/* ab, 0ab, ac, 0ac, abc, 0abc, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd 
						    6,   7, 10,  11,  14,   15,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31
						*/
						if ((i < 22) && (i != 6) && (i != 7) && (i != 10) && (i != 11) && (i != 14) && (i != 15)) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}
					else
					if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
						/* ab, 0ab, abc, 0abc, bd, 0bd, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd 
						    6,   7,  14,   15, 20,  21,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31
						*/

						if ((i < 20) && (i != 6) && (i != 7) && (i != 14) && (i != 15)) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}
					else
					if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 4) {
						/* ab, 0ab, abc, 0abc, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd 
						    6,   7,  14,   15,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31
						*/

						if ((i < 22) && (i != 6) && (i != 7) && (i != 14) && (i != 15)) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}

#endif


#if (modelno == 24)

					/* Case 24: 
						Adding specific checks:
OLD:

						1. a in (a + d) AND b or d in (a + d)
						2. a or b or c or d in (b + c) XXXX
						3. b or c or d in (c + c)
NEW:
						...
				

					*/

					/* a + d */
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 4) {
#if 0
						/* OLD */
						if ( !(i & 2) )  {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}

						if ( !(i & 4) && !(i & 16) ) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
#endif
						/* a + d not subset of 0ad, 0bcd, 0ac */
						/* ab, 0ab, abc, 0abc, abd, 0abd, acd, 0acd, abcd, 0abcd */
						/*  6,   7,  14,   15,  22,   23,  26,   27,   30,    31 */ 

						if ((i != 6) && (i != 7) && (i != 14) && (i != 15) && (i != 22) &&
							(i != 23) && (i != 26) && (i != 27) && (i != 30) && (i != 31)) {

							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}

					}
					else
					/* b + c */
					if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 3) {
#if 0
						if ( !(i & 2) && !(i & 4) && !(i & 8) && !(i & 16) )  {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
#endif
						/* b + c not subset of 0ad 0bc 0ac */
						/* ab, 0ab, abc, 0abc, bd, 0bd, abd, 0abd, cd, 0cd, acd, 0acd, bcd, 0bcd, abcd, 0abcd */
						/*  6,   7,  14,   15, 20,  21,  22,   23, 24,  25,  26,   27,  28,   29,   30,    31 */
						if ((i < 6) || (i > 7 && i < 14) || (i>16 && i < 20)) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}
					else
					/* c + c */
					if (it_cand->candidate_pair.first == 3 && it_cand->candidate_pair.second == 3) {
#if 0
						if ( !(i & 4) && !(i & 8) && !(i & 16) ) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
#endif
						/* c + c not subset of 0ad 0ac */
						if (((i & 19) == i) || ((i & 11) == i)) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}
					

					/* 
						---------------------------
						end Case 24 specific checks
						--------------------------- 

					*/
#endif

#if (modelno == 25)
					/* Case 25: 
						Adding specific checks:

						1. a + a \subseteq {0, a, b}
						2. a + b \subseteq {0, a, b, c}
						3. d \in a + c
						4. d \in b + b

					*/
					
					/* a + a */
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 1) {

						if (i > 7) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							return;
						}
					
					}	
					else
					/* a + b */
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {

						if (i > 15) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							return;
						}
					
					}	
					else
					/* a + c */
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 3) {

						if ( !(i & 16) ) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}	
					else
					/* b + b */
					if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {

						if ( !(i & 16) ) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}
							continue;
						}
					}	

					/* 
						---------------------------
						end Case 25 specific checks
						--------------------------- 

					*/

#endif

#if (modelno == 26)
					/* Case 26

						a + 0 = a
						|
						  a + b = ac, 0ac, abc,0abc,abd,0abd, acd, 0acd, abcd,0abcd
						        = 10, 11,   14,  15, 22,  23,  26,   27,   30,   31

						a + 0 = 0a
						|
						  a + b = 14, 15, 22, 23, 30, 31	


					 */

					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 2) {
						if ( (op[0][1] == &setCandidates[2] && i != 10 && i != 11 && i != 14 && 
					              i != 15 && i != 22 && i != 23 && i != 26 && i != 27 && i != 30 && i != 31) ||
						     (op[0][1] == &setCandidates[3] && i != 14 && i != 15 && i != 22 && 
							i != 23 && i != 30 && i != 31)
						   ) {

							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}
					else
					/*
						a + 0 = a
						|
						  a + c = abd,0abd, acd, 0acd, abcd,0abcd
						        =  22,  23,  26,   27,   30,   31

						a + 0 = 0a
						|
						  a + c = 26, 27, 30, 31	
					
					 */					
					if (it_cand->candidate_pair.first == 1 && it_cand->candidate_pair.second == 3) {
						if ( (op[0][1] == &setCandidates[2] && i != 22 && i != 23 &&
						    i != 26 && i != 27 && i != 30 && i != 31) ||
						     (op[0][1] == &setCandidates[3] && i != 26 && i != 27 && i != 30 && i != 31)
						   ) {

							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}
					else
					if (it_cand->candidate_pair.first == 2 && it_cand->candidate_pair.second == 2) {
						/* b + b not subset of 0ab 0ad */

						if ( ((i & 7) == i) || ((i & 19) == i) ) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}
					else
					if (it_cand->candidate_pair.first == 4 && it_cand->candidate_pair.second == 4) {
						/* d + d not subset of 0bcd, 0abc, 0ab (contained in 0abc) */

						if ( (i < 16) || ((i & 29) == i) ) {
							if (!was_cleared) {
								clear_used_from(current_setup, level);
								was_cleared = true;
							}

							continue;
						}
					}

					

					/* 
						---------------------------
						end Case 26 specific checks
						--------------------------- 

					*/

#endif



					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[i];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[i];

					if (!check_summability(candidates, it_cand->candidate_pair, op))
						continue;

#if 0
					bool valid = true;
					if (last_cand_for_x.find(it_cand->candidate_pair) != last_cand_for_x.end()) {
						// x is summable with all elements in y + 0, if x + y is summable
						for (int i = it_cand->candidate_pair.first; i < n - 1 && valid; i++)
							if (hea::is_summable(op, it_cand->candidate_pair.first, i))
								for (set<int>::const_iterator it = op[0][i]->begin(); it != op[0][i]->end(); ++it)
									if (!hea::is_summable(op, it_cand->candidate_pair.first, *it)) {
										valid = false;
										break;
									}

					}
					
					if (!valid)
						continue;

#endif

					set<Candidate>::const_iterator it_new = it_cand;
					it_new++;
					solve_candidates_state_setup(it_new, level + 1, candidates, last_cand_for_x, n, op, complement, setCandidates, current_setup);

					// unset
					op[it_cand->candidate_pair.first][it_cand->candidate_pair.second] = &setCandidates[0];
					if (it_cand->candidate_pair.first != it_cand->candidate_pair.second)
						op[it_cand->candidate_pair.second][it_cand->candidate_pair.first] = &setCandidates[0];
				}
			}

		}
	}
}



void init_last_cand_for_x_set(const int n, const set<Candidate> &candidates, set<pair<int, int> > &last_cand_for_x) {
	set<Candidate>::const_iterator it_next = candidates.begin();
	for (set<Candidate>::const_iterator it_cand = candidates.begin(); it_cand != candidates.end(); ++it_cand) {
		++it_next;
		
		if (it_next == candidates.end() ||
			it_cand->candidate_pair.first < it_next->candidate_pair.first)
			last_cand_for_x.insert(it_cand->candidate_pair);	

	}
}

void gen_rel(int x, int y, const int n, int (&relation)[MAX][MAX], vector<Relation> &relations) {
	if (x == n - 1) {

		if (admits_reverse_of_relation(n, relation)) {

			if (!is_isomorphic(n, relation, relations)) {

				/* disp relation */

				for (int i = 0; i < n; i++) {
					for (int j = 0; j < n; j++)
						cerr << relation[i][j];
					cerr << endl;
				}
				cerr << "-R-"<<endl;

				relations.push_back(Relation(n, relation));

				{

//					vector< pair< set< pair<int,int> >, set< pair<int,int> > > > HHHH;
					vector<Zoznamy> HHHH;
					vector<hea::Permutation> IIII;


					vector<hea::Permutation> vec_ortho;

					get_vector_of_orthosupplement_candidates(n, relation, vec_ortho);

					for (vector<hea::Permutation>::const_iterator it = vec_ortho.begin(); it != vec_ortho.end(); ++it) {

					set<pair<int, int> > one_pairs;

					for (int i = 0; i < n; i++) {
						one_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[i]));
						if (i != it->permutation[i])
							one_pairs.insert(pair<int, int>(it->permutation[i] + 1, 1 + i));

					}

/*
					if (one_pairs.size() == (unsigned int) n) {
*/

					set<pair<int, int> > other_pairs;
					
					for (int i = 0; i < n; i++)
					for (int j = 0; j < n; j++)
						if (i != j && relation[i][j]) {

							other_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[j]));
							if (i != it->permutation[j])

								other_pairs.insert(pair<int, int>(1 + it->permutation[j], i + 1));
						}




					if (!is_similar_vec_pairs(n, HHHH, one_pairs, other_pairs)) {
					
						IIII.push_back(hea::Permutation(*it));
						HHHH.push_back(Zoznamy(one_pairs, other_pairs));

					}

				}
/*
			cout << "IIII size: " << IIII.size() << endl;
			cout << "HHHH size: " << HHHH.size() << endl;
*/
			/* display candidates */
			cout << "----------------------------" << endl;
			vector<hea::Permutation>::const_iterator it_perm = IIII.begin();
			vector<Zoznamy>::const_iterator it_zozn = HHHH.begin();
			while (it_perm != IIII.end()) {
				cout << "Candidate: ";
				for (int i = 0; i < n; i++)
					cout << 1 + it_perm->permutation[i]; 
				cout << endl;

				cout << "Candidates for sets containing 1: " ;


				for (set<pair<int, int> >::const_iterator it_set = it_zozn->one_pairs.begin(); it_set != it_zozn->one_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;

				cout << "Candidates for other summable elements: " ;
				for (set<pair<int, int> >::const_iterator it_set = it_zozn->other_pairs.begin(); it_set != it_zozn->other_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;



				Zoznamy reduced;
				reduce_summable_pairs(*it_zozn, reduced);

				set<Candidate> candidates;

				for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
					candidates.insert(Candidate(*it_set, true));

				for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 

					candidates.insert(Candidate(*it_set, false));

				/* insert pairs (0, x) into candidates */
				for (int i = 1; i <=n; i++)
					candidates.insert(Candidate(pair<int,int>(0, i), false));
				candidates.insert(Candidate(pair<int, int>(0, n + 1), true));


				/* display candidates */

				cout << "Candidates: ";

				for (set<Candidate>::const_iterator it_cand = candidates.begin(); it_cand != candidates.end(); ++it_cand) {
					
					cout << "(" << it_cand->candidate_pair.first << ", " << it_cand->candidate_pair.second << ")";
					if (it_cand->is_one)
						cout << "*";

					cout << ", ";

				}
				cout << endl;
				

				set<pair<int, int> > last_of_x;
				init_last_cand_for_x_set(n, candidates, last_of_x); 
				cout << "Last of x pairs: " ;
				
				for (set<pair<int, int> >::const_iterator it_x = last_of_x.begin(); it_x != last_of_x.end(); ++it_x) {
					
					cout << "(" << it_x->first << ", " << it_x->second << "), ";

				}
				cout << endl;


/*
				cout << "Reduced candidates for sets containing 1: " ;


				for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;

				cout << "Reduced candidates for other summable elements: " ;
				for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;
*/

				cout << "----------------------------" << endl;

				{
					set<int> setCandidates[1 << MAX];
					const set<int>* op[MAX][MAX];
					int complement[MAX];

					// init
					for (int i = 0; i < 1 << (n + 2); i++) {
						setCandidates[i] = set<int>();
						hea::init_set_i(setCandidates[i], i);
					}
					

			/*
					for (int i = 0 ; i < 1 << n; i++) {
						cout << "Set " << i << ": ";
						disp_set(setCandidates[i]);		
						cout << endl;
					}
			*/
					// just to have initialize op
					for (int i = 0; i < n + 2; i++)
					for (int j = 0; j < n + 2; j++)
						op[i][j] = &setCandidates[0];
					

					// 0 + 0  = {0}
					op[0][0] = &setCandidates[1];

					complement[0] = n + 1;
					complement[n + 1] = 0;
				
					solve_candidates(candidates.begin(), candidates, last_of_x, n + 2, op, complement, setCandidates);
				}

				++it_perm;
				++it_zozn;
			}

			}


		}

		}
	}
	else {
		for (int i = 0; i < 3; i++) {
			switch (i) {
				case 0:
					relation[x][y] = 0;
					relation[y][x] = 0;
					break;
				case 1:
					relation[x][y] = 1;
					relation[y][x] = 0;
					break;
				case 2:
					relation[x][y] = 0;
					relation[y][x] = 1;
					break;
			}

			if (y < n - 1)
				gen_rel(x, y + 1, n, relation, relations);
			else
				gen_rel(x + 1, x + 2, n, relation, relations);
		}
	}
}

void gen_rel_ith(int x, int y, const int n, int (&relation)[MAX][MAX], vector<Relation> &relations, int ith, int &current_count) {
	if (x == n - 1) {

		if (admits_reverse_of_relation(n, relation)) {

			if (!is_isomorphic(n, relation, relations)) {

				/* disp relation */

/*
				for (int i = 0; i < n; i++) {
					for (int j = 0; j < n; j++)
						cerr << relation[i][j];
					cerr << endl;
				}
				cerr << "-R-"<<endl;
*/

				relations.push_back(Relation(n, relation));

				{

//					vector< pair< set< pair<int,int> >, set< pair<int,int> > > > HHHH;
					vector<Zoznamy> HHHH;
					vector<hea::Permutation> IIII;


					vector<hea::Permutation> vec_ortho;

					get_vector_of_orthosupplement_candidates(n, relation, vec_ortho);

					for (vector<hea::Permutation>::const_iterator it = vec_ortho.begin(); it != vec_ortho.end(); ++it) {

					set<pair<int, int> > one_pairs;

					for (int i = 0; i < n; i++) {
						one_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[i]));
						if (i != it->permutation[i])
							one_pairs.insert(pair<int, int>(it->permutation[i] + 1, 1 + i));

					}

/*
					if (one_pairs.size() == (unsigned int) n) {
*/

					set<pair<int, int> > other_pairs;
					
					for (int i = 0; i < n; i++)
					for (int j = 0; j < n; j++)
						if (i != j && relation[i][j]) {

							other_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[j]));
							if (i != it->permutation[j])

								other_pairs.insert(pair<int, int>(1 + it->permutation[j], i + 1));
						}




					if (!is_similar_vec_pairs(n, HHHH, one_pairs, other_pairs)) {
					
						IIII.push_back(hea::Permutation(*it));
						HHHH.push_back(Zoznamy(one_pairs, other_pairs));

					}

				}
/*
			cout << "IIII size: " << IIII.size() << endl;
			cout << "HHHH size: " << HHHH.size() << endl;
*/
			/* display candidates */
			cerr << "----------------------------" << endl;
			vector<hea::Permutation>::const_iterator it_perm = IIII.begin();
			vector<Zoznamy>::const_iterator it_zozn = HHHH.begin();
			while (it_perm != IIII.end()) {
				current_count++;
				if (ith == current_count)
				{

					cerr << "Candidate: ";
					for (int i = 0; i < n; i++)
						cerr << 1 + it_perm->permutation[i]; 
					cerr << endl;

					cerr << "Candidates for sets containing 1: " ;


					for (set<pair<int, int> >::const_iterator it_set = it_zozn->one_pairs.begin(); it_set != it_zozn->one_pairs.end(); ++it_set) 
						cerr << "(" << it_set->first << ", " << it_set->second << "), ";

					cerr << endl;

					cerr << "Candidates for other summable elements: " ;
					for (set<pair<int, int> >::const_iterator it_set = it_zozn->other_pairs.begin(); it_set != it_zozn->other_pairs.end(); ++it_set) 
						cerr << "(" << it_set->first << ", " << it_set->second << "), ";

					cerr << endl;



					Zoznamy reduced;
					reduce_summable_pairs(*it_zozn, reduced);

					set<Candidate> candidates;

					for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
						candidates.insert(Candidate(*it_set, true));

					for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 

						candidates.insert(Candidate(*it_set, false));

					/* insert pairs (0, x) into candidates */
					for (int i = 1; i <=n; i++)
						candidates.insert(Candidate(pair<int,int>(0, i), false));
					candidates.insert(Candidate(pair<int, int>(0, n + 1), true));


					/* display candidates */

					cerr << "Candidates: ";

					for (set<Candidate>::const_iterator it_cand = candidates.begin(); it_cand != candidates.end(); ++it_cand) {
						
						cerr << "(" << it_cand->candidate_pair.first << ", " << it_cand->candidate_pair.second << ")";
						if (it_cand->is_one)
							cerr << "*";

						cerr << ", ";

					}
					cerr << endl;

					

					set<pair<int, int> > last_of_x;
					init_last_cand_for_x_set(n, candidates, last_of_x); 
					cerr << "Last of x pairs: " ;
					
					for (set<pair<int, int> >::const_iterator it_x = last_of_x.begin(); it_x != last_of_x.end(); ++it_x) {
						
						cerr << "(" << it_x->first << ", " << it_x->second << "), ";

					}
					cerr << endl;
		


/*
				cout << "Reduced candidates for sets containing 1: " ;


				for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;

				cout << "Reduced candidates for other summable elements: " ;
				for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;
*/

					cerr << "----------------------------" << endl;

					cerr << "\t ***" << endl;
					set<int> setCandidates[1 << MAX];
					const set<int>* op[MAX][MAX];
					int complement[MAX];

					// init
					for (int i = 0; i < 1 << (n + 2); i++) {
						setCandidates[i] = set<int>();
						hea::init_set_i(setCandidates[i], i);
					}
					

			/*
					for (int i = 0 ; i < 1 << n; i++) {
						cout << "Set " << i << ": ";
						disp_set(setCandidates[i]);		
						cout << endl;
					}
			*/
					// just to have initialize op
					for (int i = 0; i < n + 2; i++)
					for (int j = 0; j < n + 2; j++)
						op[i][j] = &setCandidates[0];
					

					// 0 + 0  = {0}
					op[0][0] = &setCandidates[1];

					complement[0] = n + 1;
					complement[n + 1] = 0;

					/* init complement */
					for (set<Candidate>::const_iterator it_compl = candidates.begin(); it_compl != candidates.end(); ++it_compl)
						if (it_compl->is_one) {
							complement[it_compl->candidate_pair.first] = it_compl->candidate_pair.second;
							if (it_compl->candidate_pair.first != it_compl->candidate_pair.second)
								complement[it_compl->candidate_pair.second] = it_compl->candidate_pair.first;
						}
					/**/
				
					solve_candidates(candidates.begin(), candidates, last_of_x, n + 2, op, complement, setCandidates);
				}

				++it_perm;
				++it_zozn;
			}

			}


		}

		}
	}
	else {
		for (int i = 0; i < 3; i++) {
			switch (i) {
				case 0:
					relation[x][y] = 0;
					relation[y][x] = 0;
					break;
				case 1:
					relation[x][y] = 1;
					relation[y][x] = 0;
					break;
				case 2:
					relation[x][y] = 0;
					relation[y][x] = 1;
					break;
			}

			if (y < n - 1)
				gen_rel_ith(x, y + 1, n, relation, relations, ith, current_count);
			else
				gen_rel_ith(x + 1, x + 2, n, relation, relations, ith, current_count);
		}
	}

}

void gen_rel_ith_from_state(int x, int y, const int n, int (&relation)[MAX][MAX], vector<Relation> &relations, int ith, int &current_count, int *from_state, bool *used_from_state) {
	if (x == n - 1) {

		if (admits_reverse_of_relation(n, relation)) {

			if (!is_isomorphic(n, relation, relations)) {

				/* disp relation */

/*
				for (int i = 0; i < n; i++) {
					for (int j = 0; j < n; j++)
						cerr << relation[i][j];
					cerr << endl;
				}
				cerr << "-R-"<<endl;
*/

				relations.push_back(Relation(n, relation));

				{

//					vector< pair< set< pair<int,int> >, set< pair<int,int> > > > HHHH;
					vector<Zoznamy> HHHH;
					vector<hea::Permutation> IIII;


					vector<hea::Permutation> vec_ortho;

					get_vector_of_orthosupplement_candidates(n, relation, vec_ortho);

					for (vector<hea::Permutation>::const_iterator it = vec_ortho.begin(); it != vec_ortho.end(); ++it) {

					set<pair<int, int> > one_pairs;

					for (int i = 0; i < n; i++) {
						one_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[i]));
						if (i != it->permutation[i])
							one_pairs.insert(pair<int, int>(it->permutation[i] + 1, 1 + i));

					}

/*
					if (one_pairs.size() == (unsigned int) n) {
*/

					set<pair<int, int> > other_pairs;
					
					for (int i = 0; i < n; i++)
					for (int j = 0; j < n; j++)
						if (i != j && relation[i][j]) {

							other_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[j]));
							if (i != it->permutation[j])

								other_pairs.insert(pair<int, int>(1 + it->permutation[j], i + 1));
						}




					if (!is_similar_vec_pairs(n, HHHH, one_pairs, other_pairs)) {
					
						IIII.push_back(hea::Permutation(*it));
						HHHH.push_back(Zoznamy(one_pairs, other_pairs));

					}

				}
/*
			cout << "IIII size: " << IIII.size() << endl;
			cout << "HHHH size: " << HHHH.size() << endl;
*/
			/* display candidates */
			cerr << "----------------------------" << endl;
			vector<hea::Permutation>::const_iterator it_perm = IIII.begin();
			vector<Zoznamy>::const_iterator it_zozn = HHHH.begin();
			while (it_perm != IIII.end()) {
				current_count++;
				if (ith == current_count)
				{

					cerr << "Candidate: ";
					for (int i = 0; i < n; i++)
						cerr << 1 + it_perm->permutation[i]; 
					cerr << endl;

					cerr << "Candidates for sets containing 1: " ;


					for (set<pair<int, int> >::const_iterator it_set = it_zozn->one_pairs.begin(); it_set != it_zozn->one_pairs.end(); ++it_set) 
						cerr << "(" << it_set->first << ", " << it_set->second << "), ";

					cerr << endl;

					cerr << "Candidates for other summable elements: " ;
					for (set<pair<int, int> >::const_iterator it_set = it_zozn->other_pairs.begin(); it_set != it_zozn->other_pairs.end(); ++it_set) 
						cerr << "(" << it_set->first << ", " << it_set->second << "), ";

					cerr << endl;



					Zoznamy reduced;
					reduce_summable_pairs(*it_zozn, reduced);

					set<Candidate> candidates;

					for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
						candidates.insert(Candidate(*it_set, true));

					for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 

						candidates.insert(Candidate(*it_set, false));

					/* insert pairs (0, x) into candidates */
					for (int i = 1; i <=n; i++)
						candidates.insert(Candidate(pair<int,int>(0, i), false));
					candidates.insert(Candidate(pair<int, int>(0, n + 1), true));


					/* display candidates */

					cerr << "Candidates: ";

					for (set<Candidate>::const_iterator it_cand = candidates.begin(); it_cand != candidates.end(); ++it_cand) {
						
						cerr << "(" << it_cand->candidate_pair.first << ", " << it_cand->candidate_pair.second << ")";
						if (it_cand->is_one)
							cerr << "*";

						cerr << ", ";

					}
					cerr << endl;
					
					set<pair<int, int> > last_of_x;
					init_last_cand_for_x_set(n, candidates, last_of_x); 
					cerr << "Last of x pairs: " ;
					
					for (set<pair<int, int> >::const_iterator it_x = last_of_x.begin(); it_x != last_of_x.end(); ++it_x) {
						
						cerr << "(" << it_x->first << ", " << it_x->second << "), ";

					}
					cerr << endl;
		


/*
				cout << "Reduced candidates for sets containing 1: " ;


				for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;

				cout << "Reduced candidates for other summable elements: " ;
				for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;
*/

					cerr << "----------------------------" << endl;

					cerr << "\t ***" << endl;
					set<int> setCandidates[1 << MAX];
					const set<int>* op[MAX][MAX];
					int complement[MAX];

					// init
					for (int i = 0; i < 1 << (n + 2); i++) {
						setCandidates[i] = set<int>();
						hea::init_set_i(setCandidates[i], i);
					}
					

			/*
					for (int i = 0 ; i < 1 << n; i++) {
						cout << "Set " << i << ": ";
						disp_set(setCandidates[i]);		
						cout << endl;
					}
			*/
					// just to have initialize op
					for (int i = 0; i < n + 2; i++)
					for (int j = 0; j < n + 2; j++)
						op[i][j] = &setCandidates[0];
					

					// 0 + 0  = {0}
					op[0][0] = &setCandidates[1];

					complement[0] = n + 1;
					complement[n + 1] = 0;

					/* init complement */
					for (set<Candidate>::const_iterator it_compl = candidates.begin(); it_compl != candidates.end(); ++it_compl)
						if (it_compl->is_one) {
							complement[it_compl->candidate_pair.first] = it_compl->candidate_pair.second;
							if (it_compl->candidate_pair.first != it_compl->candidate_pair.second)
								complement[it_compl->candidate_pair.second] = it_compl->candidate_pair.first;
						}
					/**/
				
					solve_candidates_from_state(candidates.begin(), 0, candidates, last_of_x, n + 2, op, complement, setCandidates, from_state, used_from_state);
				}

				++it_perm;
				++it_zozn;
			}

			}


		}

		}
	}
	else {
		for (int i = 0; i < 3; i++) {
			switch (i) {
				case 0:
					relation[x][y] = 0;
					relation[y][x] = 0;
					break;
				case 1:
					relation[x][y] = 1;
					relation[y][x] = 0;
					break;
				case 2:
					relation[x][y] = 0;
					relation[y][x] = 1;
					break;
			}

			if (y < n - 1)
				gen_rel_ith_from_state(x, y + 1, n, relation, relations, ith, current_count, from_state, used_from_state);
			else
				gen_rel_ith_from_state(x + 1, x + 2, n, relation, relations, ith, current_count, from_state, used_from_state);
		}
	}

}

void gen_rel_ith_from_state_to_state(int x, int y, const int n, int (&relation)[MAX][MAX], vector<Relation> &relations, int ith, int &current_count, int *from_state, bool *used_from_state, int *to_state, bool *used_to_state) {
	if (x == n - 1) {

		if (admits_reverse_of_relation(n, relation)) {

			if (!is_isomorphic(n, relation, relations)) {

				/* disp relation */

/*
				for (int i = 0; i < n; i++) {
					for (int j = 0; j < n; j++)
						cerr << relation[i][j];
					cerr << endl;
				}
				cerr << "-R-"<<endl;
*/

				relations.push_back(Relation(n, relation));

				{

//					vector< pair< set< pair<int,int> >, set< pair<int,int> > > > HHHH;
					vector<Zoznamy> HHHH;
					vector<hea::Permutation> IIII;


					vector<hea::Permutation> vec_ortho;

					get_vector_of_orthosupplement_candidates(n, relation, vec_ortho);

					for (vector<hea::Permutation>::const_iterator it = vec_ortho.begin(); it != vec_ortho.end(); ++it) {

					set<pair<int, int> > one_pairs;

					for (int i = 0; i < n; i++) {
						one_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[i]));
						if (i != it->permutation[i])
							one_pairs.insert(pair<int, int>(it->permutation[i] + 1, 1 + i));

					}

/*
					if (one_pairs.size() == (unsigned int) n) {
*/

					set<pair<int, int> > other_pairs;
					
					for (int i = 0; i < n; i++)
					for (int j = 0; j < n; j++)
						if (i != j && relation[i][j]) {

							other_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[j]));
							if (i != it->permutation[j])

								other_pairs.insert(pair<int, int>(1 + it->permutation[j], i + 1));
						}




					if (!is_similar_vec_pairs(n, HHHH, one_pairs, other_pairs)) {
					
						IIII.push_back(hea::Permutation(*it));
						HHHH.push_back(Zoznamy(one_pairs, other_pairs));

					}

				}
/*
			cout << "IIII size: " << IIII.size() << endl;
			cout << "HHHH size: " << HHHH.size() << endl;
*/
			/* display candidates */
			cerr << "----------------------------" << endl;
			vector<hea::Permutation>::const_iterator it_perm = IIII.begin();
			vector<Zoznamy>::const_iterator it_zozn = HHHH.begin();
			while (it_perm != IIII.end()) {
				current_count++;
				if (ith == current_count)
				{

					cerr << "Candidate: ";
					for (int i = 0; i < n; i++)
						cerr << 1 + it_perm->permutation[i]; 
					cerr << endl;

					cerr << "Candidates for sets containing 1: " ;


					for (set<pair<int, int> >::const_iterator it_set = it_zozn->one_pairs.begin(); it_set != it_zozn->one_pairs.end(); ++it_set) 
						cerr << "(" << it_set->first << ", " << it_set->second << "), ";

					cerr << endl;

					cerr << "Candidates for other summable elements: " ;
					for (set<pair<int, int> >::const_iterator it_set = it_zozn->other_pairs.begin(); it_set != it_zozn->other_pairs.end(); ++it_set) 
						cerr << "(" << it_set->first << ", " << it_set->second << "), ";

					cerr << endl;



					Zoznamy reduced;
					reduce_summable_pairs(*it_zozn, reduced);

					set<Candidate> candidates;

					for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
						candidates.insert(Candidate(*it_set, true));

					for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 

						candidates.insert(Candidate(*it_set, false));

					/* insert pairs (0, x) into candidates */
					for (int i = 1; i <=n; i++)
						candidates.insert(Candidate(pair<int,int>(0, i), false));
					candidates.insert(Candidate(pair<int, int>(0, n + 1), true));


					/* display candidates */

					cerr << "Candidates: ";

					for (set<Candidate>::const_iterator it_cand = candidates.begin(); it_cand != candidates.end(); ++it_cand) {
						
						cerr << "(" << it_cand->candidate_pair.first << ", " << it_cand->candidate_pair.second << ")";
						if (it_cand->is_one)
							cerr << "*";

						cerr << ", ";

					}
					cerr << endl;
					
					set<pair<int, int> > last_of_x;
					init_last_cand_for_x_set(n, candidates, last_of_x); 
					cerr << "Last of x pairs: " ;
					
					for (set<pair<int, int> >::const_iterator it_x = last_of_x.begin(); it_x != last_of_x.end(); ++it_x) {
						
						cerr << "(" << it_x->first << ", " << it_x->second << "), ";

					}
					cerr << endl;
		


/*
				cout << "Reduced candidates for sets containing 1: " ;


				for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;

				cout << "Reduced candidates for other summable elements: " ;
				for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;
*/

					cerr << "----------------------------" << endl;

					cerr << "\t ***" << endl;
					set<int> setCandidates[1 << MAX];
					const set<int>* op[MAX][MAX];
					int complement[MAX];

					// init
					for (int i = 0; i < 1 << (n + 2); i++) {
						setCandidates[i] = set<int>();
						hea::init_set_i(setCandidates[i], i);
					}
					

			/*
					for (int i = 0 ; i < 1 << n; i++) {
						cout << "Set " << i << ": ";
						disp_set(setCandidates[i]);		
						cout << endl;
					}
			*/
					// just to have initialize op
					for (int i = 0; i < n + 2; i++)
					for (int j = 0; j < n + 2; j++)
						op[i][j] = &setCandidates[0];
					

					// 0 + 0  = {0}
					op[0][0] = &setCandidates[1];

					complement[0] = n + 1;
					complement[n + 1] = 0;

					/* init complement */
					for (set<Candidate>::const_iterator it_compl = candidates.begin(); it_compl != candidates.end(); ++it_compl)
						if (it_compl->is_one) {
							complement[it_compl->candidate_pair.first] = it_compl->candidate_pair.second;
							if (it_compl->candidate_pair.first != it_compl->candidate_pair.second)
								complement[it_compl->candidate_pair.second] = it_compl->candidate_pair.first;
						}
					/**/
				
					solve_candidates_from_state_to_state(candidates.begin(), 0, candidates, last_of_x, n + 2, op, complement, setCandidates, from_state, used_from_state, to_state, used_to_state);
				}

				++it_perm;
				++it_zozn;
			}

			}


		}

		}
	}
	else {
		for (int i = 0; i < 3; i++) {
			switch (i) {
				case 0:
					relation[x][y] = 0;
					relation[y][x] = 0;
					break;
				case 1:
					relation[x][y] = 1;
					relation[y][x] = 0;
					break;
				case 2:
					relation[x][y] = 0;
					relation[y][x] = 1;
					break;
			}

			if (y < n - 1)
				gen_rel_ith_from_state_to_state(x, y + 1, n, relation, relations, ith, current_count, from_state, used_from_state, to_state, used_to_state);
			else
				gen_rel_ith_from_state_to_state(x + 1, x + 2, n, relation, relations, ith, current_count, from_state, used_from_state, to_state, used_to_state);
		}
	}

}

void gen_rel_ith_state_setup(int x, int y, const int n, int (&relation)[MAX][MAX], vector<Relation> &relations, int ith, int &current_count, fromToState &current_setup) {
	if (x == n - 1) {

		if (admits_reverse_of_relation(n, relation)) {

			if (!is_isomorphic(n, relation, relations)) {

				/* disp relation */

/*
				for (int i = 0; i < n; i++) {
					for (int j = 0; j < n; j++)
						cerr << relation[i][j];
					cerr << endl;
				}
				cerr << "-R-"<<endl;
*/

				relations.push_back(Relation(n, relation));

				{

//					vector< pair< set< pair<int,int> >, set< pair<int,int> > > > HHHH;
					vector<Zoznamy> HHHH;
					vector<hea::Permutation> IIII;


					vector<hea::Permutation> vec_ortho;

					get_vector_of_orthosupplement_candidates(n, relation, vec_ortho);

					for (vector<hea::Permutation>::const_iterator it = vec_ortho.begin(); it != vec_ortho.end(); ++it) {

					set<pair<int, int> > one_pairs;

					for (int i = 0; i < n; i++) {
						one_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[i]));
						if (i != it->permutation[i])
							one_pairs.insert(pair<int, int>(it->permutation[i] + 1, 1 + i));

					}

/*
					if (one_pairs.size() == (unsigned int) n) {
*/

					set<pair<int, int> > other_pairs;
					
					for (int i = 0; i < n; i++)
					for (int j = 0; j < n; j++)
						if (i != j && relation[i][j]) {

							other_pairs.insert(pair<int, int>(i + 1, 1 + it->permutation[j]));
							if (i != it->permutation[j])

								other_pairs.insert(pair<int, int>(1 + it->permutation[j], i + 1));
						}




					if (!is_similar_vec_pairs(n, HHHH, one_pairs, other_pairs)) {
					
						IIII.push_back(hea::Permutation(*it));
						HHHH.push_back(Zoznamy(one_pairs, other_pairs));

					}

				}
/*
			cout << "IIII size: " << IIII.size() << endl;
			cout << "HHHH size: " << HHHH.size() << endl;
*/
			/* display candidates */
			cerr << "----------------------------" << endl;
			vector<hea::Permutation>::const_iterator it_perm = IIII.begin();
			vector<Zoznamy>::const_iterator it_zozn = HHHH.begin();
			while (it_perm != IIII.end()) {
				current_count++;
				if (ith == current_count)
				{

					cerr << "Candidate: ";
					for (int i = 0; i < n; i++)
						cerr << 1 + it_perm->permutation[i]; 
					cerr << endl;

					cerr << "Candidates for sets containing 1: " ;


					for (set<pair<int, int> >::const_iterator it_set = it_zozn->one_pairs.begin(); it_set != it_zozn->one_pairs.end(); ++it_set) 
						cerr << "(" << it_set->first << ", " << it_set->second << "), ";

					cerr << endl;

					cerr << "Candidates for other summable elements: " ;
					for (set<pair<int, int> >::const_iterator it_set = it_zozn->other_pairs.begin(); it_set != it_zozn->other_pairs.end(); ++it_set) 
						cerr << "(" << it_set->first << ", " << it_set->second << "), ";

					cerr << endl;



					Zoznamy reduced;
					reduce_summable_pairs(*it_zozn, reduced);

					set<Candidate> candidates;

					for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
						candidates.insert(Candidate(*it_set, true));

					for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 

						candidates.insert(Candidate(*it_set, false));

					/* insert pairs (0, x) into candidates */
					for (int i = 1; i <=n; i++)
						candidates.insert(Candidate(pair<int,int>(0, i), false));
					candidates.insert(Candidate(pair<int, int>(0, n + 1), true));


					/* display candidates */

					cerr << "Candidates: ";

					for (set<Candidate>::const_iterator it_cand = candidates.begin(); it_cand != candidates.end(); ++it_cand) {
						
						cerr << "(" << it_cand->candidate_pair.first << ", " << it_cand->candidate_pair.second << ")";
						if (it_cand->is_one)
							cerr << "*";

						cerr << ", ";

					}
					cerr << endl;
					
					set<pair<int, int> > last_of_x;
					init_last_cand_for_x_set(n, candidates, last_of_x); 
					cerr << "Last of x pairs: " ;
					
					for (set<pair<int, int> >::const_iterator it_x = last_of_x.begin(); it_x != last_of_x.end(); ++it_x) {
						
						cerr << "(" << it_x->first << ", " << it_x->second << "), ";

					}
					cerr << endl;
		


/*
				cout << "Reduced candidates for sets containing 1: " ;


				for (set<pair<int, int> >::const_iterator it_set = reduced.one_pairs.begin(); it_set != reduced.one_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;

				cout << "Reduced candidates for other summable elements: " ;
				for (set<pair<int, int> >::const_iterator it_set = reduced.other_pairs.begin(); it_set != reduced.other_pairs.end(); ++it_set) 
					cout << "(" << it_set->first << ", " << it_set->second << "), ";

				cout << endl;
*/

					cerr << "----------------------------" << endl;

					cerr << "\t ***" << endl;
					set<int> setCandidates[1 << MAX];
					const set<int>* op[MAX][MAX];
					int complement[MAX];

					// init
					for (int i = 0; i < 1 << (n + 2); i++) {
						setCandidates[i] = set<int>();
						hea::init_set_i(setCandidates[i], i);
					}
					

			/*
					for (int i = 0 ; i < 1 << n; i++) {
						cout << "Set " << i << ": ";
						disp_set(setCandidates[i]);		
						cout << endl;
					}
			*/
					// just to have initialize op
					for (int i = 0; i < n + 2; i++)
					for (int j = 0; j < n + 2; j++)
						op[i][j] = &setCandidates[0];
					

					// 0 + 0  = {0}
					op[0][0] = &setCandidates[1];

					complement[0] = n + 1;
					complement[n + 1] = 0;

					/* init complement */
					for (set<Candidate>::const_iterator it_compl = candidates.begin(); it_compl != candidates.end(); ++it_compl)
						if (it_compl->is_one) {
							complement[it_compl->candidate_pair.first] = it_compl->candidate_pair.second;
							if (it_compl->candidate_pair.first != it_compl->candidate_pair.second)
								complement[it_compl->candidate_pair.second] = it_compl->candidate_pair.first;
						}
					/**/
				
					solve_candidates_state_setup(candidates.begin(), 0, candidates, last_of_x, n + 2, op, complement, setCandidates, current_setup);
				}

				++it_perm;
				++it_zozn;
			}

			}


		}

		}
	}
	else {
		for (int i = 0; i < 3; i++) {
			switch (i) {
				case 0:
					relation[x][y] = 0;
					relation[y][x] = 0;
					break;
				case 1:
					relation[x][y] = 1;
					relation[y][x] = 0;
					break;
				case 2:
					relation[x][y] = 0;
					relation[y][x] = 1;
					break;
			}

			if (y < n - 1)
				gen_rel_ith_state_setup(x, y + 1, n, relation, relations, ith, current_count, current_setup);
			else
				gen_rel_ith_state_setup(x + 1, x + 2, n, relation, relations, ith, current_count, current_setup);
		}
	}

}




int main(int argc, char **argv) {

	fromToState current_state_setup;

	//bool bFromState = false;
	//bool bToState = false;
	if (argc < 2) {
		cout << "Usage: " << argv[0] << " <n> [<from_state> [<to_state>]]" << endl
		     << "\t<n> - computes models corresponding to the <n>-th model, (n = 1-...)" << endl
		     << "\t<from_state> - comma separated list of numbers" << endl 
		     << "\t<to_state> - comma separated list of numbers" << endl;
		return 0;
	}

	//int from_state[MAX*MAX];
	//bool used_from_state[MAX*MAX];

	//int to_state[MAX*MAX];
	//bool used_to_state[MAX*MAX];

	//for (int i = 0; i < MAX*MAX ; i++)
		//used_from_state[i] = false;

	//for (int i = 0; i < MAX*MAX; i++)
		//used_to_state[i] = false;

	if (argc >= 3) {
		current_state_setup.bFromState = true;
		char *pom = argv[2];
		int idx = 0;

		while (*pom) {
			int val = 0;
			while (*pom>='0' && *pom<='9') {
				val=10*val + (int)(*pom-'0');
				pom++;
			}
			if (*pom==',') 
				pom++;
			else 
				if (*pom != '\0') {
					cerr << "Invalid character '"<< *pom << "'." << endl;
					return -1;
				}
			
			current_state_setup.from_state[idx] = val;
			current_state_setup.used_from_state[idx] = true;
			idx++;
		}
	}

	if (argc == 4) {
		current_state_setup.bToState = true;
		char *pom = argv[3];
		int idx = 0;

		while (*pom) {
			int val = 0;
			while (*pom>='0' && *pom<='9') {
				val=10*val + (int)(*pom-'0');
				pom++;
			}
			if (*pom==',') 
				pom++;
			else 
				if (*pom != '\0') {
					cerr << "Invalid character '"<< *pom << "'." << endl;
					return -1;
				}
			
			current_state_setup.to_state[idx] = val;
			current_state_setup.used_to_state[idx] = true;
			idx++;
		}
	}
		

	int relation[MAX][MAX];
	const int n = 4;
	vector<Relation> relations;

	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		relation[i][j] = 0;

	for (int i = 0; i < n; i++)
		relation[i][i] = 1;

	int current_count = 0;

	if (current_state_setup.bFromState) {
		cerr << "Computing " << argv[1] << "th model\n\t from state: " << endl;
		int idx = 0;
		while (idx < MAX*MAX && current_state_setup.used_from_state[idx]) {
			cerr << current_state_setup.from_state[idx] << ", ";
			idx++;
		}
		cerr << endl;
		
		if (current_state_setup.bToState) {
			cerr << "\t to state: " << endl;
			idx = 0;
			while (idx < MAX*MAX && current_state_setup.used_to_state[idx]) {
				cerr << current_state_setup.to_state[idx] << ", ";
				idx++;
			}
			cerr << endl;

			//gen_rel_ith_from_state_to_state(0, 1, n, relation, relations, atoi(argv[1]), current_count, from_state, used_from_state, to_state, used_to_state);

		}
		//else 
			//gen_rel_ith_from_state(0, 1, n, relation, relations, atoi(argv[1]), current_count, from_state, used_from_state);

	}
	//else
		//gen_rel_ith(0, 1, n, relation, relations, atoi(argv[1]), current_count);

	time_t time_start;
	time_t time_end;
	time(&time_start);
	gen_rel_ith_state_setup(0, 1, n, relation, relations, atoi(argv[1]), current_count, current_state_setup);
	time(&time_end);

	cerr << "Last possibilities: " << number_of_tested_configurations << endl;
	cerr << "Start time: " << time_start << endl;
	cerr << "End   time: " << time_end << endl;


/*
	map<pair<int, int>, bool> is_last_of_x;
	is_last_of_x.insert(pair<pair<int, int>, bool>(pair<int, int>(1,2), true));
*/

}


